#include "common.h"
#include "common.c"

scalar ComputeEG2c(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar t    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar Psi1 = psiParam[0];
	
	return -((beta*t)/Psi1);
}

scalar ComputeEG2G2cAW(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	
	return -(((-phi + 2*Psi1)*(2*Psi1 - Psi2)*Exponential(-(Psi1*r))*Power2(beta)*Power2(-1 + Exponential(Psi1*r))*Power4(1/Psi1))/(phi*Psi2));
}

scalar ComputeEG2G2c(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return Power2(ComputeEG2c(hParam,thetaParam,psiParam));
}

scalar ComputeEG4c(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar t    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	
	return (Power2(beta)*Power2(1/phi)*(12*phi*Psi1*(2*Psi1 - Psi2)*(1 + Psi1*t - Exponential(Psi1*t)) + 3*Power2(phi)*(-2*Psi2*(-1 + Exponential(Psi1*t)) + 2*Psi1*(-2 + Psi2*t + 2*Exponential(Psi1*t)) + t*(-4 + Psi2*t)*Power2(Psi1)) + 2*(-2*Psi1 + Psi2)*t*Power3(Psi1))*Power4(1/Psi1))/Psi2;
}

scalar ComputeEG2G2G2cAWAD(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	
	return ((-phi + 2*Psi1)*(2*Psi1 - Psi2)*r*Exponential(-(Psi1*r))*Power2(1/phi)*(-3*phi*Psi1 + phi*Psi2 - 2*Psi1*Psi2 + 4*Power2(Psi1))*Power2(-1 + Exponential(Psi1*r))*Power3(beta)*Power5(1/Psi1))/((Psi1 - Psi2)*Psi2);
}

scalar ComputeEG2G2G2cAWBD(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	
	return (2*(-phi + 2*Psi1 - Psi2)*(-phi + Psi1 + Psi2)*(3*Psi1 - 3*Psi2 + Psi3)*(-1 + Exponential(Psi1*r))*(-1 + Exponential(Psi2*r))*(-Exponential(Psi1*r) + Exponential(Psi2*r))*Exponential(-((Psi1 + Psi2)*r))*Power2(1/phi)*Power2(1/Psi1)*Power2(1/(Psi1 - Psi2))*Power2(1/Psi2)*Power3(beta))/Psi3;
}

scalar ComputeEG2G2G2cAW(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG2c(hParam,thetaParam,psiParam)*ComputeEG2G2cAW(hParam,thetaParam,psiParam);
}

scalar ComputeEG2G2G2cAD(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG2c(hParam,thetaParam,psiParam)*ComputeEG2G2cAW(hParam,thetaParam,psiParam);
}

scalar ComputeEG2G2G2c(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return Power3(ComputeEG2c(hParam,thetaParam,psiParam));
}

scalar ComputeEG4G2cAD(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	
	return (2*(-phi + 2*Psi1)*(2*Psi1 - Psi2)*Exponential(-(Psi1*r))*(-1 + Exponential(Psi1*r))*(6*phi*Psi1*(2*Psi1 - Psi2)*(1 + (-1 + Psi1*r)*Exponential(Psi1*r)) + (2*Psi1 - Psi2)*(-1 + Exponential(Psi1*r))*Power2(Psi1) - 3*Power2(phi)*(2*Psi2*(-1 + Exponential(Psi1*r)) - Psi1*(-4 + Psi2*r + (4 + Psi2*r)*Exponential(Psi1*r)) + r*(1 + 3*Exponential(Psi1*r))*Power2(Psi1)))*Power3(beta)*Power3(1/phi)*Power6(1/Psi1))/((Psi1 - Psi2)*Psi2);
}

scalar ComputeEG4G2cBD(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	
	return (2*(-phi + Psi1 + Psi2)*(-3*Psi1 + 3*Psi2 - Psi3)*Exponential(-(Psi2*r))*(-1 + Exponential(Psi2*r))*Power2(1/Psi1)*Power2(1/(Psi1 - Psi2))*(-6*phi*(2*Psi1 - Psi2)*(-(Psi2*(-1 + Exponential(Psi1*r))) + Psi1*(-1 + Exponential(Psi2*r))) + 6*(-(Psi2*(-1 + Exponential(Psi1*r))) + Psi1*(-1 + Exponential(Psi2*r)))*Power2(phi) + Psi1*(-1 + Exponential(Psi2*r))*(-3*Psi1*Psi2 + 2*Power2(Psi1) + Power2(Psi2)))*Power3(beta)*Power3(1/phi)*Power3(1/Psi2))/Psi3;
}

scalar ComputeEG4G2c(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG2c(hParam,thetaParam,psiParam)*ComputeEG4c(hParam,thetaParam,psiParam);
}

scalar ComputeEG2G4cAW(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	
	return (2*Exponential(-(Psi1*r))*(-1 + Exponential(Psi1*r))*Power2(1/phi)*Power2(1/(Psi1 - Psi2))*Power2(1/Psi2)*Power3(beta)*(3*Power2(phi)*(2*Psi3*(-8 + 2*Psi2*r + (8 + 3*Psi2*r)*Exponential(Psi1*r))*Power2(Psi1)*Power2(Psi2) - Psi2*(10*Psi3*(-1 + Exponential(Psi1*r)) + Psi2*(6 + 5*Psi3*r + (-6 + 11*Psi3*r)*Exponential(Psi1*r)))*Power3(Psi1) - Psi1*Psi3*(-10 + Psi2*r + (10 + Psi2*r)*Exponential(Psi1*r))*Power3(Psi2) + 2*(Psi2*(6 + Psi3*r + 3*(-1 + Psi3*r)*Exponential(Psi1*r) - 3*Exponential(Psi2*r)) + Psi3*(-1 + Exponential(Psi2*r)))*Power4(Psi1) + 2*Psi3*(-1 + Exponential(Psi1*r))*Power4(Psi2) + 6*(-1 + Exponential(Psi2*r))*Power5(Psi1)) - 6*phi*Psi1*(Psi3*(4*(-6 + Psi2*r) + (24 + 11*Psi2*r)*Exponential(Psi1*r))*Power2(Psi1)*Power2(Psi2) - Psi2*(15*Psi3*(-1 + Exponential(Psi1*r)) + Psi2*(9 + 5*Psi3*r + (-9 + 19*Psi3*r)*Exponential(Psi1*r)))*Power3(Psi1) - Psi1*Psi3*(-15 + Psi2*r + (15 + 2*Psi2*r)*Exponential(Psi1*r))*Power3(Psi2) + (Psi2*(2*(9 + Psi3*r) + (-9 + 10*Psi3*r)*Exponential(Psi1*r) - 9*Exponential(Psi2*r)) + 3*Psi3*(-1 + Exponential(Psi2*r)))*Power4(Psi1) + 3*Psi3*(-1 + Exponential(Psi1*r))*Power4(Psi2) + 9*(-1 + Exponential(Psi2*r))*Power5(Psi1)) + Power2(Psi1)*((Psi3*(-95 + 101*Exponential(Psi1*r) - 6*Exponential(Psi2*r)) + 3*Psi2*(-13 + (7 + 20*Psi3*r)*Exponential(Psi1*r) + 6*Exponential(Psi2*r)))*Power2(Psi1)*Power2(Psi2) - Psi2*(Psi3*(-59 + 65*Exponential(Psi1*r) - 6*Exponential(Psi2*r)) + 6*Psi2*(-1 + (-5 + 16*Psi3*r)*Exponential(Psi1*r) + 6*Exponential(Psi2*r)))*Power3(Psi1) - Psi1*(61*Psi3*(-1 + Exponential(Psi1*r)) + 12*Psi2*(-1 + (1 + Psi3*r)*Exponential(Psi1*r)))*Power3(Psi2) + 3*(Psi2*(19 + (-13 + 16*Psi3*r)*Exponential(Psi1*r) - 6*Exponential(Psi2*r)) + 4*Psi3*(-1 + Exponential(Psi2*r)))*Power4(Psi1) + 13*Psi3*(-1 + Exponential(Psi1*r))*Power4(Psi2) + 36*(-1 + Exponential(Psi2*r))*Power5(Psi1)))*Power6(1/Psi1))/Psi3;
}

scalar ComputeEG2G4c(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG2c(hParam,thetaParam,psiParam)*ComputeEG4c(hParam,thetaParam,psiParam);
}

scalar ComputeEG6c(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar t    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	
	return (-3*Power2(1/(Psi1 - Psi2))*Power3(beta)*Power3(1/phi)*Power3(1/Psi2)*Power6(1/Psi1)*(2*(Psi1 - Psi2)*Power3(Psi1)*(-(Power2(Psi1)*Power2(Psi2)*(Psi2*(15 + 36*Psi3*t - 15*Exponential(Psi2*t)) + 5*Psi3*(-9 + 8*Exponential(Psi1*t) + Exponential(Psi2*t)) + 12*t*Power2(Psi2))) + Psi2*(5*Psi3*(-1 + Exponential(Psi2*t)) - 6*Psi2*(-5 - 6*Psi3*t + 5*Exponential(Psi2*t)) + 24*t*Power2(Psi2))*Power3(Psi1) + 10*Psi1*Psi3*(-4 + Psi2*t + 4*Exponential(Psi1*t))*Power3(Psi2) + (10*Psi3*(-1 + Exponential(Psi2*t)) - 5*Psi2*(-3 + 2*Psi3*t + 3*Exponential(Psi2*t)) + 18*t*Power2(Psi2))*Power4(Psi1) - 10*Psi3*(-1 + Exponential(Psi1*t))*Power4(Psi2) + 30*(-1 - Psi2*t + Exponential(Psi2*t))*Power5(Psi1)) + 60*Psi1*Power2(phi)*(Power2(Psi2)*(-27*Psi3*(-1 + Exponential(Psi1*t)) - Psi2*(9 + 29*Psi3*t + (-9 + 19*Psi3*t)*Exponential(Psi1*t)) + 4*Psi3*Power2(Psi2)*Power2(t))*Power3(Psi1) - Psi3*Power2(Psi1)*(48 - 19*Psi2*t - (48 + 11*Psi2*t)*Exponential(Psi1*t) + Power2(Psi2)*Power2(t))*Power3(Psi2) - Power2(Psi2)*(-9 + 9*Psi2*t - 17*Psi3*t + (9 - 10*Psi3*t)*Exponential(Psi1*t) + 5*Psi2*Psi3*Power2(t))*Power4(Psi1) - 2*Psi1*Psi3*(-15 + 2*Psi2*t + (15 + Psi2*t)*Exponential(Psi1*t))*Power4(Psi2) + (3*Psi3*(-1 + Exponential(Psi2*t)) - 3*Psi2*(-3 + Psi3*t + 3*Exponential(Psi2*t)) + 2*t*(9 + Psi3*t)*Power2(Psi2))*Power5(Psi1) + 6*Psi3*(-1 + Exponential(Psi1*t))*Power5(Psi2) + 9*(-1 - Psi2*t + Exponential(Psi2*t))*Power6(Psi1)) - 5*Power3(phi)*(-6*Psi3*Power2(Psi1)*(32 - 14*Psi2*t - 2*(16 + 3*Psi2*t)*Exponential(Psi1*t) + Power2(Psi2)*Power2(t))*Power3(Psi2) - Power2(Psi2)*Power3(Psi1)*(108*Psi3*(-1 + Exponential(Psi1*t)) + 6*Psi2*(6 + 21*Psi3*t + (-6 + 11*Psi3*t)*Exponential(Psi1*t)) - 24*Psi3*Power2(Psi2)*Power2(t) + Psi3*Power3(Psi2)*Power3(t)) + 2*Power2(Psi2)*(18 - 18*Psi2*t + 36*Psi3*t + 18*(-1 + Psi3*t)*Exponential(Psi1*t) - 15*Psi2*Psi3*Power2(t) + Psi3*Power2(Psi2)*Power3(t))*Power4(Psi1) - 6*Psi1*Psi3*(-20 + 3*Psi2*t + (20 + Psi2*t)*Exponential(Psi1*t))*Power4(Psi2) + (12*Psi3*(-1 + Exponential(Psi2*t)) - 12*Psi2*(-3 + Psi3*t + 3*Exponential(Psi2*t)) + 12*t*(6 + Psi3*t)*Power2(Psi2) - Psi3*Power3(Psi2)*Power3(t))*Power5(Psi1) + 24*Psi3*(-1 + Exponential(Psi1*t))*Power5(Psi2) + 36*(-1 - Psi2*t + Exponential(Psi2*t))*Power6(Psi1)) - 10*phi*Power2(Psi1)*(Power2(Psi2)*(-(Psi3*(-122 + 117*Exponential(Psi1*t) + 5*Exponential(Psi2*t))) + Psi2*(-45 - 104*Psi3*t + (30 - 96*Psi3*t)*Exponential(Psi1*t) + 15*Exponential(Psi2*t)) + 4*t*(-9 + Psi3*t)*Power2(Psi2))*Power3(Psi1) - Power2(Psi1)*(-205*Psi3*(-1 + Exponential(Psi1*t)) - 3*Psi2*(-7 + 22*Psi3*t + (7 + 20*Psi3*t)*Exponential(Psi1*t)) + t*(-12 + Psi3*t)*Power2(Psi2))*Power3(Psi2) - Psi2*(-3*Psi2*(21 + 22*Psi3*t + (-13 + 16*Psi3*t)*Exponential(Psi1*t) - 8*Exponential(Psi2*t)) - 3*Psi3*(-1 + Exponential(Psi2*t)) + t*(6 + 5*Psi3*t)*Power2(Psi2))*Power4(Psi1) - 2*Psi1*(63*Psi3*(-1 + Exponential(Psi1*t)) + Psi2*(-6 + 7*Psi3*t + 6*(1 + Psi3*t)*Exponential(Psi1*t)))*Power4(Psi2) + (Psi2*(33 - 14*Psi3*t - 33*Exponential(Psi2*t)) + 14*Psi3*(-1 + Exponential(Psi2*t)) + 2*t*(36 + Psi3*t)*Power2(Psi2))*Power5(Psi1) + 26*Psi3*(-1 + Exponential(Psi1*t))*Power5(Psi2) + 42*(-1 - Psi2*t + Exponential(Psi2*t))*Power6(Psi1))))/Psi3;
}

scalar ComputeEG2G2G2G2cAWADAH(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	
	return -(((-phi + 2*Psi1)*(2*Psi1 - Psi2)*Exponential(-(Psi1*r))*Power2(1/(Psi1 - Psi2))*Power2(r)*Power2(-1 + Exponential(Psi1*r))*Power2(-3*phi*Psi1 + phi*Psi2 - 2*Psi1*Psi2 + 4*Power2(Psi1))*Power3(1/phi)*Power4(beta)*Power6(1/Psi1))/Psi2);
}

scalar ComputeEG2G2G2G2cAWADBH(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	
	return (2*(-phi + 2*Psi1 - Psi2)*(-phi + Psi1 + Psi2)*(3*Psi1 - 3*Psi2 + Psi3)*r*(-1 + Exponential(Psi1*r))*(Exponential(Psi1*r) - Exponential(Psi2*r))*(-1 + Exponential(Psi2*r))*Exponential(-((Psi1 + Psi2)*r))*(-3*phi*Psi1 + phi*Psi2 - 2*Psi1*Psi2 + 4*Power2(Psi1))*Power2(1/Psi2)*Power3(1/phi)*Power3(1/Psi1)*Power3(1/(Psi1 - Psi2))*Power4(beta))/Psi3;
}

scalar ComputeEG2G2G2G2cAWAD(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG2c(hParam,thetaParam,psiParam)*ComputeEG2G2G2cAWAD(hParam,thetaParam,psiParam);
}

scalar ComputeEG2G2G2G2cAWAHBD(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	
	return (2*(-phi + 2*Psi1)*(2*Psi1 - Psi2)*(-phi + 2*Psi1 - Psi2)*(-phi + Psi2)*Exponential(-((2*Psi1 + Psi2)*r))*(Psi2*(3*Psi2 - Psi3) - Psi1*(2*Psi2 + Psi3) + Power2(Psi1))*Power2(1/Psi2)*Power2(-1 + Exponential(Psi1*r))*Power2(Exponential(Psi1*r) - Exponential(Psi2*r))*Power3(1/phi)*Power4(beta)*Power4(1/Psi1)*Power4(1/(Psi1 - Psi2)))/(Psi1 - Psi3);
}

scalar ComputeEG2G2G2G2cAWAH(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return Power2(ComputeEG2G2cAW(hParam,thetaParam,psiParam));
}

scalar ComputeEG2G2G2G2cAWBDBH(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	
	return (2*(-phi + 2*Psi1 - Psi2)*(-phi + Psi1 + Psi2)*(3*Psi1 - 3*Psi2 + Psi3)*r*(-1 + Exponential(Psi1*r))*(Exponential(Psi1*r) - Exponential(Psi2*r))*(-1 + Exponential(Psi2*r))*Exponential(-((Psi1 + Psi2)*r))*Power2(1/Psi1)*(phi*(-3*Psi1 + 5*Psi2 - 2*Psi3) + Psi1*(-4*Psi2 + Psi3) + Psi2*(-Psi2 + Psi3) + 3*Power2(Psi1))*Power2(1/Psi2)*Power3(1/phi)*Power3(1/(Psi1 - Psi2))*Power4(beta))/((Psi2 - Psi3)*Psi3);
}

scalar ComputeEG2G2G2G2cAWBDCH(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	const scalar Psi4 = psiParam[3];
	
	return (-6*(-phi + 2*Psi1 - Psi2)*(-phi + Psi1 + Psi2 - Psi3)*(-phi + Psi1 + Psi3)*(Psi3*(4*Psi3 - Psi4) - Psi2*(4*Psi3 + Psi4) + Psi1*(4*Psi2 - 4*Psi3 + 3*Psi4))*(-1 + Exponential(Psi1*r))*(Exponential(Psi1*r) - Exponential(Psi2*r))*(Exponential(Psi2*r) - Exponential(Psi3*r))*(-1 + Exponential(Psi3*r))*Exponential(-((Psi1 + Psi2 + Psi3)*r))*Power2(1/Psi1)*Power2(1/(Psi2 - Psi3))*Power2(1/Psi3)*Power3(1/phi)*Power4(beta))/((Psi1 - Psi2)*Psi2*(Psi1 - Psi3)*Psi4);
}

scalar ComputeEG2G2G2G2cAWBD(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG2c(hParam,thetaParam,psiParam)*ComputeEG2G2G2cAWBD(hParam,thetaParam,psiParam);
}

scalar ComputeEG2G2G2G2cAW(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG2G2cAW(hParam,thetaParam,psiParam)*Power2(ComputeEG2c(hParam,thetaParam,psiParam));
}

scalar ComputeEG2G2G2G2cADAH(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG2c(hParam,thetaParam,psiParam)*ComputeEG2G2G2cAWAD(hParam,thetaParam,psiParam);
}

scalar ComputeEG2G2G2G2cADBH(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG2c(hParam,thetaParam,psiParam)*ComputeEG2G2G2cAWBD(hParam,thetaParam,psiParam);
}

scalar ComputeEG2G2G2G2cAD(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG2G2cAW(hParam,thetaParam,psiParam)*Power2(ComputeEG2c(hParam,thetaParam,psiParam));
}

scalar ComputeEG2G2G2G2cAH(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG2G2cAW(hParam,thetaParam,psiParam)*Power2(ComputeEG2c(hParam,thetaParam,psiParam));
}

scalar ComputeEG2G2G2G2c(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return Power4(ComputeEG2c(hParam,thetaParam,psiParam));
}

scalar ComputeEG4G2G2cADAH(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	
	return (-2*(-phi + 2*Psi1)*(2*Psi1 - Psi2)*r*Exponential(-(Psi1*r))*(-1 + Exponential(Psi1*r))*(-3*phi*Psi1 + phi*Psi2 - 2*Psi1*Psi2 + 4*Power2(Psi1))*(6*phi*Psi1*(2*Psi1 - Psi2)*(1 + (-1 + Psi1*r)*Exponential(Psi1*r)) + (2*Psi1 - Psi2)*(-1 + Exponential(Psi1*r))*Power2(Psi1) - 3*Power2(phi)*(2*Psi2*(-1 + Exponential(Psi1*r)) - Psi1*(-4 + Psi2*r + (4 + Psi2*r)*Exponential(Psi1*r)) + r*(1 + 3*Exponential(Psi1*r))*Power2(Psi1)))*Power2(1/(Psi1 - Psi2))*Power4(beta)*Power4(1/phi)*Power7(1/Psi1))/Psi2;
}

scalar ComputeEG4G2G2cADBH(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	
	return (4*(phi - 2*Psi1 + Psi2)*(-phi + Psi1 + Psi2)*(-3*Psi1 + 3*Psi2 - Psi3)*(-1 + Exponential(Psi2*r))*(-Exponential(Psi1*r) + Exponential(Psi2*r))*Exponential(-((Psi1 + Psi2)*r))*(-6*phi*Psi1*(2*Psi1 - Psi2)*(1 + (-1 + Psi1*r)*Exponential(Psi1*r)) - (2*Psi1 - Psi2)*(-1 + Exponential(Psi1*r))*Power2(Psi1) + 3*Power2(phi)*(2*Psi2*(-1 + Exponential(Psi1*r)) - Psi1*(-4 + Psi2*r + (4 + Psi2*r)*Exponential(Psi1*r)) + r*(1 + 3*Exponential(Psi1*r))*Power2(Psi1)))*Power2(1/Psi2)*Power3(1/(Psi1 - Psi2))*Power4(beta)*Power4(1/phi)*Power4(1/Psi1))/Psi3;
}

scalar ComputeEG4G2G2cAD(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG2c(hParam,thetaParam,psiParam)*ComputeEG4G2cAD(hParam,thetaParam,psiParam);
}

scalar ComputeEG4G2G2cAHBD(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	
	return (2*(-phi + 2*Psi1)*(2*Psi1 - Psi2)*(-phi + Psi2)*(-1 + Exponential(Psi1*r))*(Exponential(Psi1*r) - Exponential(Psi2*r))*Exponential(-((Psi1 + Psi2)*r))*(Psi2*(3*Psi2 - Psi3) - Psi1*(2*Psi2 + Psi3) + Power2(Psi1))*(-6*phi*(2*Psi1 - Psi2)*(-(Psi2*(-1 + Exponential(Psi1*r))) + Psi1*(-1 + Exponential(Psi2*r))) + 6*(-(Psi2*(-1 + Exponential(Psi1*r))) + Psi1*(-1 + Exponential(Psi2*r)))*Power2(phi) + Psi1*(-1 + Exponential(Psi2*r))*(-3*Psi1*Psi2 + 2*Power2(Psi1) + Power2(Psi2)))*Power3(1/Psi2)*Power4(beta)*Power4(1/phi)*Power4(1/Psi1)*Power4(1/(Psi1 - Psi2)))/(Psi1 - Psi3);
}

scalar ComputeEG4G2G2cAH(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG2G2cAW(hParam,thetaParam,psiParam)*ComputeEG4c(hParam,thetaParam,psiParam);
}

scalar ComputeEG4G2G2cBDBH(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	
	return (-2*(-phi + Psi1 + Psi2)*(-3*Psi1 + 3*Psi2 - Psi3)*r*Exponential(-(Psi2*r))*(-1 + Exponential(Psi2*r))*Power2(1/Psi1)*(Psi2*(Psi2 - Psi3) + Psi1*(4*Psi2 - Psi3) + phi*(3*Psi1 - 5*Psi2 + 2*Psi3) - 3*Power2(Psi1))*(-6*phi*(2*Psi1 - Psi2)*(-(Psi2*(-1 + Exponential(Psi1*r))) + Psi1*(-1 + Exponential(Psi2*r))) + 6*(-(Psi2*(-1 + Exponential(Psi1*r))) + Psi1*(-1 + Exponential(Psi2*r)))*Power2(phi) + Psi1*(-1 + Exponential(Psi2*r))*(-3*Psi1*Psi2 + 2*Power2(Psi1) + Power2(Psi2)))*Power3(1/Psi2)*Power3(1/(-Psi1 + Psi2))*Power4(beta)*Power4(1/phi))/((Psi2 - Psi3)*Psi3);
}

scalar ComputeEG4G2G2cBDCH(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	const scalar Psi4 = psiParam[3];
	
	return (-6*(-phi + Psi1 + Psi2 - Psi3)*(-phi + Psi1 + Psi3)*(Psi1*(-4*Psi2 + 4*Psi3 - 3*Psi4) + Psi3*(-4*Psi3 + Psi4) + Psi2*(4*Psi3 + Psi4))*(-1 + Exponential(Psi3*r))*(-Exponential(Psi2*r) + Exponential(Psi3*r))*Exponential(-((Psi2 + Psi3)*r))*Power2(1/Psi1)*Power2(1/Psi2)*(-6*phi*(2*Psi1 - Psi2)*(-(Psi2*(-1 + Exponential(Psi1*r))) + Psi1*(-1 + Exponential(Psi2*r))) + 6*(-(Psi2*(-1 + Exponential(Psi1*r))) + Psi1*(-1 + Exponential(Psi2*r)))*Power2(phi) + Psi1*(-1 + Exponential(Psi2*r))*(-3*Psi1*Psi2 + 2*Power2(Psi1) + Power2(Psi2)))*Power2(1/(Psi2 - Psi3))*Power2(1/Psi3)*Power4(beta)*Power4(1/phi))/((Psi1 - Psi2)*(Psi1 - Psi3)*Psi4);
}

scalar ComputeEG4G2G2cBD(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG2c(hParam,thetaParam,psiParam)*ComputeEG4G2cBD(hParam,thetaParam,psiParam);
}

scalar ComputeEG4G2G2c(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG4c(hParam,thetaParam,psiParam)*Power2(ComputeEG2c(hParam,thetaParam,psiParam));
}

scalar ComputeEG2G4G2cAWAH(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	
	return -(((-phi + 2*Psi1)*(2*Psi1 - Psi2)*Exponential(-2*Psi1*r)*Power2(1/Psi2)*Power2(-1 + Exponential(Psi1*r))*Power3(1/phi)*Power4(beta)*Power4(1/(Psi1 - Psi2))*(2*Psi2*Power3(Psi1)*(6*Psi1*(-2*Psi1 + Psi2)*Exponential(Psi2*r)*(Psi2*(3*Psi2 - Psi3) - Psi1*(2*Psi2 + Psi3) + Power2(Psi1)) + Exponential(Psi1*r)*(-(Psi1*Power2(Psi2)*(-6*Psi3 + 2*Psi2*(9 + Psi3*r) + r*(13 + 6*Psi3*r)*Power2(Psi2))) + 2*Psi2*Power2(Psi1)*(-3*Psi3 + 3*Psi2*(8 + Psi3*r) + r*(25 + 18*Psi3*r)*Power2(Psi2) + 3*Power2(r)*Power3(Psi2)) - 2*Power3(Psi1)*(6*Psi3 + Psi2*(15 + 11*Psi3*r) + 3*r*(10 + 13*Psi3*r)*Power2(Psi2) + 18*Power2(r)*Power3(Psi2)) + (12 + 17*(2*Psi2 + Psi3)*r + 6*Psi2*(13*Psi2 + 12*Psi3)*Power2(r))*Power4(Psi1) + Psi3*r*Power4(Psi2) - r*(11 + 24*(3*Psi2 + Psi3)*r)*Power5(Psi1) + 24*Power2(r)*Power6(Psi1))) - 12*phi*Psi1*(-(Power2(Psi1)*(14*Psi3*(-1 + Exponential(Psi1*r)) + 6*Psi2*(-1 + (1 + Psi3*r)*Exponential(Psi1*r)) + r*(1 + Psi3*r)*Exponential(Psi1*r)*Power2(Psi2))*Power3(Psi2)) + Power2(Psi2)*Power3(Psi1)*(16*Psi3*(-1 + Exponential(Psi1*r)) + 14*Psi2*(-1 + (1 + Psi3*r)*Exponential(Psi1*r)) + r*(6 + 7*Psi3*r)*Exponential(Psi1*r)*Power2(Psi2) + Exponential(Psi1*r)*Power2(r)*Power3(Psi2)) - Psi2*(Psi3*(-9 + 11*Exponential(Psi1*r) - 2*Exponential(Psi2*r)) + 2*Psi2*(-8 + (5 + 9*Psi3*r)*Exponential(Psi1*r) + 3*Exponential(Psi2*r)) + r*(8 + 17*Psi3*r)*Exponential(Psi1*r)*Power2(Psi2) + 7*Exponential(Psi1*r)*Power2(r)*Power3(Psi2))*Power4(Psi1) + Psi1*(6*Psi3*(-1 + Exponential(Psi1*r)) + Psi2*(-1 + (1 + Psi3*r)*Exponential(Psi1*r)))*Power4(Psi2) + (2*Psi3*(-1 + Exponential(Psi2*r)) + Psi2*(-9 + (5 + 9*Psi3*r)*Exponential(Psi1*r) + 4*Exponential(Psi2*r)) + r*(6 + 17*Psi3*r)*Exponential(Psi1*r)*Power2(Psi2) + 17*Exponential(Psi1*r)*Power2(r)*Power3(Psi2))*Power5(Psi1) - Psi3*(-1 + Exponential(Psi1*r))*Power5(Psi2) - (-2 + Psi2*r*(3 + 17*Psi2*r + 6*Psi3*r)*Exponential(Psi1*r) + 2*Exponential(Psi2*r))*Power6(Psi1) + 6*Psi2*Exponential(Psi1*r)*Power2(r)*Power7(Psi1)) + 3*Power2(phi)*(-(Power2(Psi1)*(28*Psi3*(-1 + Exponential(Psi1*r)) + 12*Psi2*(-1 + (1 + Psi3*r)*Exponential(Psi1*r)) + r*(2 + Psi3*r)*Exponential(Psi1*r)*Power2(Psi2))*Power3(Psi2)) + Power2(Psi2)*Power3(Psi1)*(32*Psi3*(-1 + Exponential(Psi1*r)) + 28*Psi2*(-1 + (1 + Psi3*r)*Exponential(Psi1*r)) + 4*r*(3 + 2*Psi3*r)*Exponential(Psi1*r)*Power2(Psi2) + Exponential(Psi1*r)*Power2(r)*Power3(Psi2)) - 2*Psi2*(Psi3*(-9 + 11*Exponential(Psi1*r) - 2*Exponential(Psi2*r)) + 2*Psi2*(-8 + (5 + 9*Psi3*r)*Exponential(Psi1*r) + 3*Exponential(Psi2*r)) + r*(8 + 11*Psi3*r)*Exponential(Psi1*r)*Power2(Psi2) + 4*Exponential(Psi1*r)*Power2(r)*Power3(Psi2))*Power4(Psi1) + 2*Psi1*(6*Psi3*(-1 + Exponential(Psi1*r)) + Psi2*(-1 + (1 + Psi3*r)*Exponential(Psi1*r)))*Power4(Psi2) + 2*(2*Psi3*(-1 + Exponential(Psi2*r)) + Psi2*(-9 + (5 + 9*Psi3*r)*Exponential(Psi1*r) + 4*Exponential(Psi2*r)) + 6*r*(1 + 2*Psi3*r)*Exponential(Psi1*r)*Power2(Psi2) + 11*Exponential(Psi1*r)*Power2(r)*Power3(Psi2))*Power5(Psi1) - 2*Psi3*(-1 + Exponential(Psi1*r))*Power5(Psi2) - (-4 + 3*Psi2*r*(2 + 8*Psi2*r + 3*Psi3*r)*Exponential(Psi1*r) + 4*Exponential(Psi2*r))*Power6(Psi1) + 9*Psi2*Exponential(Psi1*r)*Power2(r)*Power7(Psi1)))*Power8(1/Psi1))/(Psi1 - Psi3));
}

scalar ComputeEG2G4G2cAWBH(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	
	return (-2*(-phi + Psi1 + Psi2)*(3*Psi1 - 3*Psi2 + Psi3)*(-1 + Exponential(Psi1*r))*Exponential(-((3*Psi1 + Psi2)*r))*Power2(1/Psi2)*Power3(1/phi)*Power3(1/Psi1)*Power4(beta)*Power4(1/(Psi1 - Psi2))*(Exponential(3*Psi1*r)*(Psi1*Power2(Psi2)*(13*Psi3 + 4*Psi2*(-4 + 3*Psi3*r) - 12*r*Power2(Psi2)) + Psi2*Power2(Psi1)*(-53*Psi3 + Psi2*(80 - 60*Psi3*r) + 60*r*Power2(Psi2)) + 6*Power2(phi)*(Psi2*(Psi2 - Psi3)*(1 + Psi2*r) + 3*(1 + (Psi2 - Psi3)*r)*Power2(Psi1) + Psi1*(5*Psi3 + 4*Psi2*(-2 + Psi3*r) - 4*r*Power2(Psi2))) + (65*Psi3 + 32*Psi2*(-4 + 3*Psi3*r) - 96*r*Power2(Psi2))*Power3(Psi1) - 6*phi*(-((Psi2 - Psi3)*(1 + Psi2*r)*Power2(Psi2)) + Power2(Psi1)*(15*Psi3 + Psi2*(-27 + 17*Psi3*r) - 17*r*Power2(Psi2)) + Psi1*Psi2*(-8*Psi3 + Psi2*(11 - 8*Psi3*r) + 8*r*Power2(Psi2)) + (9 + 10*(Psi2 - Psi3)*r)*Power3(Psi1)) + (Psi2 - Psi3)*Power3(Psi2) + (39 + 48*(Psi2 - Psi3)*r)*Power4(Psi1)) - Exponential((3*Psi1 + Psi2)*r)*(Psi1*Power2(Psi2)*(13*Psi3 + 4*Psi2*(-4 + 3*Psi3*r) - 12*r*Power2(Psi2)) + Psi2*Power2(Psi1)*(-53*Psi3 + Psi2*(80 - 60*Psi3*r) + 60*r*Power2(Psi2)) + 6*Power2(phi)*(Psi2*(Psi2 - Psi3)*(1 + Psi2*r) + 3*(1 + (Psi2 - Psi3)*r)*Power2(Psi1) + Psi1*(5*Psi3 + 4*Psi2*(-2 + Psi3*r) - 4*r*Power2(Psi2))) + (65*Psi3 + 32*Psi2*(-4 + 3*Psi3*r) - 96*r*Power2(Psi2))*Power3(Psi1) - 6*phi*(-((Psi2 - Psi3)*(1 + Psi2*r)*Power2(Psi2)) + Power2(Psi1)*(15*Psi3 + Psi2*(-27 + 17*Psi3*r) - 17*r*Power2(Psi2)) + Psi1*Psi2*(-8*Psi3 + Psi2*(11 - 8*Psi3*r) + 8*r*Power2(Psi2)) + (9 + 10*(Psi2 - Psi3)*r)*Power3(Psi1)) + (Psi2 - Psi3)*Power3(Psi2) + (39 + 48*(Psi2 - Psi3)*r)*Power4(Psi1)) + Exponential((2*Psi1 + Psi2)*r)*(Psi1*Power2(Psi2)*(-13*Psi3 + Psi2*(16 - 6*Psi3*r) + 6*r*Power2(Psi2)) + Psi2*Power2(Psi1)*(53*Psi3 + 4*Psi2*(-20 + 3*Psi3*r) + 6*r*Power2(Psi2)) + (-65*Psi3 + 2*Psi2*(64 + 3*Psi3*r) - 78*r*Power2(Psi2))*Power3(Psi1) - 6*Power2(phi)*(Psi2*(Psi2 - Psi3) + (3 - 8*Psi2*r + 2*Psi3*r)*Power2(Psi1) + Psi1*(5*Psi3 - 2*Psi2*(4 + Psi3*r) + 5*r*Power2(Psi2)) + 3*r*Power3(Psi1)) + (-Psi2 + Psi3)*Power3(Psi2) + 3*(-13 + 34*Psi2*r - 4*Psi3*r)*Power4(Psi1) + 6*phi*((-Psi2 + Psi3)*Power2(Psi2) + Psi1*Psi2*(-8*Psi3 + Psi2*(11 + Psi3*r) - 4*r*Power2(Psi2)) + 3*Power2(Psi1)*(5*Psi3 - Psi2*(9 + 2*Psi3*r) + 7*r*Power2(Psi2)) + (9 - 26*Psi2*r + 5*Psi3*r)*Power3(Psi1) + 9*r*Power4(Psi1)) - 36*r*Power5(Psi1)) + Exponential(2*(Psi1 + Psi2)*r)*(Psi1*Power2(Psi2)*(13*Psi3 + 2*Psi2*(-8 + 3*Psi3*r) - 6*r*Power2(Psi2)) - Psi2*Power2(Psi1)*(53*Psi3 + 4*Psi2*(-20 + 3*Psi3*r) + 6*r*Power2(Psi2)) + (65*Psi3 - 2*Psi2*(64 + 3*Psi3*r) + 78*r*Power2(Psi2))*Power3(Psi1) + 6*Power2(phi)*(Psi2*(Psi2 - Psi3) + (3 - 8*Psi2*r + 2*Psi3*r)*Power2(Psi1) + Psi1*(5*Psi3 - 2*Psi2*(4 + Psi3*r) + 5*r*Power2(Psi2)) + 3*r*Power3(Psi1)) + (Psi2 - Psi3)*Power3(Psi2) + (39 - 102*Psi2*r + 12*Psi3*r)*Power4(Psi1) - 6*phi*((-Psi2 + Psi3)*Power2(Psi2) + Psi1*Psi2*(-8*Psi3 + Psi2*(11 + Psi3*r) - 4*r*Power2(Psi2)) + 3*Power2(Psi1)*(5*Psi3 - Psi2*(9 + 2*Psi3*r) + 7*r*Power2(Psi2)) + (9 - 26*Psi2*r + 5*Psi3*r)*Power3(Psi1) + 9*r*Power4(Psi1)) + 36*r*Power5(Psi1))))/((Psi2 - Psi3)*Psi3);
}

scalar ComputeEG2G4G2cAWCH(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	const scalar Psi4 = psiParam[3];
	
	return (6*(-phi + Psi1 + Psi3)*(Psi3*(4*Psi3 - Psi4) - Psi2*(4*Psi3 + Psi4) + Psi1*(4*Psi2 - 4*Psi3 + 3*Psi4))*(-1 + Exponential(Psi1*r))*(-1 + Exponential(Psi3*r))*Exponential(-((Psi1 + Psi3)*r))*Power2(1/Psi1)*Power2(1/(Psi1 - Psi3))*Power2(1/(Psi2 - Psi3))*Power2(1/Psi3)*(-6*(phi - 2*Psi1 + Psi2)*(Psi1 - Psi3)*(-phi + Psi1 + Psi2 - Psi3)*Exponential(Psi2*r) + (Psi2 - Psi3)*Exponential(Psi1*r)*(18*phi*Psi1 - 3*Psi1*Psi2 - 6*phi*Psi3 + 11*Psi1*Psi3 - 5*Psi2*Psi3 - 6*Power2(phi) - 13*Power2(Psi1) + 4*Power2(Psi2)) - (Psi1 - Psi2)*Exponential(Psi3*r)*(Psi1*(5*Psi2 - 11*Psi3) + 3*Psi2*Psi3 + 6*phi*(-3*Psi1 + Psi3) + 6*Power2(phi) + 12*Power2(Psi1) - 4*Power2(Psi2) + Power2(Psi3)))*Power3(1/phi)*Power4(beta))/((Psi1 - Psi2)*Psi2*Psi4);
}

scalar ComputeEG2G4G2cAW(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG2c(hParam,thetaParam,psiParam)*ComputeEG2G4cAW(hParam,thetaParam,psiParam);
}

scalar ComputeEG2G4G2cAH(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG2c(hParam,thetaParam,psiParam)*ComputeEG4G2cAD(hParam,thetaParam,psiParam);
}

scalar ComputeEG2G4G2cBH(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG2c(hParam,thetaParam,psiParam)*ComputeEG4G2cBD(hParam,thetaParam,psiParam);
}

scalar ComputeEG2G4G2c(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG4c(hParam,thetaParam,psiParam)*Power2(ComputeEG2c(hParam,thetaParam,psiParam));
}

scalar ComputeEG6G2cAH(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	
	return (3*(phi - 2*Psi1)*(2*Psi1 - Psi2)*Exponential(-2*Psi1*r)*Power3(1/Psi2)*Power4(beta)*Power4(1/phi)*Power4(1/(Psi1 - Psi2))*(Exponential(Psi1*r)*Power3(Psi1 - Psi2)*(4*Psi2*Power3(Psi1)*(Psi1*Psi2*(11*Psi2 + 13*Psi3) - (26*Psi2 + 5*Psi3)*Power2(Psi1) - 5*Psi3*Power2(Psi2) + 5*Power3(Psi1)) - 10*phi*Power2(Psi1)*(2*Psi1*(22*Psi3 + Psi2*(14 - Psi3*r))*Power2(Psi2) + Psi2*Power2(Psi1)*(-17*Psi3 + Psi2*(-71 + 5*Psi3*r) + 2*r*Power2(Psi2)) - (2*Psi3 + Psi2*(-11 + 2*Psi3*r) + 5*r*Power2(Psi2))*Power3(Psi1) - 16*Psi3*Power3(Psi2) + 2*(1 + Psi2*r)*Power4(Psi1)) + 60*Psi1*Power2(phi)*(2*Psi1*(9*Psi3 + Psi2*(3 - Psi3*r))*Power2(Psi2) + Psi2*Power2(Psi1)*(-8*Psi3 + Psi2*(-18 + 5*Psi3*r) + 2*r*Power2(Psi2)) - (2*Psi3 + 2*Psi2*(-1 + Psi3*r) + 5*r*Power2(Psi2))*Power3(Psi1) - 6*Psi3*Power3(Psi2) + 2*(1 + Psi2*r)*Power4(Psi1)) + 15*Power3(phi)*(6*Psi1*(-6*Psi3 + Psi2*(-2 + Psi3*r))*Power2(Psi2) + Psi2*Power2(Psi1)*(16*Psi3 + Psi2*(36 - 14*Psi3*r) + r*(-6 + Psi3*r)*Power2(Psi2)) + 12*Psi3*Power3(Psi2) - Power3(Psi1)*(-4*Psi3 + Psi2*(4 - 4*Psi3*r) + r*(-14 + Psi3*r)*Power2(Psi2) + Power2(r)*Power3(Psi2)) + (-4 - 4*Psi2*r + Power2(Psi2)*Power2(r))*Power4(Psi1))) + 10*(phi - Psi2)*Exponential((Psi1 + Psi2)*r)*(Psi2*(3*Psi2 - Psi3) - Psi1*(2*Psi2 + Psi3) + Power2(Psi1))*(-3*Psi1*Psi2 + 6*phi*(-2*Psi1 + Psi2) + 6*Power2(phi) + 2*Power2(Psi1) + Power2(Psi2))*Power5(Psi1) - 10*(phi - Psi2)*Exponential((2*Psi1 + Psi2)*r)*(Psi2*(3*Psi2 - Psi3) - Psi1*(2*Psi2 + Psi3) + Power2(Psi1))*(-3*Psi1*Psi2 + 6*phi*(-2*Psi1 + Psi2) + 6*Power2(phi) + 2*Power2(Psi1) + Power2(Psi2))*Power5(Psi1) + Exponential(3*Psi1*r)*Power2(Psi2)*(2*(Psi1 - Psi2)*Power3(Psi1)*(2*Psi1*(23*Psi3 + Psi2*(11 + 5*Psi3*r))*Power2(Psi2) - Psi2*Power2(Psi1)*(77*Psi3 + Psi2*(81 + 50*Psi3*r) + 10*r*Power2(Psi2)) + (51*Psi3 + 16*Psi2*(6 + 5*Psi3*r) + 50*r*Power2(Psi2))*Power3(Psi1) - 10*Psi3*Power3(Psi2) - (47 + 40*(2*Psi2 + Psi3)*r)*Power4(Psi1) + 40*r*Power5(Psi1)) - 60*Psi1*Power2(phi)*(-(Power2(Psi1)*Power2(Psi2)*(80*Psi3 + Psi2*(36 + 25*Psi3*r) + r*(4 + Psi3*r)*Power2(Psi2))) + 2*Psi1*(18*Psi3 + Psi2*(3 + 2*Psi3*r))*Power3(Psi2) + Psi2*Power3(Psi1)*(82*Psi3 + Psi2*(74 + 57*Psi3*r) + r*(25 + 7*Psi3*r)*Power2(Psi2) + Power2(r)*Power3(Psi2)) - (36*Psi3 + Psi2*(64 + 59*Psi3*r) + 17*r*(3 + Psi3*r)*Power2(Psi2) + 7*Power2(r)*Power3(Psi2))*Power4(Psi1) - 6*Psi3*Power4(Psi2) + (24 + (47*Psi2 + 23*Psi3)*r + 17*Psi2*(Psi2 + Psi3)*Power2(r))*Power5(Psi1) - r*(17 + 17*Psi2*r + 6*Psi3*r)*Power6(Psi1) + 6*Power2(r)*Power7(Psi1)) + 15*Power3(phi)*(-(Power2(Psi1)*Power2(Psi2)*(160*Psi3 + 8*Psi2*(9 + 5*Psi3*r) + r*(6 + Psi3*r)*Power2(Psi2))) + 6*Psi1*(12*Psi3 + Psi2*(2 + Psi3*r))*Power3(Psi2) + Psi2*Power3(Psi1)*(164*Psi3 + 4*Psi2*(37 + 24*Psi3*r) + 8*r*(5 + Psi3*r)*Power2(Psi2) + Power2(r)*Power3(Psi2)) - 2*(36*Psi3 + Psi2*(64 + 52*Psi3*r) + r*(42 + 11*Psi3*r)*Power2(Psi2) + 4*Power2(r)*Power3(Psi2))*Power4(Psi1) - 12*Psi3*Power4(Psi2) + (48 + (80*Psi2 + 42*Psi3)*r + 2*Psi2*(11*Psi2 + 12*Psi3)*Power2(r))*Power5(Psi1) - 3*r*(10 + 8*Psi2*r + 3*Psi3*r)*Power6(Psi1) + 9*Power2(r)*Power7(Psi1)) + 10*phi*Power2(Psi1)*(-(Power2(Psi1)*Power2(Psi2)*(197*Psi3 + Psi2*(155 + 81*Psi3*r) + 2*r*(13 + 3*Psi3*r)*Power2(Psi2))) + 2*Psi1*(46*Psi3 + 7*Psi2*(2 + Psi3*r))*Power3(Psi2) + Psi2*Power3(Psi1)*(202*Psi3 + Psi2*(293 + 179*Psi3*r) + 3*r*(43 + 12*Psi3*r)*Power2(Psi2) + 6*Power2(r)*Power3(Psi2)) - (93*Psi3 + Psi2*(235 + 183*Psi3*r) + r*(233 + 78*Psi3*r)*Power2(Psi2) + 36*Power2(r)*Power3(Psi2))*Power4(Psi1) - 16*Psi3*Power4(Psi2) + (81 + (195*Psi2 + 71*Psi3)*r + 6*Psi2*(13*Psi2 + 12*Psi3)*Power2(r))*Power5(Psi1) - r*(65 + 24*(3*Psi2 + Psi3)*r)*Power6(Psi1) + 24*Power2(r)*Power7(Psi1))) - 2*Exponential(2*Psi1*r)*((Psi1 - Psi2)*Psi2*Power3(Psi1)*(-(Power2(Psi1)*Power2(Psi2)*(149*Psi3 + Psi2*(177 + 50*Psi3*r) + 10*r*Power2(Psi2))) + Psi2*(97*Psi3 + 8*Psi2*(29 + 10*Psi3*r) + 50*r*Power2(Psi2))*Power3(Psi1) + 2*Psi1*(46*Psi3 + Psi2*(22 + 5*Psi3*r))*Power3(Psi2) - (10*Psi3 + Psi2*(119 + 40*Psi3*r) + 80*r*Power2(Psi2))*Power4(Psi1) - 20*Psi3*Power4(Psi2) + 10*(1 + 4*Psi2*r)*Power5(Psi1)) + 15*Power3(phi)*(Power3(Psi1)*Power3(Psi2)*(164*Psi3 + 4*Psi2*(37 + 4*Psi3*r) + 2*r*(2 + 3*Psi3*r)*Power2(Psi2) + Power2(r)*Power3(Psi2)) - 2*Power2(Psi2)*(36*Psi3 + Psi2*(64 + 11*Psi3*r) + r*(5 + 7*Psi3*r)*Power2(Psi2) + 3*Power2(r)*Power3(Psi2))*Power4(Psi1) - Power2(Psi1)*(160*Psi3 + 4*Psi2*(18 + Psi3*r) + Psi3*Power2(Psi2)*Power2(r))*Power4(Psi2) + 2*Psi2*(Psi3 + Psi2*(21 + 4*Psi3*r) + r*(5 + 7*Psi3*r)*Power2(Psi2) + 7*Power2(r)*Power3(Psi2))*Power5(Psi1) + 12*Psi1*(Psi2 + 6*Psi3)*Power5(Psi2) + (2*Psi3 + 2*Psi2*(2 + Psi3*r) - r*(2 + 5*Psi3*r)*Power2(Psi2) - 14*Power2(r)*Power3(Psi2))*Power6(Psi1) - 12*Psi3*Power6(Psi2) + (-2 - 2*Psi2*r + 5*Power2(Psi2)*Power2(r))*Power7(Psi1)) - 30*Psi1*Power2(phi)*(Power3(Psi1)*Power3(Psi2)*(164*Psi3 + 2*Psi2*(74 + 17*Psi3*r) + 7*r*(2 + Psi3*r)*Power2(Psi2) + Power2(r)*Power3(Psi2)) - Power2(Psi2)*(72*Psi3 + 4*Psi2*(32 + 9*Psi3*r) + r*(28 + 17*Psi3*r)*Power2(Psi2) + 7*Power2(r)*Power3(Psi2))*Power4(Psi1) - Power2(Psi1)*(160*Psi3 + 2*Psi2*(36 + 7*Psi3*r) + r*(2 + Psi3*r)*Power2(Psi2))*Power4(Psi2) + Psi2*(2*Psi3 + 6*Psi2*(7 + 2*Psi3*r) + r*(24 + 17*Psi3*r)*Power2(Psi2) + 17*Power2(r)*Power3(Psi2))*Power5(Psi1) + 2*Psi1*(36*Psi3 + Psi2*(6 + Psi3*r))*Power5(Psi2) + (2*Psi3 + 2*Psi2*(2 + Psi3*r) - 6*r*(1 + Psi3*r)*Power2(Psi2) - 17*Power2(r)*Power3(Psi2))*Power6(Psi1) - 12*Psi3*Power6(Psi2) + (-2 - 2*Psi2*r + 6*Power2(Psi2)*Power2(r))*Power7(Psi1)) + 5*phi*Power2(Psi1)*(Power3(Psi1)*Power3(Psi2)*(399*Psi3 + Psi2*(601 + 156*Psi3*r) + 2*r*(59 + 18*Psi3*r)*Power2(Psi2) + 6*Power2(r)*Power3(Psi2)) - Power2(Psi2)*(182*Psi3 + Psi2*(507 + 160*Psi3*r) + 6*r*(35 + 13*Psi3*r)*Power2(Psi2) + 36*Power2(r)*Power3(Psi2))*Power4(Psi1) - 2*Power2(Psi1)*(197*Psi3 + 5*Psi2*(31 + 7*Psi3*r) + 3*r*(4 + Psi3*r)*Power2(Psi2))*Power4(Psi2) + Psi2*(11*Psi3 + Psi2*(179 + 60*Psi3*r) + 4*r*(43 + 18*Psi3*r)*Power2(Psi2) + 78*Power2(r)*Power3(Psi2))*Power5(Psi1) + 4*Psi1*(46*Psi3 + Psi2*(14 + 3*Psi3*r))*Power5(Psi2) - (-2*Psi3 + Psi2*(5 - 2*Psi3*r) + 6*r*(9 + 4*Psi3*r)*Power2(Psi2) + 72*Power2(r)*Power3(Psi2))*Power6(Psi1) - 32*Psi3*Power6(Psi2) + (-2 - 2*Psi2*r + 24*Power2(Psi2)*Power2(r))*Power7(Psi1))))*Power8(1/Psi1))/(Psi1 - Psi3);
}

scalar ComputeEG6G2cBH(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	
	return (-6*(-phi + Psi1 + Psi2)*(-3*Psi1 + 3*Psi2 - Psi3)*Exponential(-(Psi2*r))*(-1 + Exponential(Psi2*r))*Power4(beta)*Power4(1/phi)*Power4(1/Psi1)*Power4(1/(Psi1 - Psi2))*Power4(1/Psi2)*(-30*Power3(phi)*(Power2(Psi1)*Power2(Psi2)*(-9*Psi3*(-1 + Exponential(Psi1*r)) - Psi2*(3*(4 + Psi3*r) + 4*(-3 + Psi3*r)*Exponential(Psi1*r)) + r*(3 + 4*Exponential(Psi1*r))*Power2(Psi2)) - Psi2*(-5*Psi3*(-1 + Exponential(Psi2*r)) + Psi2*(-14 - 3*Psi3*r + (3 - 3*Psi3*r)*Exponential(Psi1*r) + (11 + 2*Psi3*r)*Exponential(Psi2*r)) + r*(3 + 3*Exponential(Psi1*r) - 5*Exponential(Psi2*r))*Power2(Psi2))*Power3(Psi1) - Psi1*(Psi2 - Psi3)*(-7 + Psi2*r + (7 + Psi2*r)*Exponential(Psi1*r))*Power3(Psi2) + (-(Psi3*(-1 + Exponential(Psi2*r))) + Psi2*(-10 - Psi3*r + 2*(5 + Psi3*r)*Exponential(Psi2*r)) + (r - 8*r*Exponential(Psi2*r))*Power2(Psi2))*Power4(Psi1) + 2*(Psi2 - Psi3)*(-1 + Exponential(Psi1*r))*Power4(Psi2) + 3*(1 + (-1 + Psi2*r)*Exponential(Psi2*r))*Power5(Psi1)) + (Psi1 - Psi2)*Power2(Psi1)*(-(Power2(Psi1)*Power2(Psi2)*(8*Psi3*(-9 + 5*Exponential(Psi1*r) + 4*Exponential(Psi2*r)) + Psi2*(94 - 40*Exponential(Psi1*r) + 2*(-27 + 5*Psi3*r)*Exponential(Psi2*r)) + 5*r*Exponential(Psi2*r)*Power2(Psi2))) + Psi2*(46*Psi3*(-1 + Exponential(Psi2*r)) + Psi2*(114 - (114 + 5*Psi3*r)*Exponential(Psi2*r)) + 65*r*Exponential(Psi2*r)*Power2(Psi2))*Power3(Psi1) - Psi1*(-2*Psi3*(-23 + 20*Exponential(Psi1*r) + 3*Exponential(Psi2*r)) + Psi2*(-48 + 40*Exponential(Psi1*r) + (8 - 5*Psi3*r)*Exponential(Psi2*r)) + 5*r*Exponential(Psi2*r)*Power2(Psi2))*Power3(Psi2) + (-10*Psi3*(-1 + Exponential(Psi2*r)) + 2*Psi2*(-44 + (44 + 5*Psi3*r)*Exponential(Psi2*r)) - 85*r*Exponential(Psi2*r)*Power2(Psi2))*Power4(Psi1) + 10*(Psi2 - Psi3)*(-1 + Exponential(Psi1*r))*Power4(Psi2) + 30*(1 + (-1 + Psi2*r)*Exponential(Psi2*r))*Power5(Psi1)) + 30*Power2(phi)*(Power2(Psi2)*(-(Psi3*(-32 + 27*Exponential(Psi1*r) + 5*Exponential(Psi2*r))) + Psi2*(-50 - 9*Psi3*r + (39 - 17*Psi3*r)*Exponential(Psi1*r) + (11 + Psi3*r)*Exponential(Psi2*r)) + r*(9 + 17*Exponential(Psi1*r) - 4*Exponential(Psi2*r))*Power2(Psi2))*Power3(Psi1) - Power2(Psi1)*(-30*Psi3*(-1 + Exponential(Psi1*r)) - Psi2*(33 + 5*Psi3*r + (-33 + 8*Psi3*r)*Exponential(Psi1*r)) + r*(5 + 8*Exponential(Psi1*r))*Power2(Psi2))*Power3(Psi2) - Psi2*(-16*Psi3*(-1 + Exponential(Psi2*r)) + Psi2*(-52 - 7*Psi3*r + (9 - 10*Psi3*r)*Exponential(Psi1*r) + (43 + 6*Psi3*r)*Exponential(Psi2*r)) + r*(7 + 10*Exponential(Psi1*r) - 21*Exponential(Psi2*r))*Power2(Psi2))*Power4(Psi1) + Psi1*(Psi2 - Psi3)*(-13 + Psi2*r + (13 + Psi2*r)*Exponential(Psi1*r))*Power4(Psi2) + (-3*Psi3*(-1 + Exponential(Psi2*r)) + Psi2*(-33 - 2*Psi3*r + (33 + 5*Psi3*r)*Exponential(Psi2*r)) - 2*r*(-1 + 13*Exponential(Psi2*r))*Power2(Psi2))*Power5(Psi1) - 2*(Psi2 - Psi3)*(-1 + Exponential(Psi1*r))*Power5(Psi2) + 9*(1 + (-1 + Psi2*r)*Exponential(Psi2*r))*Power6(Psi1)) - 5*phi*Psi1*(Power2(Psi2)*(-(Psi3*(-163 + 117*Exponential(Psi1*r) + 46*Exponential(Psi2*r))) - Psi2*(286 + 9*Psi3*r + 12*(-15 + 8*Psi3*r)*Exponential(Psi1*r) + 2*(-53 + 2*Psi3*r)*Exponential(Psi2*r)) + r*(9 + 96*Exponential(Psi1*r) - 29*Exponential(Psi2*r))*Power2(Psi2))*Power3(Psi1) - Power2(Psi1)*(-(Psi3*(-161 + 155*Exponential(Psi1*r) + 6*Exponential(Psi2*r))) - Psi2*(194 + 5*Psi3*r + 2*(-91 + 30*Psi3*r)*Exponential(Psi1*r) + 4*(-3 + Psi3*r)*Exponential(Psi2*r)) + r*(5 + 60*Exponential(Psi1*r) + Exponential(Psi2*r))*Power2(Psi2))*Power3(Psi2) - Psi2*(-78*Psi3*(-1 + Exponential(Psi2*r)) + Psi2*(-279 - 7*Psi3*r + (39 - 48*Psi3*r)*Exponential(Psi1*r) + 16*(15 + Psi3*r)*Exponential(Psi2*r)) + r*(7 + 48*Exponential(Psi1*r) - 115*Exponential(Psi2*r))*Power2(Psi2))*Power4(Psi1) + Psi1*(-75*Psi3*(-1 + Exponential(Psi1*r)) - Psi2*(78 + Psi3*r + 6*(-13 + 2*Psi3*r)*Exponential(Psi1*r)) + r*(1 + 12*Exponential(Psi1*r))*Power2(Psi2))*Power4(Psi2) + (-14*Psi3*(-1 + Exponential(Psi2*r)) + 2*Psi2*(-82 - Psi3*r + (82 + 8*Psi3*r)*Exponential(Psi2*r)) - r*(-2 + 127*Exponential(Psi2*r))*Power2(Psi2))*Power5(Psi1) - 13*(Psi2 - Psi3)*(-1 + Exponential(Psi1*r))*Power5(Psi2) + 42*(1 + (-1 + Psi2*r)*Exponential(Psi2*r))*Power6(Psi1))))/((Psi2 - Psi3)*Psi3);
}

scalar ComputeEG6G2cCH(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	const scalar Psi4 = psiParam[3];
	
	return (-6*(-phi + Psi1 + Psi3)*(Psi3*(4*Psi3 - Psi4) - Psi2*(4*Psi3 + Psi4) + Psi1*(4*Psi2 - 4*Psi3 + 3*Psi4))*Exponential(-(Psi3*r))*(-1 + Exponential(Psi3*r))*Power2(1/Psi1)*Power2(1/Psi2)*Power2(1/(Psi1 - Psi3))*Power2(1/(Psi2 - Psi3))*Power3(1/Psi3)*Power4(beta)*Power4(1/phi)*(90*(3*Psi1 - Psi3)*Power2(phi)*(Psi2*Psi3*(-Psi2 + Psi3)*(-1 + Exponential(Psi1*r)) + (Psi3*(-1 + Exponential(Psi2*r)) - Psi2*(-1 + Exponential(Psi3*r)))*Power2(Psi1) + Psi1*((-1 + Exponential(Psi3*r))*Power2(Psi2) - (-1 + Exponential(Psi2*r))*Power2(Psi3))) - Psi1*(Psi1 - Psi2)*(Psi1 - Psi3)*(-(Psi2*(Psi2 - Psi3)*(12*Psi2*(-1 + Exponential(Psi3*r)) + Psi3*(14 - 15*Exponential(Psi2*r) + Exponential(Psi3*r)))) + 30*(-(Psi3*(-1 + Exponential(Psi2*r))) + Psi2*(-1 + Exponential(Psi3*r)))*Power2(Psi1) + 3*Psi1*(-(Psi2*Psi3*(-14 + 5*Exponential(Psi2*r) + 9*Exponential(Psi3*r))) + 4*(-1 + Exponential(Psi3*r))*Power2(Psi2) + 10*(-1 + Exponential(Psi2*r))*Power2(Psi3))) + 90*(Psi2*(Psi2 - Psi3)*Psi3*(-1 + Exponential(Psi1*r)) + (-(Psi3*(-1 + Exponential(Psi2*r))) + Psi2*(-1 + Exponential(Psi3*r)))*Power2(Psi1) + Psi1*(-((-1 + Exponential(Psi3*r))*Power2(Psi2)) + (-1 + Exponential(Psi2*r))*Power2(Psi3)))*Power3(phi) + 15*phi*(-(Psi3*(-1 + Exponential(Psi1*r))*Power2(Psi2)*(-9*Psi2*Psi3 + 4*Power2(Psi2) + 5*Power2(Psi3))) + (-(Psi2*Psi3*(-16 + 3*Exponential(Psi2*r) + 13*Exponential(Psi3*r))) - 10*(-1 + Exponential(Psi3*r))*Power2(Psi2) + 26*(-1 + Exponential(Psi2*r))*Power2(Psi3))*Power3(Psi1) + Power2(Psi1)*(Psi3*(-35 + 13*Exponential(Psi1*r) + 5*Exponential(Psi2*r) + 17*Exponential(Psi3*r))*Power2(Psi2) - Psi2*(-15 + 13*Exponential(Psi1*r) + 3*Exponential(Psi2*r) - Exponential(Psi3*r))*Power2(Psi3) - 8*(-1 + Exponential(Psi3*r))*Power3(Psi2) - 12*(-1 + Exponential(Psi2*r))*Power3(Psi3)) + Psi1*Psi2*(Psi3*(1 + 3*Exponential(Psi1*r) - 4*Exponential(Psi3*r))*Power2(Psi2) - Psi2*(-20 + 14*Exponential(Psi1*r) + 5*Exponential(Psi2*r) + Exponential(Psi3*r))*Power2(Psi3) + 4*(-1 + Exponential(Psi3*r))*Power3(Psi2) + (-17 + 11*Exponential(Psi1*r) + 6*Exponential(Psi2*r))*Power3(Psi3)) + 14*(-(Psi3*(-1 + Exponential(Psi2*r))) + Psi2*(-1 + Exponential(Psi3*r)))*Power4(Psi1))))/((Psi1 - Psi2)*Psi4);
}

scalar ComputeEG6G2c(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG2c(hParam,thetaParam,psiParam)*ComputeEG6c(hParam,thetaParam,psiParam);
}

scalar ComputeEG2G2G4cAWAD(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	
	return (-2*r*Exponential(-(Psi1*r))*(-1 + Exponential(Psi1*r))*(-3*phi*Psi1 + phi*Psi2 - 2*Psi1*Psi2 + 4*Power2(Psi1))*Power2(1/Psi2)*Power3(1/phi)*Power3(1/(Psi1 - Psi2))*Power4(beta)*(3*Power2(phi)*(2*Psi3*(-8 + 2*Psi2*r + (8 + 3*Psi2*r)*Exponential(Psi1*r))*Power2(Psi1)*Power2(Psi2) - Psi2*(10*Psi3*(-1 + Exponential(Psi1*r)) + Psi2*(6 + 5*Psi3*r + (-6 + 11*Psi3*r)*Exponential(Psi1*r)))*Power3(Psi1) - Psi1*Psi3*(-10 + Psi2*r + (10 + Psi2*r)*Exponential(Psi1*r))*Power3(Psi2) + 2*(Psi2*(6 + Psi3*r + 3*(-1 + Psi3*r)*Exponential(Psi1*r) - 3*Exponential(Psi2*r)) + Psi3*(-1 + Exponential(Psi2*r)))*Power4(Psi1) + 2*Psi3*(-1 + Exponential(Psi1*r))*Power4(Psi2) + 6*(-1 + Exponential(Psi2*r))*Power5(Psi1)) - 6*phi*Psi1*(Psi3*(4*(-6 + Psi2*r) + (24 + 11*Psi2*r)*Exponential(Psi1*r))*Power2(Psi1)*Power2(Psi2) - Psi2*(15*Psi3*(-1 + Exponential(Psi1*r)) + Psi2*(9 + 5*Psi3*r + (-9 + 19*Psi3*r)*Exponential(Psi1*r)))*Power3(Psi1) - Psi1*Psi3*(-15 + Psi2*r + (15 + 2*Psi2*r)*Exponential(Psi1*r))*Power3(Psi2) + (Psi2*(2*(9 + Psi3*r) + (-9 + 10*Psi3*r)*Exponential(Psi1*r) - 9*Exponential(Psi2*r)) + 3*Psi3*(-1 + Exponential(Psi2*r)))*Power4(Psi1) + 3*Psi3*(-1 + Exponential(Psi1*r))*Power4(Psi2) + 9*(-1 + Exponential(Psi2*r))*Power5(Psi1)) + Power2(Psi1)*((Psi3*(-95 + 101*Exponential(Psi1*r) - 6*Exponential(Psi2*r)) + 3*Psi2*(-13 + (7 + 20*Psi3*r)*Exponential(Psi1*r) + 6*Exponential(Psi2*r)))*Power2(Psi1)*Power2(Psi2) - Psi2*(Psi3*(-59 + 65*Exponential(Psi1*r) - 6*Exponential(Psi2*r)) + 6*Psi2*(-1 + (-5 + 16*Psi3*r)*Exponential(Psi1*r) + 6*Exponential(Psi2*r)))*Power3(Psi1) - Psi1*(61*Psi3*(-1 + Exponential(Psi1*r)) + 12*Psi2*(-1 + (1 + Psi3*r)*Exponential(Psi1*r)))*Power3(Psi2) + 3*(Psi2*(19 + (-13 + 16*Psi3*r)*Exponential(Psi1*r) - 6*Exponential(Psi2*r)) + 4*Psi3*(-1 + Exponential(Psi2*r)))*Power4(Psi1) + 13*Psi3*(-1 + Exponential(Psi1*r))*Power4(Psi2) + 36*(-1 + Exponential(Psi2*r))*Power5(Psi1)))*Power7(1/Psi1))/Psi3;
}

scalar ComputeEG2G2G4cAWBD(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	const scalar Psi4 = psiParam[3];
	
	return (4*(-phi + 2*Psi1 - Psi2)*(-1 + Exponential(Psi1*r))*(Exponential(Psi1*r) - Exponential(Psi2*r))*Exponential(-((Psi1 + Psi2)*r))*Power2(1/(Psi2 - Psi3))*Power2(1/Psi3)*Power3(1/phi)*Power3(1/Psi2)*Power4(beta)*Power4(1/Psi1)*Power4(1/(Psi1 - Psi2))*(3*Power2(phi)*(Psi1*Psi4*Power2(Psi2)*(-(Psi3*(-1 + Exponential(Psi1*r))) + Psi2*(-8 - Psi3*r + 8*Exponential(Psi1*r)) + 3*r*Power2(Psi2))*Power2(Psi2 - Psi3)*Power2(Psi3) - (3*Psi2 - Psi3)*Psi4*(-1 + Exponential(Psi1*r))*Power2(Psi2 - Psi3)*Power2(Psi3)*Power3(Psi2) - 3*(-(Psi3*(Psi4*(-19 + 20*Exponential(Psi2*r) - Exponential(Psi3*r)) + Psi3*(4 - 2*Psi4*r + (-8 + 11*Psi4*r)*Exponential(Psi2*r) + 4*Exponential(Psi3*r)))*Power2(Psi2)) + Psi2*(8*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(-4 - Psi4*r + 4*Exponential(Psi2*r)))*Power2(Psi3) + (Psi3*(20 - Psi4*r + (-12 + 11*Psi4*r)*Exponential(Psi2*r) - 8*Exponential(Psi3*r)) + 10*Psi4*(-1 + Exponential(Psi3*r)))*Power3(Psi2) + Psi4*(-1 + Exponential(Psi2*r))*Power3(Psi3) + 12*(-1 + Exponential(Psi3*r))*Power4(Psi2))*Power5(Psi1) + Power4(Psi1)*(-((Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(36 + 5*Psi4*r + 4*(-9 + 5*Psi4*r)*Exponential(Psi2*r)))*Power2(Psi2)*Power2(Psi3)) - Psi3*(3*Psi4*(-19 + 22*Exponential(Psi2*r) - 3*Exponential(Psi3*r)) + 2*Psi3*(-18 - 8*Psi4*r + 13*Psi4*r*Exponential(Psi2*r) + 18*Exponential(Psi3*r)))*Power3(Psi2) + Psi2*Psi4*(-29 - 2*Psi3*r + (29 + 7*Psi3*r)*Exponential(Psi2*r))*Power3(Psi3) + (Psi3*(36 - 9*Psi4*r + (-36 + 39*Psi4*r)*Exponential(Psi2*r)) + 36*Psi4*(-1 + Exponential(Psi3*r)))*Power4(Psi2) - 7*Psi4*(-1 + Exponential(Psi2*r))*Power4(Psi3) + 36*(-1 + Exponential(Psi3*r))*Power5(Psi2)) + Psi2*Power2(Psi1)*(-((Psi4*(-29 + 5*Exponential(Psi1*r) + 24*Exponential(Psi2*r)) + Psi3*(12 - 15*Psi4*r + 2*(-6 + 13*Psi4*r)*Exponential(Psi2*r)))*Power2(Psi3)*Power3(Psi2)) + Psi4*(-12*(3 + Psi3*r) + 8*Exponential(Psi1*r) + (28 + 13*Psi3*r)*Exponential(Psi2*r))*Power2(Psi2)*Power3(Psi3) + Psi3*(Psi3*(24 - 2*Psi4*r + 3*(-4 + 5*Psi4*r)*Exponential(Psi2*r) - 12*Exponential(Psi3*r)) + 3*Psi4*(-1 + Exponential(Psi3*r)))*Power4(Psi2) - Psi2*Psi4*(-2*(6 + Psi3*r) + Exponential(Psi1*r) + (11 + 2*Psi3*r)*Exponential(Psi2*r))*Power4(Psi3) + 3*(Psi4*(-1 + Exponential(Psi3*r)) + Psi3*(-4 - Psi4*r + 4*Exponential(Psi3*r)))*Power5(Psi2) - Psi4*(-1 + 2*Exponential(Psi1*r) - Exponential(Psi2*r))*Power5(Psi3)) + 3*(-(Psi2*Psi3*(6*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(4 + (-4 + 3*Psi4*r)*Exponential(Psi2*r)))) + (Psi3*(8 + (-4 + 3*Psi4*r)*Exponential(Psi2*r) - 4*Exponential(Psi3*r)) + 3*Psi4*(-1 + Exponential(Psi3*r)))*Power2(Psi2) + 3*Psi4*(-1 + Exponential(Psi2*r))*Power2(Psi3) + 4*(-1 + Exponential(Psi3*r))*Power3(Psi2))*Power6(Psi1) + Power3(Psi1)*(2*(Psi4*(-20 + Exponential(Psi1*r) + 19*Exponential(Psi2*r)) + Psi3*(18 - 3*Psi4*r + (-18 + 23*Psi4*r)*Exponential(Psi2*r)))*Power2(Psi3)*Power3(Psi2) - Psi4*(-52 - 9*Psi3*r + 4*Exponential(Psi1*r) + 4*(12 + 5*Psi3*r)*Exponential(Psi2*r))*Power2(Psi2)*Power3(Psi3) - Psi3*(Psi3*(60 + 11*Psi4*r + (-24 + 13*Psi4*r)*Exponential(Psi2*r) - 36*Exponential(Psi3*r)) - 3*Psi4*(-5 + 8*Exponential(Psi2*r) - 3*Exponential(Psi3*r)))*Power4(Psi2) + Psi2*Psi4*(-14 - Psi3*r + 2*Exponential(Psi1*r) + 2*(6 + Psi3*r)*Exponential(Psi2*r))*Power4(Psi3) - 3*(6*Psi4*(-1 + Exponential(Psi3*r)) + Psi3*(-4 - 3*Psi4*r + (-4 + 5*Psi4*r)*Exponential(Psi2*r) + 8*Exponential(Psi3*r)))*Power5(Psi2) + Psi4*(-1 + Exponential(Psi2*r))*Power5(Psi3) - 12*(-1 + Exponential(Psi3*r))*Power6(Psi2))) - 3*phi*(Psi1*Psi4*(Psi3*(-1 + Exponential(Psi1*r)) + Psi2*(-2 - Psi3*r + 2*Exponential(Psi1*r)) + 3*r*Power2(Psi2))*Power2(Psi2 - Psi3)*Power2(Psi3)*Power3(Psi2) - (3*Psi2 - Psi3)*Psi4*(-1 + Exponential(Psi1*r))*Power2(Psi2 - Psi3)*Power2(Psi3)*Power4(Psi2) + Power5(Psi1)*(-((26*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(60 + 2*Psi4*r + 15*(-4 + 3*Psi4*r)*Exponential(Psi2*r)))*Power2(Psi2)*Power2(Psi3)) - Psi3*(3*Psi4*(-19 + 24*Exponential(Psi2*r) - 5*Exponential(Psi3*r)) + Psi3*(-2*(42 + 5*Psi4*r) + 3*(8 + Psi4*r)*Exponential(Psi2*r) + 60*Exponential(Psi3*r)))*Power3(Psi2) + Psi2*Psi4*(-55 - 2*Psi3*r + (55 + 12*Psi3*r)*Exponential(Psi2*r))*Power3(Psi3) + 6*(7*Psi4*(-1 + Exponential(Psi3*r)) + Psi3*(2 - Psi4*r + 6*(-1 + Psi4*r)*Exponential(Psi2*r) + 4*Exponential(Psi3*r)))*Power4(Psi2) - 14*Psi4*(-1 + Exponential(Psi2*r))*Power4(Psi3) + 36*(-1 + Exponential(Psi3*r))*Power5(Psi2)) + Power2(Psi1)*Power2(Psi2)*(-((-(Psi4*(13 + 11*Exponential(Psi1*r) - 24*Exponential(Psi2*r))) + Psi3*(12 - 8*Psi4*r + 3*(-4 + 11*Psi4*r)*Exponential(Psi2*r)))*Power2(Psi3)*Power3(Psi2)) + Psi4*(-2 - 7*Psi3*r - 26*Exponential(Psi1*r) + 2*(14 + 9*Psi3*r)*Exponential(Psi2*r))*Power2(Psi2)*Power3(Psi3) + Psi3*(Psi3*(24 + Psi4*r + 6*(-2 + 3*Psi4*r)*Exponential(Psi2*r) - 12*Exponential(Psi3*r)) + 3*Psi4*(-1 + Exponential(Psi3*r)))*Power4(Psi2) - Psi2*Psi4*(8 - Psi3*r - 19*Exponential(Psi1*r) + (11 + 3*Psi3*r)*Exponential(Psi2*r))*Power4(Psi3) + 3*(Psi4*(-1 + Exponential(Psi3*r)) + Psi3*(-4 - Psi4*r + 4*Exponential(Psi3*r)))*Power5(Psi2) - Psi4*(-3 + 4*Exponential(Psi1*r) - Exponential(Psi2*r))*Power5(Psi3)) - 3*(-(Psi3*(2*Psi4*(-16 + 17*Exponential(Psi2*r) - Exponential(Psi3*r)) + Psi3*(4 - 2*Psi4*r + (-12 + 17*Psi4*r)*Exponential(Psi2*r) + 8*Exponential(Psi3*r)))*Power2(Psi2)) - Psi2*(-13*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(8 + Psi4*r + (-8 + Psi4*r)*Exponential(Psi2*r)))*Power2(Psi3) + (Psi3*(32 - Psi4*r + 2*(-10 + 9*Psi4*r)*Exponential(Psi2*r) - 12*Exponential(Psi3*r)) + 17*Psi4*(-1 + Exponential(Psi3*r)))*Power3(Psi2) + 2*Psi4*(-1 + Exponential(Psi2*r))*Power3(Psi3) + 20*(-1 + Exponential(Psi3*r))*Power4(Psi2))*Power6(Psi1) + Psi2*Power3(Psi1)*((-2*Psi4*(-9 + 4*Exponential(Psi1*r) + 5*Exponential(Psi2*r)) + 3*Psi3*(4 + 3*Psi4*r + (-4 + 3*Psi4*r)*Exponential(Psi2*r)))*Power2(Psi3)*Power3(Psi2) - Psi4*(20 + 3*Psi3*r - 12*Exponential(Psi1*r) + (-8 + 6*Psi3*r)*Exponential(Psi2*r))*Power2(Psi2)*Power3(Psi3) + Psi3*(3*Psi4*(-7 + 8*Exponential(Psi2*r) - Exponential(Psi3*r)) + Psi3*(-12 - 13*Psi4*r + 15*Psi4*r*Exponential(Psi2*r) + 12*Exponential(Psi3*r)))*Power4(Psi2) + Psi2*Psi4*(10 + Psi3*r - 10*Exponential(Psi2*r))*Power4(Psi3) - 6*(Psi3*(2 - Psi4*r + (-2 + 3*Psi4*r)*Exponential(Psi2*r)) + 2*Psi4*(-1 + Exponential(Psi3*r)))*Power5(Psi2) - Psi4*(-1 + 4*Exponential(Psi1*r) - 3*Exponential(Psi2*r))*Power5(Psi3) - 12*(-1 + Exponential(Psi3*r))*Power6(Psi2)) + Power4(Psi1)*((Psi4*(-79 + 4*Exponential(Psi1*r) + 75*Exponential(Psi2*r)) + Psi3*(36 - 11*Psi4*r + 6*(-6 + 11*Psi4*r)*Exponential(Psi2*r)))*Power2(Psi3)*Power3(Psi2) - Psi4*(-75 - 7*Psi3*r + 8*Exponential(Psi1*r) + (67 + 24*Psi3*r)*Exponential(Psi2*r))*Power2(Psi2)*Power3(Psi3) - Psi3*(Psi3*(84 - 5*Psi4*r + (-48 + 63*Psi4*r)*Exponential(Psi2*r) - 36*Exponential(Psi3*r)) + 9*Psi4*(-3 + 2*Exponential(Psi2*r) + Exponential(Psi3*r)))*Power4(Psi2) + Psi2*Psi4*(-21 - Psi3*r + 4*Exponential(Psi1*r) + (17 + 3*Psi3*r)*Exponential(Psi2*r))*Power4(Psi3) + 6*Psi3*(10 + (-2 + 3*Psi4*r)*Exponential(Psi2*r) - 8*Exponential(Psi3*r))*Power5(Psi2) + 2*Psi4*(-1 + Exponential(Psi2*r))*Power5(Psi3) + 12*(-1 + Exponential(Psi3*r))*Power6(Psi2)) + 6*(-(Psi2*Psi3*(6*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(4 + (-4 + 3*Psi4*r)*Exponential(Psi2*r)))) + (Psi3*(8 + (-4 + 3*Psi4*r)*Exponential(Psi2*r) - 4*Exponential(Psi3*r)) + 3*Psi4*(-1 + Exponential(Psi3*r)))*Power2(Psi2) + 3*Psi4*(-1 + Exponential(Psi2*r))*Power2(Psi3) + 4*(-1 + Exponential(Psi3*r))*Power3(Psi2))*Power7(Psi1)) + Psi1*(Psi1*(Psi2 - Psi3)*Psi3*Power3(Psi2)*(-3*Psi3*(-(Psi4*(-15 + 16*Exponential(Psi1*r) - 4*Exponential(Psi2*r) + 3*Exponential(Psi3*r))) + 2*Psi3*(-9 + (3 + 2*Psi4*r)*Exponential(Psi2*r) + 6*Exponential(Psi3*r)))*Power2(Psi2) + Psi2*(-(Psi4*(-61 + 54*Exponential(Psi1*r) + 7*Exponential(Psi2*r))) + 3*Psi3*(-8 + (8 + Psi4*r)*Exponential(Psi2*r)))*Power2(Psi3) + (9*Psi4*(-1 + Exponential(Psi3*r)) + Psi3*(-30 + (-6 + 9*Psi4*r)*Exponential(Psi2*r) + 36*Exponential(Psi3*r)))*Power3(Psi2) + Psi4*(-7 + 6*Exponential(Psi1*r) + Exponential(Psi2*r))*Power3(Psi3)) - 6*(3*Psi2 - Psi3)*Psi4*(-1 + Exponential(Psi1*r))*Power2(Psi2 - Psi3)*Power2(Psi3)*Power4(Psi2) - ((6*Psi3*(25 + (-19 + 11*Psi4*r)*Exponential(Psi2*r) - 6*Exponential(Psi3*r)) + Psi4*(-107 + 80*Exponential(Psi2*r) + 27*Exponential(Psi3*r)))*Power2(Psi2)*Power2(Psi3) - 3*Psi3*(Psi3*(76 + (-28 + 17*Psi4*r)*Exponential(Psi2*r) - 48*Exponential(Psi3*r)) - 3*Psi4*(2 + 3*Exponential(Psi2*r) - 5*Exponential(Psi3*r)))*Power3(Psi2) - Psi2*(88*Psi4*(-1 + Exponential(Psi2*r)) + 3*Psi3*(8 + (-8 + 5*Psi4*r)*Exponential(Psi2*r)))*Power3(Psi3) + 6*(Psi3*(17 + Exponential(Psi2*r) - 18*Exponential(Psi3*r)) - 3*Psi4*(-1 + Exponential(Psi3*r)))*Power4(Psi2) + 17*Psi4*(-1 + Exponential(Psi2*r))*Power4(Psi3))*Power5(Psi1) + Power2(Psi1)*Power2(Psi2)*(-((6*Psi3*(35 + (-5 + 9*Psi4*r)*Exponential(Psi2*r) - 30*Exponential(Psi3*r)) + Psi4*(-47 + 30*Exponential(Psi1*r) + 44*Exponential(Psi2*r) - 27*Exponential(Psi3*r)))*Power2(Psi3)*Power3(Psi2)) + (6*Psi3*(35 + (-17 + 3*Psi4*r)*Exponential(Psi2*r) - 18*Exponential(Psi3*r)) + Psi4*(-152 + 48*Exponential(Psi1*r) + 77*Exponential(Psi2*r) + 27*Exponential(Psi3*r)))*Power2(Psi2)*Power3(Psi3) + 3*Psi3*(Psi4*(11 + 4*Exponential(Psi2*r) - 15*Exponential(Psi3*r)) + 2*Psi3*(13 + (-1 + 8*Psi4*r)*Exponential(Psi2*r) - 12*Exponential(Psi3*r)))*Power4(Psi2) - Psi2*(2*Psi4*(-38 + 3*Exponential(Psi1*r) + 35*Exponential(Psi2*r)) + 3*Psi3*(24 + (-24 + Psi4*r)*Exponential(Psi2*r)))*Power4(Psi3) + (Psi3*(-6 + (6 - 9*Psi4*r)*Exponential(Psi2*r)) + 9*Psi4*(-1 + Exponential(Psi3*r)))*Power5(Psi2) - Psi4*(-5 + 12*Exponential(Psi1*r) - 7*Exponential(Psi2*r))*Power5(Psi3)) - 9*(-(Psi3*(-13*Psi4 + 2*(7*Psi4 + Psi3*(-2 + 3*Psi4*r))*Exponential(Psi2*r) + (4*Psi3 - Psi4)*Exponential(Psi3*r))*Power2(Psi2)) - Psi2*(-5*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(4 + (-4 + Psi4*r)*Exponential(Psi2*r)))*Power2(Psi3) + (Psi3*(12 + (-8 + 7*Psi4*r)*Exponential(Psi2*r) - 4*Exponential(Psi3*r)) + 7*Psi4*(-1 + Exponential(Psi3*r)))*Power3(Psi2) + Psi4*(-1 + Exponential(Psi2*r))*Power3(Psi3) + 8*(-1 + Exponential(Psi3*r))*Power4(Psi2))*Power6(Psi1) + Psi2*Power3(Psi1)*((Psi4*(58 + 12*Exponential(Psi1*r) + 11*Exponential(Psi2*r) - 81*Exponential(Psi3*r)) + 3*Psi3*(30 + (6 + 29*Psi4*r)*Exponential(Psi2*r) - 36*Exponential(Psi3*r)))*Power2(Psi3)*Power3(Psi2) - 3*(2*Psi3*(27 + (-9 + 5*Psi4*r)*Exponential(Psi2*r) - 18*Exponential(Psi3*r)) + Psi4*(-28 + 8*Exponential(Psi1*r) + 11*Exponential(Psi2*r) + 9*Exponential(Psi3*r)))*Power2(Psi2)*Power3(Psi3) - 3*Psi3*(Psi3*(-18 + (6 + 11*Psi4*r)*Exponential(Psi2*r) + 12*Exponential(Psi3*r)) - Psi4*(-43 + 13*Exponential(Psi2*r) + 30*Exponential(Psi3*r)))*Power4(Psi2) + 3*Psi2*(Psi4*(-17 + 4*Exponential(Psi1*r) + 13*Exponential(Psi2*r)) + Psi3*(24 + (-24 + Psi4*r)*Exponential(Psi2*r)))*Power4(Psi3) - 9*(Psi3*(10 + (-2 + 3*Psi4*r)*Exponential(Psi2*r) - 8*Exponential(Psi3*r)) + 5*Psi4*(-1 + Exponential(Psi3*r)))*Power5(Psi2) + 7*Psi4*(-1 + Exponential(Psi2*r))*Power5(Psi3) - 36*(-1 + Exponential(Psi3*r))*Power6(Psi2)) + Power4(Psi1)*((3*Psi3*(50 + (-38 + 15*Psi4*r)*Exponential(Psi2*r) - 12*Exponential(Psi3*r)) + Psi4*(-212 + 131*Exponential(Psi2*r) + 81*Exponential(Psi3*r)))*Power2(Psi3)*Power3(Psi2) - (Psi4*(-83 + 92*Exponential(Psi2*r) - 9*Exponential(Psi3*r)) + 6*Psi3*(-1 + (-5 + 3*Psi4*r)*Exponential(Psi2*r) + 6*Exponential(Psi3*r)))*Power2(Psi2)*Power3(Psi3) - 6*Psi3*(Psi3*(49 + (-19 + 17*Psi4*r)*Exponential(Psi2*r) - 30*Exponential(Psi3*r)) + Psi4*(-31 + 16*Exponential(Psi2*r) + 15*Exponential(Psi3*r)))*Power4(Psi2) + Psi2*(4*Psi4*(-1 + Exponential(Psi2*r)) + 3*Psi3*(-8 + (8 + Psi4*r)*Exponential(Psi2*r)))*Power4(Psi3) + 18*(Psi3*(13 + (-3 + 4*Psi4*r)*Exponential(Psi2*r) - 10*Exponential(Psi3*r)) + 3*Psi4*(-1 + Exponential(Psi3*r)))*Power5(Psi2) - Psi4*(-1 + Exponential(Psi2*r))*Power5(Psi3) + 72*(-1 + Exponential(Psi3*r))*Power6(Psi2)) + 9*(-(Psi2*Psi3*(6*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(4 + (-4 + 3*Psi4*r)*Exponential(Psi2*r)))) + (Psi3*(8 + (-4 + 3*Psi4*r)*Exponential(Psi2*r) - 4*Exponential(Psi3*r)) + 3*Psi4*(-1 + Exponential(Psi3*r)))*Power2(Psi2) + 3*Psi4*(-1 + Exponential(Psi2*r))*Power2(Psi3) + 4*(-1 + Exponential(Psi3*r))*Power3(Psi2))*Power7(Psi1))))/((Psi1 - Psi3)*Psi4);
}

scalar ComputeEG2G2G4cAW(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG2G2cAW(hParam,thetaParam,psiParam)*ComputeEG4c(hParam,thetaParam,psiParam);
}

scalar ComputeEG2G2G4cAD(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG2c(hParam,thetaParam,psiParam)*ComputeEG2G4cAW(hParam,thetaParam,psiParam);
}

scalar ComputeEG2G2G4c(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG4c(hParam,thetaParam,psiParam)*Power2(ComputeEG2c(hParam,thetaParam,psiParam));
}

scalar ComputeEG4G4cAD(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	
	return (4*Exponential(-(Psi1*r))*(-6*phi*Psi1*(2*Psi1 - Psi2)*(1 + (-1 + Psi1*r)*Exponential(Psi1*r)) - (2*Psi1 - Psi2)*(-1 + Exponential(Psi1*r))*Power2(Psi1) + 3*Power2(phi)*(2*Psi2*(-1 + Exponential(Psi1*r)) - Psi1*(-4 + Psi2*r + (4 + Psi2*r)*Exponential(Psi1*r)) + r*(1 + 3*Exponential(Psi1*r))*Power2(Psi1)))*Power2(1/Psi2)*Power3(1/(Psi1 - Psi2))*Power4(beta)*Power4(1/phi)*(3*Power2(phi)*(2*Psi3*(-8 + 2*Psi2*r + (8 + 3*Psi2*r)*Exponential(Psi1*r))*Power2(Psi1)*Power2(Psi2) - Psi2*(10*Psi3*(-1 + Exponential(Psi1*r)) + Psi2*(6 + 5*Psi3*r + (-6 + 11*Psi3*r)*Exponential(Psi1*r)))*Power3(Psi1) - Psi1*Psi3*(-10 + Psi2*r + (10 + Psi2*r)*Exponential(Psi1*r))*Power3(Psi2) + 2*(Psi2*(6 + Psi3*r + 3*(-1 + Psi3*r)*Exponential(Psi1*r) - 3*Exponential(Psi2*r)) + Psi3*(-1 + Exponential(Psi2*r)))*Power4(Psi1) + 2*Psi3*(-1 + Exponential(Psi1*r))*Power4(Psi2) + 6*(-1 + Exponential(Psi2*r))*Power5(Psi1)) - 6*phi*Psi1*(Psi3*(4*(-6 + Psi2*r) + (24 + 11*Psi2*r)*Exponential(Psi1*r))*Power2(Psi1)*Power2(Psi2) - Psi2*(15*Psi3*(-1 + Exponential(Psi1*r)) + Psi2*(9 + 5*Psi3*r + (-9 + 19*Psi3*r)*Exponential(Psi1*r)))*Power3(Psi1) - Psi1*Psi3*(-15 + Psi2*r + (15 + 2*Psi2*r)*Exponential(Psi1*r))*Power3(Psi2) + (Psi2*(2*(9 + Psi3*r) + (-9 + 10*Psi3*r)*Exponential(Psi1*r) - 9*Exponential(Psi2*r)) + 3*Psi3*(-1 + Exponential(Psi2*r)))*Power4(Psi1) + 3*Psi3*(-1 + Exponential(Psi1*r))*Power4(Psi2) + 9*(-1 + Exponential(Psi2*r))*Power5(Psi1)) + Power2(Psi1)*((Psi3*(-95 + 101*Exponential(Psi1*r) - 6*Exponential(Psi2*r)) + 3*Psi2*(-13 + (7 + 20*Psi3*r)*Exponential(Psi1*r) + 6*Exponential(Psi2*r)))*Power2(Psi1)*Power2(Psi2) - Psi2*(Psi3*(-59 + 65*Exponential(Psi1*r) - 6*Exponential(Psi2*r)) + 6*Psi2*(-1 + (-5 + 16*Psi3*r)*Exponential(Psi1*r) + 6*Exponential(Psi2*r)))*Power3(Psi1) - Psi1*(61*Psi3*(-1 + Exponential(Psi1*r)) + 12*Psi2*(-1 + (1 + Psi3*r)*Exponential(Psi1*r)))*Power3(Psi2) + 3*(Psi2*(19 + (-13 + 16*Psi3*r)*Exponential(Psi1*r) - 6*Exponential(Psi2*r)) + 4*Psi3*(-1 + Exponential(Psi2*r)))*Power4(Psi1) + 13*Psi3*(-1 + Exponential(Psi1*r))*Power4(Psi2) + 36*(-1 + Exponential(Psi2*r))*Power5(Psi1)))*Power8(1/Psi1))/Psi3;
}

scalar ComputeEG4G4cBD(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	const scalar Psi4 = psiParam[3];
	
	return (-4*Exponential(-(Psi2*r))*(-6*phi*(2*Psi1 - Psi2)*(-(Psi2*(-1 + Exponential(Psi1*r))) + Psi1*(-1 + Exponential(Psi2*r))) + 6*(-(Psi2*(-1 + Exponential(Psi1*r))) + Psi1*(-1 + Exponential(Psi2*r)))*Power2(phi) + Psi1*(-1 + Exponential(Psi2*r))*(-3*Psi1*Psi2 + 2*Power2(Psi1) + Power2(Psi2)))*Power2(1/(Psi2 - Psi3))*Power2(1/Psi3)*Power4(beta)*Power4(1/phi)*Power4(1/Psi1)*Power4(1/(Psi1 - Psi2))*Power4(1/Psi2)*(-3*Power2(phi)*(Psi1*Psi4*Power2(Psi2)*(-(Psi3*(-1 + Exponential(Psi1*r))) + Psi2*(-8 - Psi3*r + 8*Exponential(Psi1*r)) + 3*r*Power2(Psi2))*Power2(Psi2 - Psi3)*Power2(Psi3) - (3*Psi2 - Psi3)*Psi4*(-1 + Exponential(Psi1*r))*Power2(Psi2 - Psi3)*Power2(Psi3)*Power3(Psi2) - 3*(-(Psi3*(Psi4*(-19 + 20*Exponential(Psi2*r) - Exponential(Psi3*r)) + Psi3*(4 - 2*Psi4*r + (-8 + 11*Psi4*r)*Exponential(Psi2*r) + 4*Exponential(Psi3*r)))*Power2(Psi2)) + Psi2*(8*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(-4 - Psi4*r + 4*Exponential(Psi2*r)))*Power2(Psi3) + (Psi3*(20 - Psi4*r + (-12 + 11*Psi4*r)*Exponential(Psi2*r) - 8*Exponential(Psi3*r)) + 10*Psi4*(-1 + Exponential(Psi3*r)))*Power3(Psi2) + Psi4*(-1 + Exponential(Psi2*r))*Power3(Psi3) + 12*(-1 + Exponential(Psi3*r))*Power4(Psi2))*Power5(Psi1) + Power4(Psi1)*(-((Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(36 + 5*Psi4*r + 4*(-9 + 5*Psi4*r)*Exponential(Psi2*r)))*Power2(Psi2)*Power2(Psi3)) - Psi3*(3*Psi4*(-19 + 22*Exponential(Psi2*r) - 3*Exponential(Psi3*r)) + 2*Psi3*(-18 - 8*Psi4*r + 13*Psi4*r*Exponential(Psi2*r) + 18*Exponential(Psi3*r)))*Power3(Psi2) + Psi2*Psi4*(-29 - 2*Psi3*r + (29 + 7*Psi3*r)*Exponential(Psi2*r))*Power3(Psi3) + (Psi3*(36 - 9*Psi4*r + (-36 + 39*Psi4*r)*Exponential(Psi2*r)) + 36*Psi4*(-1 + Exponential(Psi3*r)))*Power4(Psi2) - 7*Psi4*(-1 + Exponential(Psi2*r))*Power4(Psi3) + 36*(-1 + Exponential(Psi3*r))*Power5(Psi2)) + Psi2*Power2(Psi1)*(-((Psi4*(-29 + 5*Exponential(Psi1*r) + 24*Exponential(Psi2*r)) + Psi3*(12 - 15*Psi4*r + 2*(-6 + 13*Psi4*r)*Exponential(Psi2*r)))*Power2(Psi3)*Power3(Psi2)) + Psi4*(-12*(3 + Psi3*r) + 8*Exponential(Psi1*r) + (28 + 13*Psi3*r)*Exponential(Psi2*r))*Power2(Psi2)*Power3(Psi3) + Psi3*(Psi3*(24 - 2*Psi4*r + 3*(-4 + 5*Psi4*r)*Exponential(Psi2*r) - 12*Exponential(Psi3*r)) + 3*Psi4*(-1 + Exponential(Psi3*r)))*Power4(Psi2) - Psi2*Psi4*(-2*(6 + Psi3*r) + Exponential(Psi1*r) + (11 + 2*Psi3*r)*Exponential(Psi2*r))*Power4(Psi3) + 3*(Psi4*(-1 + Exponential(Psi3*r)) + Psi3*(-4 - Psi4*r + 4*Exponential(Psi3*r)))*Power5(Psi2) - Psi4*(-1 + 2*Exponential(Psi1*r) - Exponential(Psi2*r))*Power5(Psi3)) + 3*(-(Psi2*Psi3*(6*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(4 + (-4 + 3*Psi4*r)*Exponential(Psi2*r)))) + (Psi3*(8 + (-4 + 3*Psi4*r)*Exponential(Psi2*r) - 4*Exponential(Psi3*r)) + 3*Psi4*(-1 + Exponential(Psi3*r)))*Power2(Psi2) + 3*Psi4*(-1 + Exponential(Psi2*r))*Power2(Psi3) + 4*(-1 + Exponential(Psi3*r))*Power3(Psi2))*Power6(Psi1) + Power3(Psi1)*(2*(Psi4*(-20 + Exponential(Psi1*r) + 19*Exponential(Psi2*r)) + Psi3*(18 - 3*Psi4*r + (-18 + 23*Psi4*r)*Exponential(Psi2*r)))*Power2(Psi3)*Power3(Psi2) - Psi4*(-52 - 9*Psi3*r + 4*Exponential(Psi1*r) + 4*(12 + 5*Psi3*r)*Exponential(Psi2*r))*Power2(Psi2)*Power3(Psi3) - Psi3*(Psi3*(60 + 11*Psi4*r + (-24 + 13*Psi4*r)*Exponential(Psi2*r) - 36*Exponential(Psi3*r)) - 3*Psi4*(-5 + 8*Exponential(Psi2*r) - 3*Exponential(Psi3*r)))*Power4(Psi2) + Psi2*Psi4*(-14 - Psi3*r + 2*Exponential(Psi1*r) + 2*(6 + Psi3*r)*Exponential(Psi2*r))*Power4(Psi3) - 3*(6*Psi4*(-1 + Exponential(Psi3*r)) + Psi3*(-4 - 3*Psi4*r + (-4 + 5*Psi4*r)*Exponential(Psi2*r) + 8*Exponential(Psi3*r)))*Power5(Psi2) + Psi4*(-1 + Exponential(Psi2*r))*Power5(Psi3) - 12*(-1 + Exponential(Psi3*r))*Power6(Psi2))) + Psi1*(-(Psi1*(Psi2 - Psi3)*Psi3*Power3(Psi2)*(-3*Psi3*(-(Psi4*(-15 + 16*Exponential(Psi1*r) - 4*Exponential(Psi2*r) + 3*Exponential(Psi3*r))) + 2*Psi3*(-9 + (3 + 2*Psi4*r)*Exponential(Psi2*r) + 6*Exponential(Psi3*r)))*Power2(Psi2) + Psi2*(-(Psi4*(-61 + 54*Exponential(Psi1*r) + 7*Exponential(Psi2*r))) + 3*Psi3*(-8 + (8 + Psi4*r)*Exponential(Psi2*r)))*Power2(Psi3) + (9*Psi4*(-1 + Exponential(Psi3*r)) + Psi3*(-30 + (-6 + 9*Psi4*r)*Exponential(Psi2*r) + 36*Exponential(Psi3*r)))*Power3(Psi2) + Psi4*(-7 + 6*Exponential(Psi1*r) + Exponential(Psi2*r))*Power3(Psi3))) + 6*(3*Psi2 - Psi3)*Psi4*(-1 + Exponential(Psi1*r))*Power2(Psi2 - Psi3)*Power2(Psi3)*Power4(Psi2) + ((6*Psi3*(25 + (-19 + 11*Psi4*r)*Exponential(Psi2*r) - 6*Exponential(Psi3*r)) + Psi4*(-107 + 80*Exponential(Psi2*r) + 27*Exponential(Psi3*r)))*Power2(Psi2)*Power2(Psi3) - 3*Psi3*(Psi3*(76 + (-28 + 17*Psi4*r)*Exponential(Psi2*r) - 48*Exponential(Psi3*r)) - 3*Psi4*(2 + 3*Exponential(Psi2*r) - 5*Exponential(Psi3*r)))*Power3(Psi2) - Psi2*(88*Psi4*(-1 + Exponential(Psi2*r)) + 3*Psi3*(8 + (-8 + 5*Psi4*r)*Exponential(Psi2*r)))*Power3(Psi3) + 6*(Psi3*(17 + Exponential(Psi2*r) - 18*Exponential(Psi3*r)) - 3*Psi4*(-1 + Exponential(Psi3*r)))*Power4(Psi2) + 17*Psi4*(-1 + Exponential(Psi2*r))*Power4(Psi3))*Power5(Psi1) + Power2(Psi1)*Power2(Psi2)*((6*Psi3*(35 + (-5 + 9*Psi4*r)*Exponential(Psi2*r) - 30*Exponential(Psi3*r)) + Psi4*(-47 + 30*Exponential(Psi1*r) + 44*Exponential(Psi2*r) - 27*Exponential(Psi3*r)))*Power2(Psi3)*Power3(Psi2) - (6*Psi3*(35 + (-17 + 3*Psi4*r)*Exponential(Psi2*r) - 18*Exponential(Psi3*r)) + Psi4*(-152 + 48*Exponential(Psi1*r) + 77*Exponential(Psi2*r) + 27*Exponential(Psi3*r)))*Power2(Psi2)*Power3(Psi3) - 3*Psi3*(Psi4*(11 + 4*Exponential(Psi2*r) - 15*Exponential(Psi3*r)) + 2*Psi3*(13 + (-1 + 8*Psi4*r)*Exponential(Psi2*r) - 12*Exponential(Psi3*r)))*Power4(Psi2) + Psi2*(2*Psi4*(-38 + 3*Exponential(Psi1*r) + 35*Exponential(Psi2*r)) + 3*Psi3*(24 + (-24 + Psi4*r)*Exponential(Psi2*r)))*Power4(Psi3) + (Psi3*(6 + (-6 + 9*Psi4*r)*Exponential(Psi2*r)) - 9*Psi4*(-1 + Exponential(Psi3*r)))*Power5(Psi2) + Psi4*(-5 + 12*Exponential(Psi1*r) - 7*Exponential(Psi2*r))*Power5(Psi3)) + 9*(-(Psi3*(-13*Psi4 + 2*(7*Psi4 + Psi3*(-2 + 3*Psi4*r))*Exponential(Psi2*r) + (4*Psi3 - Psi4)*Exponential(Psi3*r))*Power2(Psi2)) - Psi2*(-5*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(4 + (-4 + Psi4*r)*Exponential(Psi2*r)))*Power2(Psi3) + (Psi3*(12 + (-8 + 7*Psi4*r)*Exponential(Psi2*r) - 4*Exponential(Psi3*r)) + 7*Psi4*(-1 + Exponential(Psi3*r)))*Power3(Psi2) + Psi4*(-1 + Exponential(Psi2*r))*Power3(Psi3) + 8*(-1 + Exponential(Psi3*r))*Power4(Psi2))*Power6(Psi1) + Psi2*Power3(Psi1)*(-((Psi4*(58 + 12*Exponential(Psi1*r) + 11*Exponential(Psi2*r) - 81*Exponential(Psi3*r)) + 3*Psi3*(30 + (6 + 29*Psi4*r)*Exponential(Psi2*r) - 36*Exponential(Psi3*r)))*Power2(Psi3)*Power3(Psi2)) + 3*(2*Psi3*(27 + (-9 + 5*Psi4*r)*Exponential(Psi2*r) - 18*Exponential(Psi3*r)) + Psi4*(-28 + 8*Exponential(Psi1*r) + 11*Exponential(Psi2*r) + 9*Exponential(Psi3*r)))*Power2(Psi2)*Power3(Psi3) + 3*Psi3*(Psi3*(-18 + (6 + 11*Psi4*r)*Exponential(Psi2*r) + 12*Exponential(Psi3*r)) - Psi4*(-43 + 13*Exponential(Psi2*r) + 30*Exponential(Psi3*r)))*Power4(Psi2) - 3*Psi2*(Psi4*(-17 + 4*Exponential(Psi1*r) + 13*Exponential(Psi2*r)) + Psi3*(24 + (-24 + Psi4*r)*Exponential(Psi2*r)))*Power4(Psi3) + 9*(Psi3*(10 + (-2 + 3*Psi4*r)*Exponential(Psi2*r) - 8*Exponential(Psi3*r)) + 5*Psi4*(-1 + Exponential(Psi3*r)))*Power5(Psi2) - 7*Psi4*(-1 + Exponential(Psi2*r))*Power5(Psi3) + 36*(-1 + Exponential(Psi3*r))*Power6(Psi2)) - Power4(Psi1)*((3*Psi3*(50 + (-38 + 15*Psi4*r)*Exponential(Psi2*r) - 12*Exponential(Psi3*r)) + Psi4*(-212 + 131*Exponential(Psi2*r) + 81*Exponential(Psi3*r)))*Power2(Psi3)*Power3(Psi2) - (Psi4*(-83 + 92*Exponential(Psi2*r) - 9*Exponential(Psi3*r)) + 6*Psi3*(-1 + (-5 + 3*Psi4*r)*Exponential(Psi2*r) + 6*Exponential(Psi3*r)))*Power2(Psi2)*Power3(Psi3) - 6*Psi3*(Psi3*(49 + (-19 + 17*Psi4*r)*Exponential(Psi2*r) - 30*Exponential(Psi3*r)) + Psi4*(-31 + 16*Exponential(Psi2*r) + 15*Exponential(Psi3*r)))*Power4(Psi2) + Psi2*(4*Psi4*(-1 + Exponential(Psi2*r)) + 3*Psi3*(-8 + (8 + Psi4*r)*Exponential(Psi2*r)))*Power4(Psi3) + 18*(Psi3*(13 + (-3 + 4*Psi4*r)*Exponential(Psi2*r) - 10*Exponential(Psi3*r)) + 3*Psi4*(-1 + Exponential(Psi3*r)))*Power5(Psi2) - Psi4*(-1 + Exponential(Psi2*r))*Power5(Psi3) + 72*(-1 + Exponential(Psi3*r))*Power6(Psi2)) - 9*(-(Psi2*Psi3*(6*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(4 + (-4 + 3*Psi4*r)*Exponential(Psi2*r)))) + (Psi3*(8 + (-4 + 3*Psi4*r)*Exponential(Psi2*r) - 4*Exponential(Psi3*r)) + 3*Psi4*(-1 + Exponential(Psi3*r)))*Power2(Psi2) + 3*Psi4*(-1 + Exponential(Psi2*r))*Power2(Psi3) + 4*(-1 + Exponential(Psi3*r))*Power3(Psi2))*Power7(Psi1)) + 3*phi*(Psi1*Psi4*(Psi3*(-1 + Exponential(Psi1*r)) + Psi2*(-2 - Psi3*r + 2*Exponential(Psi1*r)) + 3*r*Power2(Psi2))*Power2(Psi2 - Psi3)*Power2(Psi3)*Power3(Psi2) - (3*Psi2 - Psi3)*Psi4*(-1 + Exponential(Psi1*r))*Power2(Psi2 - Psi3)*Power2(Psi3)*Power4(Psi2) + Power5(Psi1)*(-((26*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(60 + 2*Psi4*r + 15*(-4 + 3*Psi4*r)*Exponential(Psi2*r)))*Power2(Psi2)*Power2(Psi3)) - Psi3*(3*Psi4*(-19 + 24*Exponential(Psi2*r) - 5*Exponential(Psi3*r)) + Psi3*(-2*(42 + 5*Psi4*r) + 3*(8 + Psi4*r)*Exponential(Psi2*r) + 60*Exponential(Psi3*r)))*Power3(Psi2) + Psi2*Psi4*(-55 - 2*Psi3*r + (55 + 12*Psi3*r)*Exponential(Psi2*r))*Power3(Psi3) + 6*(7*Psi4*(-1 + Exponential(Psi3*r)) + Psi3*(2 - Psi4*r + 6*(-1 + Psi4*r)*Exponential(Psi2*r) + 4*Exponential(Psi3*r)))*Power4(Psi2) - 14*Psi4*(-1 + Exponential(Psi2*r))*Power4(Psi3) + 36*(-1 + Exponential(Psi3*r))*Power5(Psi2)) + Power2(Psi1)*Power2(Psi2)*(-((-(Psi4*(13 + 11*Exponential(Psi1*r) - 24*Exponential(Psi2*r))) + Psi3*(12 - 8*Psi4*r + 3*(-4 + 11*Psi4*r)*Exponential(Psi2*r)))*Power2(Psi3)*Power3(Psi2)) + Psi4*(-2 - 7*Psi3*r - 26*Exponential(Psi1*r) + 2*(14 + 9*Psi3*r)*Exponential(Psi2*r))*Power2(Psi2)*Power3(Psi3) + Psi3*(Psi3*(24 + Psi4*r + 6*(-2 + 3*Psi4*r)*Exponential(Psi2*r) - 12*Exponential(Psi3*r)) + 3*Psi4*(-1 + Exponential(Psi3*r)))*Power4(Psi2) - Psi2*Psi4*(8 - Psi3*r - 19*Exponential(Psi1*r) + (11 + 3*Psi3*r)*Exponential(Psi2*r))*Power4(Psi3) + 3*(Psi4*(-1 + Exponential(Psi3*r)) + Psi3*(-4 - Psi4*r + 4*Exponential(Psi3*r)))*Power5(Psi2) - Psi4*(-3 + 4*Exponential(Psi1*r) - Exponential(Psi2*r))*Power5(Psi3)) - 3*(-(Psi3*(2*Psi4*(-16 + 17*Exponential(Psi2*r) - Exponential(Psi3*r)) + Psi3*(4 - 2*Psi4*r + (-12 + 17*Psi4*r)*Exponential(Psi2*r) + 8*Exponential(Psi3*r)))*Power2(Psi2)) - Psi2*(-13*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(8 + Psi4*r + (-8 + Psi4*r)*Exponential(Psi2*r)))*Power2(Psi3) + (Psi3*(32 - Psi4*r + 2*(-10 + 9*Psi4*r)*Exponential(Psi2*r) - 12*Exponential(Psi3*r)) + 17*Psi4*(-1 + Exponential(Psi3*r)))*Power3(Psi2) + 2*Psi4*(-1 + Exponential(Psi2*r))*Power3(Psi3) + 20*(-1 + Exponential(Psi3*r))*Power4(Psi2))*Power6(Psi1) + Psi2*Power3(Psi1)*((-2*Psi4*(-9 + 4*Exponential(Psi1*r) + 5*Exponential(Psi2*r)) + 3*Psi3*(4 + 3*Psi4*r + (-4 + 3*Psi4*r)*Exponential(Psi2*r)))*Power2(Psi3)*Power3(Psi2) - Psi4*(20 + 3*Psi3*r - 12*Exponential(Psi1*r) + (-8 + 6*Psi3*r)*Exponential(Psi2*r))*Power2(Psi2)*Power3(Psi3) + Psi3*(3*Psi4*(-7 + 8*Exponential(Psi2*r) - Exponential(Psi3*r)) + Psi3*(-12 - 13*Psi4*r + 15*Psi4*r*Exponential(Psi2*r) + 12*Exponential(Psi3*r)))*Power4(Psi2) + Psi2*Psi4*(10 + Psi3*r - 10*Exponential(Psi2*r))*Power4(Psi3) - 6*(Psi3*(2 - Psi4*r + (-2 + 3*Psi4*r)*Exponential(Psi2*r)) + 2*Psi4*(-1 + Exponential(Psi3*r)))*Power5(Psi2) - Psi4*(-1 + 4*Exponential(Psi1*r) - 3*Exponential(Psi2*r))*Power5(Psi3) - 12*(-1 + Exponential(Psi3*r))*Power6(Psi2)) + Power4(Psi1)*((Psi4*(-79 + 4*Exponential(Psi1*r) + 75*Exponential(Psi2*r)) + Psi3*(36 - 11*Psi4*r + 6*(-6 + 11*Psi4*r)*Exponential(Psi2*r)))*Power2(Psi3)*Power3(Psi2) - Psi4*(-75 - 7*Psi3*r + 8*Exponential(Psi1*r) + (67 + 24*Psi3*r)*Exponential(Psi2*r))*Power2(Psi2)*Power3(Psi3) - Psi3*(Psi3*(84 - 5*Psi4*r + (-48 + 63*Psi4*r)*Exponential(Psi2*r) - 36*Exponential(Psi3*r)) + 9*Psi4*(-3 + 2*Exponential(Psi2*r) + Exponential(Psi3*r)))*Power4(Psi2) + Psi2*Psi4*(-21 - Psi3*r + 4*Exponential(Psi1*r) + (17 + 3*Psi3*r)*Exponential(Psi2*r))*Power4(Psi3) + 6*Psi3*(10 + (-2 + 3*Psi4*r)*Exponential(Psi2*r) - 8*Exponential(Psi3*r))*Power5(Psi2) + 2*Psi4*(-1 + Exponential(Psi2*r))*Power5(Psi3) + 12*(-1 + Exponential(Psi3*r))*Power6(Psi2)) + 6*(-(Psi2*Psi3*(6*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(4 + (-4 + 3*Psi4*r)*Exponential(Psi2*r)))) + (Psi3*(8 + (-4 + 3*Psi4*r)*Exponential(Psi2*r) - 4*Exponential(Psi3*r)) + 3*Psi4*(-1 + Exponential(Psi3*r)))*Power2(Psi2) + 3*Psi4*(-1 + Exponential(Psi2*r))*Power2(Psi3) + 4*(-1 + Exponential(Psi3*r))*Power3(Psi2))*Power7(Psi1))))/((Psi1 - Psi3)*Psi4);
}

scalar ComputeEG4G4c(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return Power2(ComputeEG4c(hParam,thetaParam,psiParam));
}

scalar ComputeEG2G6cAW(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar r    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	const scalar Psi4 = psiParam[3];
	
	return (-3*Exponential(-(Psi1*r))*(-1 + Exponential(Psi1*r))*Power2(1/(Psi1 - Psi3))*Power2(1/(Psi2 - Psi3))*Power2(1/Psi3)*Power3(1/phi)*Power3(1/Psi2)*Power4(beta)*Power4(1/(Psi1 - Psi2))*Power8(1/Psi1)*(-15*Power3(phi)*(12*Power11(Psi1)*(-(Psi2*Psi3*(6*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(4 + (-4 + 3*Psi4*r)*Exponential(Psi2*r)))) + (Psi3*(8 + (-4 + 3*Psi4*r)*Exponential(Psi2*r) - 4*Exponential(Psi3*r)) + 3*Psi4*(-1 + Exponential(Psi3*r)))*Power2(Psi2) + 3*Psi4*(-1 + Exponential(Psi2*r))*Power2(Psi3) + 4*(-1 + Exponential(Psi3*r))*Power3(Psi2)) + 2*Psi4*Power2(Psi2 - Psi3)*Power2(Psi3)*Power3(Psi1)*(2*Psi2*Psi3*(-143 + 32*Psi3*r + 11*(13 + 4*Psi3*r)*Exponential(Psi1*r)) + 246*(-1 + Exponential(Psi1*r))*Power2(Psi3) + Power2(Psi2)*(-48 + 44*Psi3*r - 3*Power2(Psi3)*Power2(r) + Exponential(Psi1*r)*(48 + 52*Psi3*r + 5*Power2(Psi3)*Power2(r))) + r*(3 - Psi3*r + (3 + Psi3*r)*Exponential(Psi1*r))*Power3(Psi2))*Power4(Psi2) - 12*Power10(Psi1)*(-(Psi3*(-2*Psi4*(11 + Psi3*r) + (23*Psi4 + Psi3*(-4 + 8*Psi4*r))*Exponential(Psi2*r) + (4*Psi3 - Psi4)*Exponential(Psi3*r))*Power2(Psi2)) - Psi2*(-5*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(8 + Psi4*r + (-8 + 3*Psi4*r)*Exponential(Psi2*r)))*Power2(Psi3) + (Psi3*(24 - Psi4*r + (-12 + 11*Psi4*r)*Exponential(Psi2*r) - 12*Exponential(Psi3*r)) + 13*Psi4*(-1 + Exponential(Psi3*r)))*Power3(Psi2) + 4*Psi4*(-1 + Exponential(Psi2*r))*Power3(Psi3) + 16*(-1 + Exponential(Psi3*r))*Power4(Psi2)) - Psi4*Power2(Psi2 - Psi3)*Power2(Psi3)*Power3(Psi2)*Power4(Psi1)*(12*Psi2*Psi3*(-68 + 16*Psi3*r + (68 + 25*Psi3*r)*Exponential(Psi1*r)) + 428*(-1 + Exponential(Psi1*r))*Power2(Psi3) + 2*Power2(Psi2)*(-122 + 122*Psi3*r - 7*Power2(Psi3)*Power2(r) + Exponential(Psi1*r)*(122 + 164*Psi3*r + 19*Power2(Psi3)*Power2(r))) + 4*r*(11 - 3*Psi3*r + (13 + 5*Psi3*r)*Exponential(Psi1*r))*Power3(Psi2) + (-1 + Exponential(Psi1*r))*Power2(r)*Power4(Psi2)) + 2*Psi3*Power2(Psi2)*Power2(Psi2 - Psi3)*(Psi2*Psi4*(-272 + 79*Psi3*r + (278 + 133*Psi3*r)*Exponential(Psi1*r) - 6*Exponential(Psi2*r))*Power2(Psi3) + 2*Psi3*Psi4*Power2(Psi2)*(-60 + 84*Psi3*r - 4*Power2(Psi3)*Power2(r) + Exponential(Psi1*r)*(60 + 123*Psi3*r + 17*Power2(Psi3)*Power2(r))) + 2*(-3*Psi4*(-1 + Exponential(Psi1*r)) + Psi3*(-12 + 26*Psi4*r + (12 + 35*Psi4*r)*Exponential(Psi1*r)) + Psi4*(-7 + 19*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power3(Psi2) + 2*Psi4*(-46 + 45*Exponential(Psi1*r) + Exponential(Psi2*r))*Power3(Psi3) + Psi3*Psi4*(-3 + 5*Exponential(Psi1*r))*Power2(r)*Power4(Psi2))*Power5(Psi1) - Psi4*Power2(Psi1)*Power2(Psi2 - Psi3)*Power2(Psi3)*(4*Psi2*Psi3*(-48 + 11*Psi3*r + (48 + 13*Psi3*r)*Exponential(Psi1*r)) + 304*(-1 + Exponential(Psi1*r))*Power2(Psi3) + Power2(Psi2)*(-12 + 12*Psi3*r - Power2(Psi3)*Power2(r) + Exponential(Psi1*r)*(12 + 12*Psi3*r + Power2(Psi3)*Power2(r))))*Power5(Psi2) + 6*Psi1*Psi4*(16*Psi3*(-1 + Exponential(Psi1*r)) + Psi2*(-4 + Psi3*r + (4 + Psi3*r)*Exponential(Psi1*r)))*Power2(Psi2 - Psi3)*Power3(Psi3)*Power6(Psi2) - Psi2*Power6(Psi1)*(2*(2*Psi4*(-10 + 47*Exponential(Psi1*r) - 37*Exponential(Psi2*r)) - 2*Psi3*(48 + 81*Psi4*r + (-36 + 137*Psi4*r)*Exponential(Psi1*r) + 2*(-6 + 13*Psi4*r)*Exponential(Psi2*r)) + Psi4*(-7 + 11*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power3(Psi2)*Power3(Psi3) - Power2(Psi3)*(-8*Psi4*(-4 + 7*Exponential(Psi1*r) - 3*Exponential(Psi2*r)) - 4*Psi3*(72 + 10*Psi4*r + 4*(-15 + 7*Psi4*r)*Exponential(Psi1*r) + 3*(-4 + 5*Psi4*r)*Exponential(Psi2*r)) + Psi4*(-41 + 177*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power4(Psi2) + Psi4*Power2(Psi2)*(116 + 108*Psi3*r + 4*(71 + 13*Psi3*r)*Exponential(Psi2*r) - 9*Power2(Psi3)*Power2(r) + Exponential(Psi1*r)*(-400 + 196*Psi3*r + 57*Power2(Psi3)*Power2(r)))*Power4(Psi3) + 4*Psi3*(3*(-4*Psi3 + Psi4)*Exponential(Psi3*r) + Psi4*(6 + 30*Psi3*r - Power2(Psi3)*Power2(r)) + 3*Exponential(Psi1*r)*(-3*Psi4 + Psi3*(4 + 13*Psi4*r) + 5*Psi4*Power2(Psi3)*Power2(r)))*Power5(Psi2) + 4*Psi2*Psi4*(-8 + 17*Psi3*r + 24*(2 + Psi3*r)*Exponential(Psi1*r) - 2*(20 + Psi3*r)*Exponential(Psi2*r))*Power5(Psi3) + 2*(-6*Psi3*(8 + Psi4*r + (-4 + Psi4*r)*Exponential(Psi1*r) - 4*Exponential(Psi3*r)) + 6*Psi4*(-1 + Exponential(Psi3*r)) + Psi4*(-7 + 19*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power6(Psi2) + 24*Psi4*(-1 + Exponential(Psi2*r))*Power6(Psi3)) - 12*Psi4*(-1 + Exponential(Psi1*r))*Power2(Psi2 - Psi3)*Power4(Psi3)*Power7(Psi2) + 2*Power7(Psi1)*((2*Psi4*(-73 + 15*Exponential(Psi1*r) + 58*Exponential(Psi2*r)) - Psi3*(144 + 75*Psi4*r + (-72 + 113*Psi4*r)*Exponential(Psi1*r) + (-72 + 66*Psi4*r)*Exponential(Psi2*r)) + Psi4*(-7 + 39*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power3(Psi2)*Power3(Psi3) - Power2(Psi3)*(-2*Psi4*(71 + 24*Exponential(Psi1*r) - 95*Exponential(Psi2*r)) + 2*Psi3*(-3*(16 + Psi4*r) + (36 + Psi4*r)*Exponential(Psi1*r) + (12 + 13*Psi4*r)*Exponential(Psi2*r)) + Psi4*(-9 + 71*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power4(Psi2) + Psi4*Power2(Psi2)*(-18 + 32*Psi3*r + 18*(3 + 2*Psi3*r)*Exponential(Psi2*r) - Power2(Psi3)*Power2(r) + Exponential(Psi1*r)*(-36 + 66*Psi3*r + 9*Power2(Psi3)*Power2(r)))*Power4(Psi3) - Psi3*(6*Psi4*(-4 + 7*Exponential(Psi1*r) + Exponential(Psi2*r) - 4*Exponential(Psi3*r)) + Psi3*(-216 - 55*Psi4*r + (72 - 79*Psi4*r)*Exponential(Psi1*r) + (48 - 60*Psi4*r)*Exponential(Psi2*r) + 96*Exponential(Psi3*r)) + Psi4*(-7 + 11*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power5(Psi2) - 2*Psi2*Psi4*(-19 - 3*Psi3*r + (19 + 2*Psi3*r)*Exponential(Psi2*r))*Power5(Psi3) + 2*(-3*Psi3*(4*(6 + Psi4*r) + (-12 + 5*Psi4*r)*Exponential(Psi1*r) - 12*Exponential(Psi3*r)) + 21*Psi4*(-1 + Exponential(Psi3*r)) + Psi4*(-4 + 17*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power6(Psi2) - 2*Psi4*(-1 + Exponential(Psi2*r))*Power6(Psi3) + 24*(-1 + Exponential(Psi3*r))*Power7(Psi2)) + (Power2(Psi3)*(-12*Psi4*(29 + 10*Exponential(Psi1*r) - 39*Exponential(Psi2*r)) + 4*Psi3*(48 + Psi4*r + 3*(-4 + 7*Psi4*r)*Exponential(Psi1*r) + 36*(-1 + 2*Psi4*r)*Exponential(Psi2*r)) + Psi4*(1 + 15*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power3(Psi2) - 4*Power2(Psi2)*(-(Psi4*(94 + 15*Exponential(Psi1*r) - 109*Exponential(Psi2*r))) + Psi3*(-48 - 13*Psi4*r + 12*Exponential(Psi1*r) + 36*Exponential(Psi2*r)) + Psi4*(-1 + 9*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power3(Psi3) + 2*Psi3*(-2*Psi3*(192 + 29*Psi4*r + 6*(-10 + 7*Psi4*r)*Exponential(Psi1*r) + (-60 + 52*Psi4*r)*Exponential(Psi2*r) - 72*Exponential(Psi3*r)) + 6*Psi4*(-16 + 5*Exponential(Psi1*r) + 17*Exponential(Psi2*r) - 6*Exponential(Psi3*r)) + Psi4*(-7 + 39*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power4(Psi2) - 4*Psi2*Psi4*(5 + 3*Psi3*r + 5*(-1 + Psi3*r)*Exponential(Psi2*r))*Power4(Psi3) - 3*(-4*Psi3*(16 + 6*Psi4*r + (-12 + 7*Psi4*r)*Exponential(Psi1*r) + (4 - 5*Psi4*r)*Exponential(Psi2*r) - 8*Exponential(Psi3*r)) + 72*Psi4*(-1 + Exponential(Psi3*r)) + Psi4*(-3 + 19*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power5(Psi2) + 32*Psi4*(-1 + Exponential(Psi2*r))*Power5(Psi3) - 192*(-1 + Exponential(Psi3*r))*Power6(Psi2))*Power8(Psi1) + 2*(Power2(Psi2)*Power2(Psi3)*(-56*Psi4*(-1 + Exponential(Psi2*r)) - 2*Psi3*(72 + 6*Psi4*r + 3*(-4 + 3*Psi4*r)*Exponential(Psi1*r) + (-60 + 53*Psi4*r)*Exponential(Psi2*r)) + Psi4*(-1 + 9*Exponential(Psi1*r))*Power2(Psi3)*Power2(r)) - 2*Psi3*(3*Psi4*(-29 + 33*Exponential(Psi2*r) - 4*Exponential(Psi3*r)) + Psi3*(-3*(36 + 7*Psi4*r) - 6*(-4 + 3*Psi4*r)*Exponential(Psi1*r) + (36 - 7*Psi4*r)*Exponential(Psi2*r) + 48*Exponential(Psi3*r)) + Psi4*(-1 + 9*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power3(Psi2) + 2*Psi2*(53*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(-3*(4 + Psi4*r) + (12 + 7*Psi4*r)*Exponential(Psi2*r)))*Power3(Psi3) + (132*Psi4*(-1 + Exponential(Psi3*r)) - 6*Psi3*(4*(-4 + Psi4*r) + (-4 + 3*Psi4*r)*Exponential(Psi1*r) + (12 - 13*Psi4*r)*Exponential(Psi2*r) + 8*Exponential(Psi3*r)) + Psi4*(-1 + 9*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power4(Psi2) - 8*Psi4*(-1 + Exponential(Psi2*r))*Power4(Psi3) + 144*(-1 + Exponential(Psi3*r))*Power5(Psi2))*Power9(Psi1)) + 30*Psi1*Power2(phi)*(24*Power11(Psi1)*(-(Psi2*Psi3*(6*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(4 + (-4 + 3*Psi4*r)*Exponential(Psi2*r)))) + (Psi3*(8 + (-4 + 3*Psi4*r)*Exponential(Psi2*r) - 4*Exponential(Psi3*r)) + 3*Psi4*(-1 + Exponential(Psi3*r)))*Power2(Psi2) + 3*Psi4*(-1 + Exponential(Psi2*r))*Power2(Psi3) + 4*(-1 + Exponential(Psi3*r))*Power3(Psi2)) + 2*Psi4*Power2(Psi2 - Psi3)*Power2(Psi3)*Power3(Psi1)*(Psi2*Psi3*(-572 + 109*Psi3*r + 13*(44 + 15*Psi3*r)*Exponential(Psi1*r)) + 492*(-1 + Exponential(Psi1*r))*Power2(Psi3) + Power2(Psi2)*(-96 + 74*Psi3*r - 3*Power2(Psi3)*Power2(r) + 2*Exponential(Psi1*r)*(48 + 59*Psi3*r + 7*Power2(Psi3)*Power2(r))) + r*(5 - Psi3*r + (7 + 3*Psi3*r)*Exponential(Psi1*r))*Power3(Psi2))*Power4(Psi2) - 6*Power10(Psi1)*(-2*Psi3*(-(Psi4*(44 + 3*Psi3*r)) + (46*Psi4 + Psi3*(-8 + 15*Psi4*r))*Exponential(Psi2*r) + (8*Psi3 - 2*Psi4)*Exponential(Psi3*r))*Power2(Psi2) - Psi2*(-20*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(32 + 3*Psi4*r + (-32 + 13*Psi4*r)*Exponential(Psi2*r)))*Power2(Psi3) + (Psi3*(96 - 3*Psi4*r + (-48 + 43*Psi4*r)*Exponential(Psi2*r) - 48*Exponential(Psi3*r)) + 52*Psi4*(-1 + Exponential(Psi3*r)))*Power3(Psi2) + 16*Psi4*(-1 + Exponential(Psi2*r))*Power3(Psi3) + 64*(-1 + Exponential(Psi3*r))*Power4(Psi2)) - Psi4*Power2(Psi2 - Psi3)*Power2(Psi3)*Power3(Psi2)*Power4(Psi1)*(4*Psi2*Psi3*(-408 + 83*Psi3*r + (408 + 163*Psi3*r)*Exponential(Psi1*r)) + 856*(-1 + Exponential(Psi1*r))*Power2(Psi3) + 2*Power2(Psi2)*(-244 + 209*Psi3*r - 7*Power2(Psi3)*Power2(r) + Exponential(Psi1*r)*(244 + 363*Psi3*r + 50*Power2(Psi3)*Power2(r))) + 2*r*(37 - 6*Psi3*r + (59 + 28*Psi3*r)*Exponential(Psi1*r))*Power3(Psi2) + (-1 + 3*Exponential(Psi1*r))*Power2(r)*Power4(Psi2)) + 2*Psi3*Power2(Psi2)*Power2(Psi2 - Psi3)*(Psi2*Psi4*(-544 + 139*Psi3*r + 4*(139 + 71*Psi3*r)*Exponential(Psi1*r) + (-12 + Psi3*r)*Exponential(Psi2*r))*Power2(Psi3) + Psi3*Psi4*Power2(Psi2)*(-3*Psi3*r*Exponential(Psi2*r) - 8*(30 - 37*Psi3*r + Power2(Psi3)*Power2(r)) + 5*Exponential(Psi1*r)*(48 + 107*Psi3*r + 17*Power2(Psi3)*Power2(r))) + (-12*Psi4*(-1 + Exponential(Psi1*r)) + Psi3*(-48 + 91*Psi4*r + 3*(16 + 51*Psi4*r)*Exponential(Psi1*r)) + 2*Psi4*(-7 + 50*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power3(Psi2) + 4*Psi4*(-46 + 45*Exponential(Psi1*r) + Exponential(Psi2*r))*Power3(Psi3) + Psi3*Psi4*(-3 + 14*Exponential(Psi1*r))*Power2(r)*Power4(Psi2))*Power5(Psi1) - Psi4*Power2(Psi1)*Power2(Psi2 - Psi3)*Power2(Psi3)*(2*Psi2*Psi3*(-192 + 37*Psi3*r + (192 + 59*Psi3*r)*Exponential(Psi1*r)) + 608*(-1 + Exponential(Psi1*r))*Power2(Psi3) + Power2(Psi2)*(-24 + 20*Psi3*r - Power2(Psi3)*Power2(r) + Exponential(Psi1*r)*(24 + 28*Psi3*r + 3*Power2(Psi3)*Power2(r))))*Power5(Psi2) + 2*Psi1*Psi4*(96*Psi3*(-1 + Exponential(Psi1*r)) + Psi2*(-24 + 5*Psi3*r + (24 + 7*Psi3*r)*Exponential(Psi1*r)))*Power2(Psi2 - Psi3)*Power3(Psi3)*Power6(Psi2) - Psi2*Power6(Psi1)*(2*(4*Psi4*(-10 + 47*Exponential(Psi1*r) - 37*Exponential(Psi2*r)) - Psi3*(192 + 293*Psi4*r + (-144 + 589*Psi4*r)*Exponential(Psi1*r) + (-48 + 94*Psi4*r)*Exponential(Psi2*r)) + Psi4*(-7 + 33*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power3(Psi2)*Power3(Psi3) - Power2(Psi3)*(-16*Psi4*(-4 + 7*Exponential(Psi1*r) - 3*Exponential(Psi2*r)) - 2*Psi3*(9*(32 + 3*Psi4*r) + 120*(-2 + Psi4*r)*Exponential(Psi1*r) + (-48 + 65*Psi4*r)*Exponential(Psi2*r)) + Psi4*(-41 + 443*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power4(Psi2) + Psi4*Power2(Psi2)*(232 + 204*Psi3*r + (568 + 82*Psi3*r)*Exponential(Psi2*r) - 9*Power2(Psi3)*Power2(r) + Exponential(Psi1*r)*(-800 + 426*Psi3*r + 137*Power2(Psi3)*Power2(r)))*Power4(Psi3) + 4*Psi3*(-3*Psi3*Psi4*r*Exponential(Psi2*r) + 6*(-4*Psi3 + Psi4)*Exponential(Psi3*r) + Psi4*(12 + 56*Psi3*r - Power2(Psi3)*Power2(r)) + Exponential(Psi1*r)*(-18*Psi4 + Psi3*(24 + 85*Psi4*r) + 35*Psi4*Power2(Psi3)*Power2(r)))*Power5(Psi2) + 2*Psi2*Psi4*(-32 + 61*Psi3*r + (192 + 101*Psi3*r)*Exponential(Psi1*r) - 2*(80 + 3*Psi3*r)*Exponential(Psi2*r))*Power5(Psi3) + 2*(-3*Psi3*(32 + 3*Psi4*r + (-16 + 5*Psi4*r)*Exponential(Psi1*r) - 16*Exponential(Psi3*r)) + 12*Psi4*(-1 + Exponential(Psi3*r)) + Psi4*(-7 + 50*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power6(Psi2) + 48*Psi4*(-1 + Exponential(Psi2*r))*Power6(Psi3)) - 24*Psi4*(-1 + Exponential(Psi1*r))*Power2(Psi2 - Psi3)*Power4(Psi3)*Power7(Psi2) + 2*Power7(Psi1)*((4*Psi4*(-73 + 15*Exponential(Psi1*r) + 58*Exponential(Psi2*r)) - 2*Psi3*(144 + 65*Psi4*r + 24*(-3 + 5*Psi4*r)*Exponential(Psi1*r) + (-72 + 69*Psi4*r)*Exponential(Psi2*r)) + Psi4*(-7 + 95*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power3(Psi2)*Power3(Psi3) - Power2(Psi3)*(-4*Psi4*(71 + 24*Exponential(Psi1*r) - 95*Exponential(Psi2*r)) + 2*Psi3*(-96 + 6*Psi4*r + 3*(24 + Psi4*r)*Exponential(Psi1*r) + (24 + 13*Psi4*r)*Exponential(Psi2*r)) + 3*Psi4*(-3 + 56*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power4(Psi2) + Psi4*Power2(Psi2)*(-36 + 64*Psi3*r + (108 + 65*Psi3*r)*Exponential(Psi2*r) - Power2(Psi3)*Power2(r) + Exponential(Psi1*r)*(-72 + 139*Psi3*r + 21*Power2(Psi3)*Power2(r)))*Power4(Psi3) - Psi3*(12*Psi4*(-4 + 7*Exponential(Psi1*r) + Exponential(Psi2*r) - 4*Exponential(Psi3*r)) + Psi3*(-432 - 103*Psi4*r - 16*(-9 + 11*Psi4*r)*Exponential(Psi1*r) + (96 - 109*Psi4*r)*Exponential(Psi2*r) + 192*Exponential(Psi3*r)) + Psi4*(-7 + 33*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power5(Psi2) - Psi2*Psi4*(-76 - 11*Psi3*r + (76 + 7*Psi3*r)*Exponential(Psi2*r))*Power5(Psi3) + (-3*Psi3*(12*(8 + Psi4*r) + (-48 + 23*Psi4*r)*Exponential(Psi1*r) + Psi4*r*Exponential(Psi2*r) - 48*Exponential(Psi3*r)) + 84*Psi4*(-1 + Exponential(Psi3*r)) + Psi4*(-8 + 85*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power6(Psi2) - 4*Psi4*(-1 + Exponential(Psi2*r))*Power6(Psi3) + 48*(-1 + Exponential(Psi3*r))*Power7(Psi2)) + (Power2(Psi3)*(-24*Psi4*(29 + 10*Exponential(Psi1*r) - 39*Exponential(Psi2*r)) + 2*Psi3*(192 + 19*Psi4*r + (-48 + 91*Psi4*r)*Exponential(Psi1*r) + 2*(-72 + 133*Psi4*r)*Exponential(Psi2*r)) + Psi4*(1 + 31*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power3(Psi2) - 2*Power2(Psi2)*(-4*Psi4*(94 + 15*Exponential(Psi1*r) - 109*Exponential(Psi2*r)) - Psi3*(192 + 37*Psi4*r + (-48 + Psi4*r)*Exponential(Psi1*r) + 2*(-72 + 7*Psi4*r)*Exponential(Psi2*r)) + 2*Psi4*(-1 + 21*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power3(Psi3) + 2*Psi3*(12*Psi4*(-16 + 5*Exponential(Psi1*r) + 17*Exponential(Psi2*r) - 6*Exponential(Psi3*r)) + Psi3*(-768 - 97*Psi4*r - 5*(-48 + 37*Psi4*r)*Exponential(Psi1*r) - 30*(-8 + 7*Psi4*r)*Exponential(Psi2*r) + 288*Exponential(Psi3*r)) + Psi4*(-7 + 95*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power4(Psi2) - 2*Psi2*Psi4*(20 + 13*Psi3*r + (-20 + 19*Psi3*r)*Exponential(Psi2*r))*Power4(Psi3) + (6*Psi3*(64 + 18*Psi4*r + (-48 + 31*Psi4*r)*Exponential(Psi1*r) + (16 - 17*Psi4*r)*Exponential(Psi2*r) - 32*Exponential(Psi3*r)) - 432*Psi4*(-1 + Exponential(Psi3*r)) - Psi4*(-9 + 137*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power5(Psi2) + 64*Psi4*(-1 + Exponential(Psi2*r))*Power5(Psi3) - 384*(-1 + Exponential(Psi3*r))*Power6(Psi2))*Power8(Psi1) + 2*(Power2(Psi2)*Power2(Psi3)*(-112*Psi4*(-1 + Exponential(Psi2*r)) - Psi3*(288 + 22*Psi4*r + (-48 + 39*Psi4*r)*Exponential(Psi1*r) + (-240 + 211*Psi4*r)*Exponential(Psi2*r)) + Psi4*(-1 + 21*Exponential(Psi1*r))*Power2(Psi3)*Power2(r)) - Psi3*(12*Psi4*(-29 + 33*Exponential(Psi2*r) - 4*Exponential(Psi3*r)) + Psi3*(-432 - 65*Psi4*r + (96 - 78*Psi4*r)*Exponential(Psi1*r) + (144 - 41*Psi4*r)*Exponential(Psi2*r) + 192*Exponential(Psi3*r)) + 2*Psi4*(-1 + 21*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power3(Psi2) + Psi2*(212*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(-48 - 7*Psi4*r + (48 + 23*Psi4*r)*Exponential(Psi2*r)))*Power3(Psi3) + (264*Psi4*(-1 + Exponential(Psi3*r)) - 3*Psi3*(-64 + 12*Psi4*r + (-16 + 13*Psi4*r)*Exponential(Psi1*r) + (48 - 49*Psi4*r)*Exponential(Psi2*r) + 32*Exponential(Psi3*r)) + Psi4*(-1 + 21*Exponential(Psi1*r))*Power2(Psi3)*Power2(r))*Power4(Psi2) - 16*Psi4*(-1 + Exponential(Psi2*r))*Power4(Psi3) + 288*(-1 + Exponential(Psi3*r))*Power5(Psi2))*Power9(Psi1)) - 10*phi*Power2(Psi1)*(90*Power11(Psi1)*(-(Psi2*Psi3*(6*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(4 + (-4 + 3*Psi4*r)*Exponential(Psi2*r)))) + (Psi3*(8 + (-4 + 3*Psi4*r)*Exponential(Psi2*r) - 4*Exponential(Psi3*r)) + 3*Psi4*(-1 + Exponential(Psi3*r)))*Power2(Psi2) + 3*Psi4*(-1 + Exponential(Psi2*r))*Power2(Psi3) + 4*(-1 + Exponential(Psi3*r))*Power3(Psi2)) + Psi3*Power2(Psi2 - Psi3)*Power3(Psi1)*(Psi2*(2*Psi3*(4*(6 + 71*Psi4*r) + (-24 + 803*Psi4*r)*Exponential(Psi1*r)) + 3*Psi4*(-1546 + 1541*Exponential(Psi1*r) + 5*Exponential(Psi2*r)))*Power2(Psi3) + Psi3*Power2(Psi2)*(1031*Psi4*(-1 + Exponential(Psi1*r)) + Psi3*(-48 + 445*Psi4*r + 3*(16 + 381*Psi4*r)*Exponential(Psi1*r)) + 156*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r)) + 2*(6*Psi4*(-1 + Exponential(Psi1*r)) + Psi3*(24 + 25*Psi4*r + (-24 + 49*Psi4*r)*Exponential(Psi1*r)) + 18*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power3(Psi2) + Psi4*(-3494 + 3499*Exponential(Psi1*r) - 5*Exponential(Psi2*r))*Power3(Psi3))*Power4(Psi2) - 3*Power10(Psi1)*(-(Psi3*(9*Psi4*(-70 + 73*Exponential(Psi2*r) - 3*Exponential(Psi3*r)) + 2*Psi3*(-2*(5 + 6*Psi4*r) + (-48 + 93*Psi4*r)*Exponential(Psi2*r) + 58*Exponential(Psi3*r)))*Power2(Psi2)) - 3*Psi2*(-45*Psi4*(-1 + Exponential(Psi2*r)) + 4*Psi3*(20 + Psi4*r + (-20 + 9*Psi4*r)*Exponential(Psi2*r)))*Power2(Psi3) + (2*Psi3*(340 - 6*Psi4*r + 21*(-8 + 7*Psi4*r)*Exponential(Psi2*r) - 172*Exponential(Psi3*r)) + 375*Psi4*(-1 + Exponential(Psi3*r)))*Power3(Psi2) + 120*Psi4*(-1 + Exponential(Psi2*r))*Power3(Psi3) + 460*(-1 + Exponential(Psi3*r))*Power4(Psi2)) - Psi4*Power2(Psi1)*Power2(Psi2 - Psi3)*Power2(Psi3)*(Psi2*Psi3*(-1603 + 191*Psi3*r + (1603 + 501*Psi3*r)*Exponential(Psi1*r)) + 2169*(-1 + Exponential(Psi1*r))*Power2(Psi3) + 2*Power2(Psi2)*(-74 + 32*Psi3*r + Exponential(Psi1*r)*(74 + 74*Psi3*r + 9*Power2(Psi3)*Power2(r))))*Power5(Psi2) + 2*Psi1*Psi4*(346*Psi3*(-1 + Exponential(Psi1*r)) + Psi2*(-106 + 13*Psi3*r + (106 + 31*Psi3*r)*Exponential(Psi1*r)))*Power2(Psi2 - Psi3)*Power3(Psi3)*Power6(Psi2) + Power3(Psi2)*Power4(Psi1)*((Psi3*(-540 + 1265*Psi4*r + (504 + 3191*Psi4*r)*Exponential(Psi1*r) + (96 - 156*Psi4*r)*Exponential(Psi2*r) - 60*Exponential(Psi3*r)) - Psi4*(-1483 + 1301*Exponential(Psi1*r) + 197*Exponential(Psi2*r) - 15*Exponential(Psi3*r)) + 732*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power3(Psi2)*Power3(Psi3) + 2*Power2(Psi3)*(-3*Psi4*(-414 + 410*Exponential(Psi1*r) + 5*Exponential(Psi2*r) - Exponential(Psi3*r)) + Psi3*(156 - 305*Psi4*r - (174 + 931*Psi4*r)*Exponential(Psi1*r) + 15*(-2 + 3*Psi4*r)*Exponential(Psi2*r) + 48*Exponential(Psi3*r)) + 42*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power4(Psi2) - Power2(Psi2)*(-(Psi4*(-7492 + 7091*Exponential(Psi1*r) + 401*Exponential(Psi2*r))) - 2*Psi3*(240 + 266*Psi4*r + 2*(-99 + 503*Psi4*r)*Exponential(Psi1*r) + (-42 + 39*Psi4*r)*Exponential(Psi2*r)) + 522*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power4(Psi3) - Psi3*(3*Psi4*(-15 + 16*Exponential(Psi1*r) - Exponential(Psi3*r)) + Psi3*(-36 + 293*Psi4*r + (-48 + 687*Psi4*r)*Exponential(Psi1*r) + 84*Exponential(Psi3*r)) + 276*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power5(Psi2) - Psi2*(Psi4*(-446 + 199*Exponential(Psi1*r) + 247*Exponential(Psi2*r)) + 2*Psi3*(96 + 441*Psi4*r + (-72 + 1321*Psi4*r)*Exponential(Psi1*r) + 6*(-4 + Psi4*r)*Exponential(Psi2*r)))*Power5(Psi3) - 6*(2*Psi3*(8 + Psi4*r + (-4 + Psi4*r)*Exponential(Psi1*r) - 4*Exponential(Psi3*r)) - 2*Psi4*(-1 + Exponential(Psi3*r)) + 3*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power6(Psi2) - Psi4*(-3046 + 3083*Exponential(Psi1*r) - 37*Exponential(Psi2*r))*Power6(Psi3)) - 88*Psi4*(-1 + Exponential(Psi1*r))*Power2(Psi2 - Psi3)*Power4(Psi3)*Power7(Psi2) + Power7(Psi1)*(2*(-2*Psi3*(648 + 145*Psi4*r + (-306 + 469*Psi4*r)*Exponential(Psi1*r) + 3*(-94 + 95*Psi4*r)*Exponential(Psi2*r) - 60*Exponential(Psi3*r)) + Psi4*(-676 + 219*Exponential(Psi1*r) + 487*Exponential(Psi2*r) - 30*Exponential(Psi3*r)) + 456*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power3(Psi2)*Power3(Psi3) - Power2(Psi3)*(-3*Psi4*(579 + 239*Exponential(Psi1*r) - 714*Exponential(Psi2*r) - 104*Exponential(Psi3*r)) + Psi3*(-2208 + 389*Psi4*r + (1416 + 139*Psi4*r)*Exponential(Psi1*r) + (840 - 828*Psi4*r)*Exponential(Psi2*r) - 48*Exponential(Psi3*r)) + 1560*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power4(Psi2) + Power2(Psi2)*(-(Psi4*(421 + 531*Exponential(Psi1*r) - 952*Exponential(Psi2*r))) + Psi3*(240 + 429*Psi4*r + (-36 + 1121*Psi4*r)*Exponential(Psi1*r) + 12*(-17 + 32*Psi4*r)*Exponential(Psi2*r)) + 192*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power4(Psi3) - 2*Psi3*(3*Psi4*(66 + 104*Exponential(Psi1*r) - 109*Exponential(Psi2*r) - 61*Exponential(Psi3*r)) + Psi3*(-1164 - 241*Psi4*r + (348 - 711*Psi4*r)*Exponential(Psi1*r) - 6*(-24 + 29*Psi4*r)*Exponential(Psi2*r) + 672*Exponential(Psi3*r)) + 192*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power5(Psi2) - 2*Psi2*(253*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(-8*(3 + 4*Psi4*r) + 3*(8 + 7*Psi4*r)*Exponential(Psi2*r)))*Power5(Psi3) + 6*(-(Psi3*(400 + Psi4*r + 22*(-7 + 4*Psi4*r)*Exponential(Psi1*r) + 21*(-2 + 3*Psi4*r)*Exponential(Psi2*r) - 204*Exponential(Psi3*r))) + 16*Psi4*(-1 + Exponential(Psi3*r)) + 140*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power6(Psi2) - 22*Psi4*(-1 + Exponential(Psi2*r))*Power6(Psi3) - 168*(-1 + Exponential(Psi3*r))*Power7(Psi2)) + Power2(Psi2)*Power5(Psi1)*(-((2*Psi3*(36 + 999*Psi4*r + 4*(-3 + 656*Psi4*r)*Exponential(Psi1*r) - 48*(-2 + 5*Psi4*r)*Exponential(Psi2*r) - 120*Exponential(Psi3*r)) + Psi4*(-1130 + 847*Exponential(Psi1*r) + 223*Exponential(Psi2*r) + 60*Exponential(Psi3*r)) + 636*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power3(Psi2)*Power3(Psi3)) - Power2(Psi3)*(-(Psi4*(-2915 + 2488*Exponential(Psi1*r) + 499*Exponential(Psi2*r) - 72*Exponential(Psi3*r))) + Psi3*(-888 - 301*Psi4*r + (636 - 1361*Psi4*r)*Exponential(Psi1*r) + (-60 + 66*Psi4*r)*Exponential(Psi2*r) + 312*Exponential(Psi3*r)) + 1092*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power4(Psi2) + Power2(Psi2)*(-(Psi4*(-4802 + 4537*Exponential(Psi1*r) + 265*Exponential(Psi2*r))) + 2*Psi3*(-240 + 121*Psi4*r + 2*(81 + 7*Psi4*r)*Exponential(Psi1*r) + (78 - 138*Psi4*r)*Exponential(Psi2*r)) + 840*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power4(Psi3) + Psi3*(3*Psi4*(-12 + 8*Exponential(Psi1*r) + 5*Exponential(Psi2*r) - Exponential(Psi3*r)) + 2*Psi3*(-516 + 317*Psi4*r + (312 + 755*Psi4*r)*Exponential(Psi1*r) + (60 - 90*Psi4*r)*Exponential(Psi2*r) + 144*Exponential(Psi3*r)) + 732*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power5(Psi2) + Psi2*(Psi4*(-1748 + 1543*Exponential(Psi1*r) + 205*Exponential(Psi2*r)) + 2*Psi3*(144 + 379*Psi4*r + 12*(-6 + 95*Psi4*r)*Exponential(Psi1*r) + 3*(-24 + 7*Psi4*r)*Exponential(Psi2*r)))*Power5(Psi3) + 3*(Psi3*(3*(40 + 7*Psi4*r) + (-64 + 23*Psi4*r)*Exponential(Psi1*r) - 56*Exponential(Psi3*r)) - 33*Psi4*(-1 + Exponential(Psi3*r)) + 52*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power6(Psi2) + 3*Psi4*(-444 + 443*Exponential(Psi1*r) + Exponential(Psi2*r))*Power6(Psi3) - 48*(-1 + Exponential(Psi3*r))*Power7(Psi2)) + Psi2*Power6(Psi1)*(-((Psi4*(1103 + 962*Exponential(Psi1*r) - 1975*Exponential(Psi2*r) - 90*Exponential(Psi3*r)) + Psi3*(-1848 - 1531*Psi4*r - 21*(-56 + 221*Psi4*r)*Exponential(Psi1*r) - 6*(-52 + 67*Psi4*r)*Exponential(Psi2*r) + 360*Exponential(Psi3*r)) + 384*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power3(Psi2)*Power3(Psi3)) + Power2(Psi3)*(-(Psi4*(-1105 + 476*Exponential(Psi1*r) + 857*Exponential(Psi2*r) - 228*Exponential(Psi3*r))) + Psi3*(-3072 + 272*Psi4*r + (2184 - 578*Psi4*r)*Exponential(Psi1*r) - 150*(-4 + 7*Psi4*r)*Exponential(Psi2*r) + 288*Exponential(Psi3*r)) + 2190*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power4(Psi2) - Power2(Psi2)*(Psi4*(577 + 686*Psi3*r) + 3*(709*Psi4 + 6*Psi3*(-2 + 3*Psi4*r))*Exponential(Psi2*r) + 2*Exponential(Psi1*r)*(-1352*Psi4 + Psi3*(18 + 911*Psi4*r) + 324*Psi4*Power2(Psi3)*Power2(r)))*Power4(Psi3) - Psi3*(-3*Psi4*(68 + 54*Exponential(Psi1*r) - 89*Exponential(Psi2*r) - 33*Exponential(Psi3*r)) + 2*Psi3*(-654 + 339*Psi4*r + (468 + 727*Psi4*r)*Exponential(Psi1*r) + (204 - 300*Psi4*r)*Exponential(Psi2*r) - 18*Exponential(Psi3*r)) + 636*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power5(Psi2) - Psi2*(Psi4*(-409 + 1428*Exponential(Psi1*r) - 1019*Exponential(Psi2*r)) + Psi3*(192 + 343*Psi4*r + (-48 + 805*Psi4*r)*Exponential(Psi1*r) - 12*(12 + Psi4*r)*Exponential(Psi2*r)))*Power5(Psi3) - 3*(-77*Psi4*(-1 + Exponential(Psi3*r)) + Psi3*(8*(-15 + 4*Psi4*r) + (28 - 6*Psi4*r)*Exponential(Psi1*r) + (20 - 30*Psi4*r)*Exponential(Psi2*r) + 72*Exponential(Psi3*r)) + 174*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power6(Psi2) - 193*Psi4*(-1 + Exponential(Psi2*r))*Power6(Psi3) + 252*(-1 + Exponential(Psi3*r))*Power7(Psi2)) + (2*Power2(Psi3)*(-(Psi4*(1339 + 474*Exponential(Psi1*r) - 1714*Exponential(Psi2*r) - 99*Exponential(Psi3*r))) + Psi3*(636 + 111*Psi4*r + 2*(-75 + 188*Psi4*r)*Exponential(Psi1*r) + 6*(-65 + 116*Psi4*r)*Exponential(Psi2*r) - 96*Exponential(Psi3*r)) + 60*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power3(Psi2) - Power2(Psi2)*(-(Psi4*(2513 + 474*Exponential(Psi1*r) - 3002*Exponential(Psi2*r) + 15*Exponential(Psi3*r))) + Psi3*(-9*(164 + 11*Psi4*r) + (384 - 17*Psi4*r)*Exponential(Psi1*r) - 6*(-172 + 49*Psi4*r)*Exponential(Psi2*r) + 60*Exponential(Psi3*r)) + 384*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power3(Psi3) + Psi3*(3*Psi4*(-213 + 158*Exponential(Psi1*r) + 228*Exponential(Psi2*r) - 173*Exponential(Psi3*r)) + Psi3*(-5700 - 373*Psi4*r + (1752 - 1555*Psi4*r)*Exponential(Psi1*r) - 6*(-292 + 283*Psi4*r)*Exponential(Psi2*r) + 2196*Exponential(Psi3*r)) + 912*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power4(Psi2) - 2*Psi2*(-53*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(48 + 46*Psi4*r + 6*(-8 + 11*Psi4*r)*Exponential(Psi2*r)))*Power4(Psi3) - 6*(189*Psi4*(-1 + Exponential(Psi3*r)) + Psi3*(-24*(15 + Psi4*r) + (178 - 131*Psi4*r)*Exponential(Psi1*r) + (6 - 24*Psi4*r)*Exponential(Psi2*r) + 176*Exponential(Psi3*r)) + 108*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power5(Psi2) + 224*Psi4*(-1 + Exponential(Psi2*r))*Power5(Psi3) - 888*(-1 + Exponential(Psi3*r))*Power6(Psi2))*Power8(Psi1) + (Power2(Psi2)*Power2(Psi3)*(-(Psi4*(-991 + 943*Exponential(Psi2*r) + 48*Exponential(Psi3*r))) + Psi3*(-2184 - 113*Psi4*r + (372 - 333*Psi4*r)*Exponential(Psi1*r) - 6*(-290 + 253*Psi4*r)*Exponential(Psi2*r) + 72*Exponential(Psi3*r)) + 192*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r)) - Psi3*(-2*Psi3*(1716 + 125*Psi4*r + (-372 + 333*Psi4*r)*Exponential(Psi1*r) + (-624 + 321*Psi4*r)*Exponential(Psi2*r) - 720*Exponential(Psi3*r)) + 3*Psi4*(-728 + 839*Exponential(Psi2*r) - 111*Exponential(Psi3*r)) + 384*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power3(Psi2) + 2*Psi2*(769*Psi4*(-1 + Exponential(Psi2*r)) + 4*Psi3*(-39 - Psi4*r + 3*(13 + 5*Psi4*r)*Exponential(Psi2*r)))*Power3(Psi3) + 3*(Psi3*(312 - 43*Psi4*r + (124 - 111*Psi4*r)*Exponential(Psi1*r) + 4*(-67 + 63*Psi4*r)*Exponential(Psi2*r) - 168*Exponential(Psi3*r)) + 583*Psi4*(-1 + Exponential(Psi3*r)) + 64*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power4(Psi2) - 112*Psi4*(-1 + Exponential(Psi2*r))*Power4(Psi3) + 1872*(-1 + Exponential(Psi3*r))*Power5(Psi2))*Power9(Psi1)) + 2*Power3(Psi1)*(180*Power11(Psi1)*(-(Psi2*Psi3*(6*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(4 + (-4 + 3*Psi4*r)*Exponential(Psi2*r)))) + (Psi3*(8 + (-4 + 3*Psi4*r)*Exponential(Psi2*r) - 4*Exponential(Psi3*r)) + 3*Psi4*(-1 + Exponential(Psi3*r)))*Power2(Psi2) + 3*Psi4*(-1 + Exponential(Psi2*r))*Power2(Psi3) + 4*(-1 + Exponential(Psi3*r))*Power3(Psi2)) - 15*Power10(Psi1)*(-(Psi3*(9*Psi4*(-26 + 27*Exponential(Psi2*r) - Exponential(Psi3*r)) + Psi3*(-20 + 6*(-4 + 9*Psi4*r)*Exponential(Psi2*r) + 44*Exponential(Psi3*r)))*Power2(Psi2)) - 3*Psi2*(-15*Psi4*(-1 + Exponential(Psi2*r)) + 16*Psi3*(2 + (-2 + Psi4*r)*Exponential(Psi2*r)))*Power2(Psi3) + (2*Psi3*(124 + (-60 + 51*Psi4*r)*Exponential(Psi2*r) - 64*Exponential(Psi3*r)) + 141*Psi4*(-1 + Exponential(Psi3*r)))*Power3(Psi2) + 48*Psi4*(-1 + Exponential(Psi2*r))*Power3(Psi3) + 172*(-1 + Exponential(Psi3*r))*Power4(Psi2)) + (Psi2 - Psi3)*Psi3*Power3(Psi1)*Power4(Psi2)*(-(Power2(Psi2)*Power2(Psi3)*(-5*Psi4*(-1353 + 1334*Exponential(Psi1*r) + 16*Exponential(Psi2*r) + 3*Exponential(Psi3*r)) + 2*Psi3*(-2*(529 + 25*Psi4*r) + (938 + 85*Psi4*r)*Exponential(Psi1*r) + 30*(3 + 2*Psi4*r)*Exponential(Psi2*r) + 30*Exponential(Psi3*r)) + 480*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))) + Psi3*(2*Psi3*(-362 + 75*Psi4*r + (482 + 1435*Psi4*r)*Exponential(Psi1*r) + 15*(-2 + 3*Psi4*r)*Exponential(Psi2*r) - 90*Exponential(Psi3*r)) + Psi4*(-3767 + 3662*Exponential(Psi1*r) + 30*Exponential(Psi2*r) + 75*Exponential(Psi3*r)) + 360*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power3(Psi2) - 2*Psi2*(11*Psi4*(-244 + 229*Exponential(Psi1*r) + 15*Exponential(Psi2*r)) + Psi3*(26*(24 + 5*Psi4*r) + 8*(-63 + 200*Psi4*r)*Exponential(Psi1*r) - 15*(8 + Psi4*r)*Exponential(Psi2*r)))*Power3(Psi3) + 2*(6*Psi4*(-11 + 6*Exponential(Psi1*r) + 5*Exponential(Psi3*r)) + Psi3*(-72 + 5*Psi4*r + (-48 + 250*Psi4*r)*Exponential(Psi1*r) + 120*Exponential(Psi3*r)) + 60*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power4(Psi2) - 2*Psi4*(-2648 + 2683*Exponential(Psi1*r) - 35*Exponential(Psi2*r))*Power4(Psi3)) - Power2(Psi1)*Power2(Psi2 - Psi3)*Power2(Psi3)*(Psi2*Psi3*(16*Psi3*(12 + 5*Psi4*r + (-12 + 65*Psi4*r)*Exponential(Psi1*r)) + Psi4*(-3871 + 3856*Exponential(Psi1*r) + 15*Exponential(Psi2*r))) + Psi4*(-3359 + 3364*Exponential(Psi1*r) - 5*Exponential(Psi2*r))*Power2(Psi3) + 2*Power2(Psi2)*(297*Psi4*(-1 + Exponential(Psi1*r)) + 2*Psi3*(-48 + 5*Psi4*r + 2*(24 + 65*Psi4*r)*Exponential(Psi1*r)) + 30*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r)))*Power5(Psi2) + 2*Psi1*Psi4*(558*Psi3*(-1 + Exponential(Psi1*r)) + Psi2*(-276 + 5*Psi3*r + (276 + 70*Psi3*r)*Exponential(Psi1*r)))*Power2(Psi2 - Psi3)*Power3(Psi3)*Power6(Psi2) - Power3(Psi2)*Power4(Psi1)*(-((2*Psi3*(-1602 + 260*Psi4*r + 176*(12 + 25*Psi4*r)*Exponential(Psi1*r) - 15*(-4 + 9*Psi4*r)*Exponential(Psi2*r) - 570*Exponential(Psi3*r)) + Psi4*(-4337 + 4342*Exponential(Psi1*r) - 440*Exponential(Psi2*r) + 435*Exponential(Psi3*r)) + 2040*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power3(Psi2)*Power3(Psi3)) - Power2(Psi3)*(-(Psi4*(-8468 + 8368*Exponential(Psi1*r) + 205*Exponential(Psi2*r) - 105*Exponential(Psi3*r))) - 30*Psi3*(12*(5 + Psi4*r) + (34 + 93*Psi4*r)*Exponential(Psi1*r) - (6 + 17*Psi4*r)*Exponential(Psi2*r) - 88*Exponential(Psi3*r)) + 360*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power4(Psi2) + 2*Power2(Psi2)*(-(Psi4*(-7421 + 6281*Exponential(Psi1*r) + 1110*Exponential(Psi2*r) + 30*Exponential(Psi3*r))) + Psi3*(-4*(842 + 45*Psi4*r) + (2438 - 1005*Psi4*r)*Exponential(Psi1*r) + 45*(18 + Psi4*r)*Exponential(Psi2*r) + 120*Exponential(Psi3*r)) + 750*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power4(Psi3) + Psi3*(4*Psi3*(-407 + 20*Psi4*r + (122 + 665*Psi4*r)*Exponential(Psi1*r) + 15*(-2 + 3*Psi4*r)*Exponential(Psi2*r) + 315*Exponential(Psi3*r)) + Psi4*(-769 + 274*Exponential(Psi1*r) + 15*Exponential(Psi2*r) + 480*Exponential(Psi3*r)) + 840*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power5(Psi2) + 2*Psi2*(Psi4*(-2669 + 1654*Exponential(Psi1*r) + 1015*Exponential(Psi2*r)) + Psi3*(4*(384 + 55*Psi4*r) + 4*(-234 + 655*Psi4*r)*Exponential(Psi1*r) - 15*(40 + Psi4*r)*Exponential(Psi2*r)))*Power5(Psi3) + 12*(2*Psi3*(12 + (-12 + 5*Psi4*r)*Exponential(Psi1*r)) - 5*Psi4*(-1 + Exponential(Psi3*r)) + 5*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power6(Psi2) + 2*Psi4*(-2332 + 2477*Exponential(Psi1*r) - 145*Exponential(Psi2*r))*Power6(Psi3)) - 150*Psi4*(-1 + Exponential(Psi1*r))*Power2(Psi2 - Psi3)*Power4(Psi3)*Power7(Psi2) + Power7(Psi1)*(2*(5*Psi4*(65 + 48*Exponential(Psi1*r) - 29*Exponential(Psi2*r) - 84*Exponential(Psi3*r)) + Psi3*(-5244 - 175*Psi4*r - 6*(-294 + 295*Psi4*r)*Exponential(Psi1*r) - 30*(-70 + 33*Psi4*r)*Exponential(Psi2*r) + 1380*Exponential(Psi3*r)) + 1200*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power3(Psi2)*Power3(Psi3) - Power2(Psi3)*(-(Psi4*(4999 + 1806*Exponential(Psi1*r) - 3370*Exponential(Psi2*r) - 3435*Exponential(Psi3*r))) + 6*Psi3*(14*(-81 + 5*Psi4*r) + (814 + 155*Psi4*r)*Exponential(Psi1*r) + (270 - 640*Psi4*r)*Exponential(Psi2*r) + 50*Exponential(Psi3*r)) + 3960*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power4(Psi2) + Power2(Psi2)*(20*Psi3*(139 + 16*Psi4*r + (-16 + 119*Psi4*r)*Exponential(Psi1*r) + 9*(-14 + 3*Psi4*r)*Exponential(Psi2*r) + 3*Exponential(Psi3*r)) - Psi4*(2423 + 922*Exponential(Psi1*r) - 3360*Exponential(Psi2*r) + 15*Exponential(Psi3*r)) + 480*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power4(Psi3) - 2*Psi3*(2*Psi4*(1309 + 341*Exponential(Psi1*r) - 855*Exponential(Psi2*r) - 795*Exponential(Psi3*r)) + Psi3*(-3128 - 205*Psi4*r - 8*(29 + 185*Psi4*r)*Exponential(Psi1*r) + 60*(4 + 9*Psi4*r)*Exponential(Psi2*r) + 3120*Exponential(Psi3*r)) + 600*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power5(Psi2) - 20*Psi2*(31*Psi4*(-1 + Exponential(Psi2*r)) + Psi3*(-2*(12 + Psi4*r) + 3*(8 + Psi4*r)*Exponential(Psi2*r)))*Power5(Psi3) + 6*(-235*Psi4*(-1 + Exponential(Psi3*r)) + Psi3*(-1292 + (202 - 145*Psi4*r)*Exponential(Psi1*r) - 30*(-5 + 7*Psi4*r)*Exponential(Psi2*r) + 940*Exponential(Psi3*r)) + 380*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power6(Psi2) + 20*Psi4*(-1 + Exponential(Psi2*r))*Power6(Psi3) - 1920*(-1 + Exponential(Psi3*r))*Power7(Psi2)) + Power2(Psi2)*Power5(Psi1)*(-2*(Psi3*(2152 + 545*Psi4*r + 8*(31 + 700*Psi4*r)*Exponential(Psi1*r) - 45*(16 + 23*Psi4*r)*Exponential(Psi2*r) - 1680*Exponential(Psi3*r)) + 2*Psi4*(-3212 + 2367*Exponential(Psi1*r) + 560*Exponential(Psi2*r) + 285*Exponential(Psi3*r)) + 780*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power3(Psi2)*Power3(Psi3) - Power2(Psi3)*(-(Psi4*(-9347 + 8172*Exponential(Psi1*r) + 2210*Exponential(Psi2*r) - 1035*Exponential(Psi3*r))) + 6*Psi3*(-6*(271 + 10*Psi4*r) + (576 + 50*Psi4*r)*Exponential(Psi1*r) + 5*(20 + 39*Psi4*r)*Exponential(Psi2*r) + 950*Exponential(Psi3*r)) + 3240*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power4(Psi2) + 2*Power2(Psi2)*(-(Psi4*(-2999 + 2369*Exponential(Psi1*r) + 585*Exponential(Psi2*r) + 45*Exponential(Psi3*r))) + 3*Psi3*(-804 + 10*Psi4*r + (514 + 275*Psi4*r)*Exponential(Psi1*r) - 115*(-2 + Psi4*r)*Exponential(Psi2*r) + 60*Exponential(Psi3*r)) + 1140*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power4(Psi3) + 2*Psi3*(Psi4*(-1043 + 188*Exponential(Psi1*r) + 60*Exponential(Psi2*r) + 795*Exponential(Psi3*r)) + Psi3*(26*(-102 + 5*Psi4*r) + 2*(876 + 1175*Psi4*r)*Exponential(Psi1*r) - 195*Psi4*r*Exponential(Psi2*r) + 900*Exponential(Psi3*r)) + 1020*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power5(Psi2) + 2*Psi2*(2*Psi4*(-1423 + 828*Exponential(Psi1*r) + 595*Exponential(Psi2*r)) + Psi3*(1824 + 205*Psi4*r + (-744 + 2290*Psi4*r)*Exponential(Psi1*r) + 45*(-24 + Psi4*r)*Exponential(Psi2*r)))*Power5(Psi3) + (-495*Psi4*(-1 + Exponential(Psi3*r)) + 2*Psi3*(394 + (-574 + 285*Psi4*r)*Exponential(Psi1*r) + 15*(-2 + 3*Psi4*r)*Exponential(Psi2*r) + 210*Exponential(Psi3*r)) + 480*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power6(Psi2) + 2*Psi4*(-1108 + 1173*Exponential(Psi1*r) - 65*Exponential(Psi2*r))*Power6(Psi3) - 240*(-1 + Exponential(Psi3*r))*Power7(Psi2)) + Psi2*Power6(Psi1)*(-((-5*Psi4*(-2067 + 250*Exponential(Psi1*r) + 1535*Exponential(Psi2*r) + 282*Exponential(Psi3*r)) + Psi3*(-8*(1577 + 125*Psi4*r) + (4096 - 8800*Psi4*r)*Exponential(Psi1*r) + 30*(136 + 21*Psi4*r)*Exponential(Psi2*r) + 4440*Exponential(Psi3*r)) + 1200*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power3(Psi2)*Power3(Psi3)) + Power2(Psi3)*(-(Psi4*(-3133 + 2118*Exponential(Psi1*r) + 3805*Exponential(Psi2*r) - 2790*Exponential(Psi3*r))) + 2*Psi3*(-7352 + 30*Psi4*r + (3992 + 630*Psi4*r)*Exponential(Psi1*r) - 15*(-64 + 77*Psi4*r)*Exponential(Psi2*r) + 2400*Exponential(Psi3*r)) + 5940*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power4(Psi2) - Power2(Psi2)*(-5*Psi4*(491 + 670*Exponential(Psi1*r) - 1173*Exponential(Psi2*r) + 12*Exponential(Psi3*r)) + 2*Psi3*(472 + 210*Psi4*r + (98 + 2145*Psi4*r)*Exponential(Psi1*r) - 15*(46 + 11*Psi4*r)*Exponential(Psi2*r) + 120*Exponential(Psi3*r)) + 1680*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power4(Psi3) - Psi3*(-3*Psi4*(1368 + 42*Exponential(Psi1*r) - 445*Exponential(Psi2*r) - 965*Exponential(Psi3*r)) + 2*Psi3*(-2318 + 220*Psi4*r + 4*(632 + 445*Psi4*r)*Exponential(Psi1*r) - 15*(-28 + 85*Psi4*r)*Exponential(Psi2*r) - 630*Exponential(Psi3*r)) + 1560*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power5(Psi2) - Psi2*(Psi4*(-1593 + 2608*Exponential(Psi1*r) - 1015*Exponential(Psi2*r)) + 2*Psi3*(4*(264 + 25*Psi4*r) + 4*(-54 + 205*Psi4*r)*Exponential(Psi1*r) + 15*(-56 + Psi4*r)*Exponential(Psi2*r)))*Power5(Psi3) + (Psi3*(1768 + (932 - 570*Psi4*r)*Exponential(Psi1*r) + 30*(-2 + 3*Psi4*r)*Exponential(Psi2*r) - 2640*Exponential(Psi3*r)) + 1425*Psi4*(-1 + Exponential(Psi3*r)) - 1500*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power6(Psi2) - 475*Psi4*(-1 + Exponential(Psi2*r))*Power6(Psi3) + 1260*(-1 + Exponential(Psi3*r))*Power7(Psi2)) + (Power2(Psi3)*(-(Psi4*(7102 + 2128*Exponential(Psi1*r) - 7205*Exponential(Psi2*r) - 2025*Exponential(Psi3*r))) + 2*Psi3*(4*(507 + 40*Psi4*r) + (-138 + 845*Psi4*r)*Exponential(Psi1*r) + 30*(-35 + 18*Psi4*r)*Exponential(Psi2*r) - 840*Exponential(Psi3*r)) + 240*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power3(Psi2) - Power2(Psi2)*(-(Psi4*(4341 + 1064*Exponential(Psi1*r) - 5600*Exponential(Psi2*r) + 195*Exponential(Psi3*r))) + Psi3*(-3356 + 40*Psi4*r + (896 - 40*Psi4*r)*Exponential(Psi1*r) - 900*(-2 + Psi4*r)*Exponential(Psi2*r) + 660*Exponential(Psi3*r)) + 960*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power3(Psi3) + Psi3*(Psi4*(2101 + 1064*Exponential(Psi1*r) - 1035*Exponential(Psi2*r) - 2130*Exponential(Psi3*r)) - 20*Psi3*(711 + 10*Psi4*r + (-162 + 175*Psi4*r)*Exponential(Psi1*r) + 3*(-66 + 53*Psi4*r)*Exponential(Psi2*r) - 351*Exponential(Psi3*r)) + 2400*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power4(Psi2) - 10*Psi2*(17*Psi4*(-1 + Exponential(Psi2*r)) + 8*Psi3*(12 + Psi4*r + 3*(-4 + Psi4*r)*Exponential(Psi2*r)))*Power4(Psi3) - 2*(405*Psi4*(-1 + Exponential(Psi3*r)) + Psi3*(-3824 + (1034 - 885*Psi4*r)*Exponential(Psi1*r) + (510 - 720*Psi4*r)*Exponential(Psi2*r) + 2280*Exponential(Psi3*r)) + 840*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power5(Psi2) + 320*Psi4*(-1 + Exponential(Psi2*r))*Power5(Psi3) - 120*(-1 + Exponential(Psi3*r))*Power6(Psi2))*Power8(Psi1) + (Power2(Psi2)*Power2(Psi3)*(-4*Psi3*(1291 + 20*Psi4*r + (-196 + 195*Psi4*r)*Exponential(Psi1*r) + 30*(-31 + 23*Psi4*r)*Exponential(Psi2*r) - 165*Exponential(Psi3*r)) - 5*Psi4*(-551 + 458*Exponential(Psi2*r) + 93*Exponential(Psi3*r)) + 480*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r)) - 2*Psi3*(15*Psi4*(-97 + 124*Exponential(Psi2*r) - 27*Exponential(Psi3*r)) + Psi3*(-4*(1021 + 5*Psi4*r) + (784 - 780*Psi4*r)*Exponential(Psi1*r) - 15*(-104 + 67*Psi4*r)*Exponential(Psi2*r) + 1740*Exponential(Psi3*r)) + 480*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power3(Psi2) + 20*Psi2*(151*Psi4*(-1 + Exponential(Psi2*r)) + 2*Psi3*(-6 + Psi4*r + 3*(2 + Psi4*r)*Exponential(Psi2*r)))*Power3(Psi3) + (Psi3*(116 + (784 - 780*Psi4*r)*Exponential(Psi1*r) + 210*(-4 + 3*Psi4*r)*Exponential(Psi2*r) - 60*Exponential(Psi3*r)) + 2805*Psi4*(-1 + Exponential(Psi3*r)) + 480*Psi4*Exponential(Psi1*r)*Power2(Psi3)*Power2(r))*Power4(Psi2) - 160*Psi4*(-1 + Exponential(Psi2*r))*Power4(Psi3) + 2880*(-1 + Exponential(Psi3*r))*Power5(Psi2))*Power9(Psi1))))/Psi4;
}

scalar ComputeEG2G6c(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	return ComputeEG2c(hParam,thetaParam,psiParam)*ComputeEG6c(hParam,thetaParam,psiParam);
}

scalar ComputeEG8c(const scalar *hParam, const scalar thetaParam[], const scalar psiParam[])
{
	const scalar t    = *hParam;
	const scalar beta = thetaParam[0];
	const scalar phi  = thetaParam[3];
	const scalar Psi1 = psiParam[0];
	const scalar Psi2 = psiParam[1];
	const scalar Psi3 = psiParam[2];
	const scalar Psi4 = psiParam[3];
	
	return (Power2(1/(Psi1 - Psi3))*Power2(1/(Psi2 - Psi3))*Power3(1/Psi3)*Power4(beta)*Power4(1/phi)*Power4(1/(Psi1 - Psi2))*Power4(1/Psi2)*Power8(1/Psi1)*(4*(Psi1 - Psi2)*(Psi1 - Psi3)*Power4(Psi1)*(1260*Power10(Psi1)*(-(Psi2*(9*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(4 + 3*Psi4*t + (-4 + 3*Psi4*t)*Exponential(Psi2*t)))*Power2(Psi3)) - (-4 + 4*Psi3*t - 6*Psi4*t + (4 - 3*Psi4*t)*Exponential(Psi2*t))*Power2(Psi2)*Power2(Psi3) + (Psi3*(4 - 3*Psi4*t - 4*Exponential(Psi3*t)) + 3*Psi4*(-1 + Exponential(Psi3*t)) + 8*t*Power2(Psi3))*Power3(Psi2) + 6*Psi4*(-1 + Exponential(Psi2*t))*Power3(Psi3) + 4*(-1 - Psi3*t + Exponential(Psi3*t))*Power4(Psi2)) - (Psi2 - Psi3)*Psi3*Power3(Psi1)*Power4(Psi2)*(-(Power2(Psi2)*Power2(Psi3)*(14*Psi4*(-2483 + 2498*Exponential(Psi1*t) - 12*Exponential(Psi2*t) - 3*Exponential(Psi3*t)) + 12*Psi3*(-119 - 130*Psi4*t + 770*Psi4*t*Exponential(Psi1*t) + 35*(3 + 2*Psi4*t)*Exponential(Psi2*t) + 14*Exponential(Psi3*t)) + t*(1536 + 245*Psi4*t)*Power2(Psi3))) + 6*Psi3*(Psi3*(378 - 2395*Psi4*t - 980*Psi4*t*Exponential(Psi1*t) + 35*(-2 + 3*Psi4*t)*Exponential(Psi2*t) - 308*Exponential(Psi3*t)) - 7*Psi4*(-62 + 75*Exponential(Psi1*t) - 13*Exponential(Psi3*t)) + t*(528 + 35*Psi4*t)*Power2(Psi3))*Power3(Psi2) + 14*Psi2*(Psi4*(-45 + 158*Exponential(Psi1*t) - 113*Exponential(Psi2*t)) + 3*Psi3*(-40 + 337*Psi4*t + 360*Psi4*t*Exponential(Psi1*t) + 5*(8 + Psi4*t)*Exponential(Psi2*t)))*Power3(Psi3) + (504*Psi4*(-1 + Exponential(Psi3*t)) + 672*Psi3*(-3 - 2*Psi4*t + 3*Exponential(Psi3*t)) + t*(-1632 + 35*Psi4*t)*Power2(Psi3))*Power4(Psi2) + 14*Psi4*(-2588 + 2565*Exponential(Psi1*t) + 23*Exponential(Psi2*t))*Power4(Psi3)) + 7*Psi4*Power2(Psi1)*Power2(Psi2 - Psi3)*Power2(Psi3)*(-6*Psi2*Psi3*(-470 + 137*Psi3*t + 10*(47 + 14*Psi3*t)*Exponential(Psi1*t)) - 4258*(-1 + Exponential(Psi1*t))*Power2(Psi3) + Power2(Psi2)*(120 - 394*Psi3*t - 120*(1 + Psi3*t)*Exponential(Psi1*t) + 5*Power2(Psi3)*Power2(t)))*Power5(Psi2) + 14*Psi1*Psi4*(831*Psi3*(-1 + Exponential(Psi1*t)) + Psi2*(-257 + 65*Psi3*t + (257 + 60*Psi3*t)*Exponential(Psi1*t)))*Power2(Psi2 - Psi3)*Power3(Psi3)*Power6(Psi2) + Power3(Psi2)*Power4(Psi1)*(-((14*Psi4*(-1913 + 2154*Exponential(Psi1*t) - 436*Exponential(Psi2*t) + 195*Exponential(Psi3*t)) - 6*Psi3*(-966 + 6963*Psi4*t + 2240*Psi4*t*Exponential(Psi1*t) + 70*(-9 + 2*Psi4*t)*Exponential(Psi2*t) + 1596*Exponential(Psi3*t)) + 7*t*(3264 + 155*Psi4*t)*Power2(Psi3))*Power3(Psi2)*Power3(Psi3)) + Power2(Psi3)*(-42*Psi4*(-49 + 20*Exponential(Psi1*t) - 4*Exponential(Psi2*t) + 33*Exponential(Psi3*t)) - 6*Psi3*(-3094 + 3221*Psi4*t + 2520*Psi4*t*Exponential(Psi1*t) + 70*(1 + 5*Psi4*t)*Exponential(Psi2*t) + 3024*Exponential(Psi3*t)) + 7*t*(3168 + 25*Psi4*t)*Power2(Psi3))*Power4(Psi2) + Power2(Psi2)*(14*Psi4*(-2066 + 3226*Exponential(Psi1*t) - 1151*Exponential(Psi2*t) - 9*Exponential(Psi3*t)) + 6*Psi3*(-2254 + 277*Psi4*t + 3080*Psi4*t*Exponential(Psi1*t) + 70*(31 + 2*Psi4*t)*Exponential(Psi2*t) + 84*Exponential(Psi3*t)) + t*(7968 + 665*Psi4*t)*Power2(Psi3))*Power4(Psi3) + Psi3*(3486*Psi4*(-1 + Exponential(Psi3*t)) + 6*Psi3*(-2*(637 + 620*Psi4*t) + 35*(-2 + 3*Psi4*t)*Exponential(Psi2*t) + 1344*Exponential(Psi3*t)) + 5*t*(-1536 + 49*Psi4*t)*Power2(Psi3))*Power5(Psi2) - 14*Psi2*(-2*Psi4*(-583 + 125*Exponential(Psi1*t) + 458*Exponential(Psi2*t)) + 3*Psi3*(-200 + 409*Psi4*t + 400*Psi4*t*Exponential(Psi1*t) + 5*(40 + Psi4*t)*Exponential(Psi2*t)))*Power5(Psi3) + 24*(Psi3*(16*Psi3 + 21*Psi4)*t - 21*Psi4*(-1 + Exponential(Psi3*t)))*Power6(Psi2) - 14*Psi4*(-1385 + 1262*Exponential(Psi1*t) + 123*Exponential(Psi2*t))*Power6(Psi3)) - 1750*Psi4*(-1 + Exponential(Psi1*t))*Power2(Psi2 - Psi3)*Power4(Psi3)*Power7(Psi2) + Psi2*Power6(Psi1)*(-((14*Psi4*(-2810 + 780*Exponential(Psi1*t) + 1733*Exponential(Psi2*t) + 297*Exponential(Psi3*t)) - 6*Psi3*(-4298 + 356*Psi4*t + 2240*Psi4*t*Exponential(Psi1*t) + 35*(50 + 43*Psi4*t)*Exponential(Psi2*t) + 2548*Exponential(Psi3*t)) + 7*t*(2976 + 35*Psi4*t)*Power2(Psi3))*Power3(Psi2)*Power3(Psi3)) - 2*Power2(Psi3)*(-6*Psi3*(1617 + 1483*Psi4*t - 560*Psi4*t*Exponential(Psi1*t) + 35*(-25 + 6*Psi4*t)*Exponential(Psi2*t) - 742*Exponential(Psi3*t)) - 42*Psi4*(-194 + 65*Exponential(Psi1*t) + 304*Exponential(Psi2*t) - 175*Exponential(Psi3*t)) + 35*t*(24 + 17*Psi4*t)*Power2(Psi3))*Power4(Psi2) + 2*Power2(Psi2)*(-6*Psi3*(469 + 88*Psi4*t + 560*Psi4*t*Exponential(Psi1*t) + 35*(-13 + 7*Psi4*t)*Exponential(Psi2*t) - 14*Exponential(Psi3*t)) + 7*Psi4*(-1069 + 390*Exponential(Psi1*t) + 682*Exponential(Psi2*t) - 3*Exponential(Psi3*t)) + 8*t*(726 + 35*Psi4*t)*Power2(Psi3))*Power4(Psi3) + Psi3*(14112*Psi4*(-1 + Exponential(Psi3*t)) - 12*Psi3*(-861 + 1993*Psi4*t + 105*(-5 + 7*Psi4*t)*Exponential(Psi2*t) + 1386*Exponential(Psi3*t)) + t*(27456 + 875*Psi4*t)*Power2(Psi3))*Power5(Psi2) + 14*Psi2*(-56*Psi4*(-1 + Exponential(Psi2*t)) + 3*Psi3*(280 - 66*Psi4*t + 5*(-56 + Psi4*t)*Exponential(Psi2*t)))*Power5(Psi3) - 36*(217*Psi4*(-1 + Exponential(Psi3*t)) - 7*Psi3*(-72 + 31*Psi4*t + 72*Exponential(Psi3*t)) + 684*t*Power2(Psi3))*Power6(Psi2) + 2562*Psi4*(-1 + Exponential(Psi2*t))*Power6(Psi3) - 8064*(-1 - Psi3*t + Exponential(Psi3*t))*Power7(Psi2)) + Power2(Psi2)*Power5(Psi1)*((14*Psi4*(-1646 + 1814*Exponential(Psi1*t) - 537*Exponential(Psi2*t) + 369*Exponential(Psi3*t)) - 6*Psi3*(-2618 + 6128*Psi4*t + 4480*Psi4*t*Exponential(Psi1*t) + 35*(-14 + 65*Psi4*t)*Exponential(Psi2*t) + 3108*Exponential(Psi3*t)) + 35*t*(1056 + 31*Psi4*t)*Power2(Psi3))*Power3(Psi2)*Power3(Psi3) + Power2(Psi3)*(-42*Psi4*(-166 + 165*Exponential(Psi1*t) + 187*Exponential(Psi2*t) - 186*Exponential(Psi3*t)) + 6*Psi3*(-4466 + 669*Psi4*t + 2800*Psi4*t*Exponential(Psi1*t) + 35*(2 + 43*Psi4*t)*Exponential(Psi2*t) + 4396*Exponential(Psi3*t)) + t*(-29472 + 455*Psi4*t)*Power2(Psi3))*Power4(Psi2) - Power2(Psi2)*(-6*Psi3*(3094 + 1333*Psi4*t + 560*Psi4*t*Exponential(Psi1*t) + 70*(-43 + 11*Psi4*t)*Exponential(Psi2*t) - 84*Exponential(Psi3*t)) + 14*Psi4*(-209 + 2143*Exponential(Psi1*t) - 1925*Exponential(Psi2*t) - 9*Exponential(Psi3*t)) + t*(14688 + 875*Psi4*t)*Power2(Psi3))*Power4(Psi3) - Psi3*(-6*Psi3*(966 + 2964*Psi4*t + 35*(-2 + 3*Psi4*t)*Exponential(Psi2*t) - 896*Exponential(Psi3*t)) + 9702*Psi4*(-1 + Exponential(Psi3*t)) + 5*t*(-960 + 133*Psi4*t)*Power2(Psi3))*Power5(Psi2) - 14*Psi2*(-(Psi4*(575 + 824*Exponential(Psi1*t) - 1399*Exponential(Psi2*t))) + 15*Psi3*(72 - 50*Psi4*t - 32*Psi4*t*Exponential(Psi1*t) + 3*(-24 + Psi4*t)*Exponential(Psi2*t)))*Power5(Psi3) + 24*(147*Psi4*(-1 + Exponential(Psi3*t)) - 7*Psi3*(-23 + 21*Psi4*t + 23*Exponential(Psi3*t)) + 184*t*Power2(Psi3))*Power6(Psi2) + 1078*Psi4*(-1 + Exponential(Psi2*t))*Power6(Psi3) + 2016*(-1 - Psi3*t + Exponential(Psi3*t))*Power7(Psi2)) + 2*Power7(Psi1)*(-((-2*Psi3*(7119 + 3722*Psi4*t + 105*(-57 + 26*Psi4*t)*Exponential(Psi2*t) - 1134*Exponential(Psi3*t)) - 21*Psi4*(-861 + 832*Exponential(Psi2*t) + 29*Exponential(Psi3*t)) + 28*t*(24 + 5*Psi4*t)*Power2(Psi3))*Power3(Psi2)*Power3(Psi3)) + Power2(Psi3)*(-21*Psi4*(-98 + 377*Exponential(Psi2*t) - 279*Exponential(Psi3*t)) - 3*Psi3*(8*(287 + 373*Psi4*t) + 35*(-104 + 85*Psi4*t)*Exponential(Psi2*t) + 1344*Exponential(Psi3*t)) + 2*t*(8028 + 245*Psi4*t)*Power2(Psi3))*Power4(Psi2) - 35*Power2(Psi2)*(464*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(84 + 76*Psi4*t + (-84 + 51*Psi4*t)*Exponential(Psi2*t)) + 2*t*(24 + Psi4*t)*Power2(Psi3))*Power4(Psi3) - 2*Psi3*(2856*Psi4*(-1 + Exponential(Psi3*t)) - 3*Psi3*(-2051 + 977*Psi4*t + 35*(-17 + 24*Psi4*t)*Exponential(Psi2*t) + 2646*Exponential(Psi3*t)) + 4*t*(3363 + 35*Psi4*t)*Power2(Psi3))*Power5(Psi2) + 35*Psi2*(127*Psi4*(-1 + Exponential(Psi2*t)) + 2*Psi3*(-24 + Psi4*t + 3*(8 + Psi4*t)*Exponential(Psi2*t)))*Power5(Psi3) + 12*(147*Psi4*(-1 + Exponential(Psi3*t)) - 21*Psi3*(-52 + 7*Psi4*t + 52*Exponential(Psi3*t)) + 1394*t*Power2(Psi3))*Power6(Psi2) - 280*Psi4*(-1 + Exponential(Psi2*t))*Power6(Psi3) + 3528*(-1 - Psi3*t + Exponential(Psi3*t))*Power7(Psi2)) + 2*(-(Power2(Psi3)*(63*Psi4*(-211 + 184*Exponential(Psi2*t) + 27*Exponential(Psi3*t)) - 14*Psi3*(13*(12 + Psi4*t) + 30*(-11 + 4*Psi4*t)*Exponential(Psi2*t) + 174*Exponential(Psi3*t)) + 4*t*(2922 + 35*Psi4*t)*Power2(Psi3))*Power3(Psi2)) + 7*Power2(Psi2)*(-23*Psi4*(-1 + Exponential(Psi2*t)) - Psi3*(1320 + 574*Psi4*t + 15*(-88 + 47*Psi4*t)*Exponential(Psi2*t)) + 10*t*(24 + Psi4*t)*Power2(Psi3))*Power3(Psi3) + Psi3*(21*Psi3*(624 + 330*Psi4*t + 35*(-4 + 3*Psi4*t)*Exponential(Psi2*t) - 484*Exponential(Psi3*t)) + 2457*Psi4*(-1 + Exponential(Psi3*t)) + 2*t*(7404 + 35*Psi4*t)*Power2(Psi3))*Power4(Psi2) + 7*Psi2*(1279*Psi4*(-1 + Exponential(Psi2*t)) + 10*Psi3*(24 + 17*Psi4*t + 3*(-8 + 5*Psi4*t)*Exponential(Psi2*t)))*Power4(Psi3) - 12*(-357*Psi4*(-1 + Exponential(Psi3*t)) - 7*Psi3*(-50 - 51*Psi4*t + 50*Exponential(Psi3*t)) + 106*t*Power2(Psi3))*Power5(Psi2) - 2240*Psi4*(-1 + Exponential(Psi2*t))*Power5(Psi3) + 3528*(-1 - Psi3*t + Exponential(Psi3*t))*Power6(Psi2))*Power8(Psi1) - 126*(-(Power2(Psi2)*Power2(Psi3)*(258*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(60 + 68*Psi4*t + 15*(-4 + 5*Psi4*t)*Exponential(Psi2*t)) + 40*t*Power2(Psi3))) - Psi3*(-7*Psi4*(-1 + Exponential(Psi3*t)) + Psi3*(-2*(68 + 83*Psi4*t) + (100 - 85*Psi4*t)*Exponential(Psi2*t) + 36*Exponential(Psi3*t)) + 24*t*Power2(Psi3))*Power3(Psi2) - Psi2*(-143*Psi4*(-1 + Exponential(Psi2*t)) + 10*Psi3*(4 + Psi4*t + (-4 + Psi4*t)*Exponential(Psi2*t)))*Power3(Psi3) + 4*(Psi3*(17 - 22*Psi4*t - 17*Exponential(Psi3*t)) + 22*Psi4*(-1 + Exponential(Psi3*t)) + 42*t*Power2(Psi3))*Power4(Psi2) + 20*Psi4*(-1 + Exponential(Psi2*t))*Power4(Psi3) + 104*(-1 - Psi3*t + Exponential(Psi3*t))*Power5(Psi2))*Power9(Psi1)) + 420*Power2(phi)*Power2(Psi1)*(192*Power12(Psi1)*(-(Psi2*(9*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(4 + 3*Psi4*t + (-4 + 3*Psi4*t)*Exponential(Psi2*t)))*Power2(Psi3)) - (-4 + 4*Psi3*t - 6*Psi4*t + (4 - 3*Psi4*t)*Exponential(Psi2*t))*Power2(Psi2)*Power2(Psi3) + (Psi3*(4 - 3*Psi4*t - 4*Exponential(Psi3*t)) + 3*Psi4*(-1 + Exponential(Psi3*t)) + 8*t*Power2(Psi3))*Power3(Psi2) + 6*Psi4*(-1 + Exponential(Psi2*t))*Power3(Psi3) + 4*(-1 - Psi3*t + Exponential(Psi3*t))*Power4(Psi2)) - 2*Power2(Psi2 - Psi3)*Power2(Psi3)*Power4(Psi1)*Power4(Psi2)*(Psi2*(16432*Psi4*(-1 + Exponential(Psi1*t)) + Psi3*(6*(24 + 713*Psi4*t) + (-144 + 4673*Psi4*t)*Exponential(Psi1*t)))*Power2(Psi3) + Psi3*Power2(Psi2)*(6085*Psi4*(-1 + Exponential(Psi1*t)) + Psi3*(-108 + 5743*Psi4*t + 2*(54 + 2939*Psi4*t)*Exponential(Psi1*t)) + 6*t*(-8 + 3*Psi4*t*(-24 + 29*Exponential(Psi1*t)))*Power2(Psi3)) + 7436*Psi4*(-1 + Exponential(Psi1*t))*Power3(Psi3) + 3*Power3(Psi2)*(54*Psi4*(-1 + Exponential(Psi1*t)) + 4*Psi3*(2*(6 + 55*Psi4*t) + 3*(-4 + 31*Psi4*t)*Exponential(Psi1*t)) + 8*t*(2 + Psi4*t*(-15 + 13*Exponential(Psi1*t)))*Power2(Psi3) + Psi4*Power3(Psi3)*Power3(t)) + (48 + 24*(-2*Psi3 + Psi4)*t - 45*Psi3*Psi4*Power2(t) + 6*Exponential(Psi1*t)*(-8 + 2*Psi4*t + 3*Psi3*Psi4*Power2(t)) + Psi4*Power2(Psi3)*Power3(t))*Power4(Psi2)) + Power2(Psi2 - Psi3)*Power2(Psi3)*Power3(Psi1)*(2*Psi2*(11621*Psi4*(-1 + Exponential(Psi1*t)) + Psi3*(48 + 2729*Psi4*t + (-48 + 2909*Psi4*t)*Exponential(Psi1*t)))*Power2(Psi3) + 6*Psi3*Power2(Psi2)*(812*Psi4*(-1 + Exponential(Psi1*t)) + Psi3*(-16 + 683*Psi4*t + (16 + 651*Psi4*t)*Exponential(Psi1*t)) + 2*Psi4*(-25 + 26*Exponential(Psi1*t))*Power2(Psi3)*Power2(t)) + 17902*Psi4*(-1 + Exponential(Psi1*t))*Power3(Psi3) + Power3(Psi2)*(72*Psi4*(-1 + Exponential(Psi1*t)) + 6*Psi3*(16 + 67*Psi4*t + (-16 + 49*Psi4*t)*Exponential(Psi1*t)) + 36*Psi4*(-3 + 2*Exponential(Psi1*t))*Power2(Psi3)*Power2(t) + Psi4*Power3(Psi3)*Power3(t)))*Power5(Psi2) - 6*Power11(Psi1)*(-(Power2(Psi2)*Power2(Psi3)*(1026*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(4*(29 + 30*Psi4*t) + (-116 + 211*Psi4*t)*Exponential(Psi2*t)) + 2*t*(128 + 7*Psi4*t)*Power2(Psi3))) + Psi3*(Psi3*(504 + 666*Psi4*t + (-372 + 325*Psi4*t)*Exponential(Psi2*t) - 132*Exponential(Psi3*t)) + 35*Psi4*(-1 + Exponential(Psi3*t)) + 4*t*(4 + 7*Psi4*t)*Power2(Psi3))*Power3(Psi2) - Psi2*(-331*Psi4*(-1 + Exponential(Psi2*t)) + 2*Psi3*(128 + 71*Psi4*t + (-128 + 57*Psi4*t)*Exponential(Psi2*t)))*Power3(Psi3) - 2*(-202*Psi4*(-1 + Exponential(Psi3*t)) + 2*Psi3*(-91 + 101*Psi4*t + 91*Exponential(Psi3*t)) + t*(-368 + 7*Psi4*t)*Power2(Psi3))*Power4(Psi2) + 256*Psi4*(-1 + Exponential(Psi2*t))*Power4(Psi3) + 496*(-1 - Psi3*t + Exponential(Psi3*t))*Power5(Psi2)) - 2*Psi4*Power2(Psi1)*Power2(Psi2 - Psi3)*(6*Psi2*Psi3*(-667 + 149*Psi3*t + (667 + 151*Psi3*t)*Exponential(Psi1*t)) + 5638*(-1 + Exponential(Psi1*t))*Power2(Psi3) + 3*Power2(Psi2)*(-116 + 94*Psi3*t - 7*Power2(Psi3)*Power2(t) + 2*Exponential(Psi1*t)*(58 + 41*Psi3*t + 3*Power2(Psi3)*Power2(t))))*Power3(Psi3)*Power6(Psi2) + 2*Power10(Psi1)*(-(Power2(Psi2)*(1753*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(7*(276 + 235*Psi4*t) + (-1932 + 1651*Psi4*t)*Exponential(Psi2*t)) + 42*t*(8 + Psi4*t)*Power2(Psi3))*Power3(Psi3)) - Power2(Psi3)*Power3(Psi2)*(6*Psi4*(-621 + 614*Exponential(Psi2*t) + 7*Exponential(Psi3*t)) - Psi3*(4*(306 + 305*Psi4*t) + (-1272 + 575*Psi4*t)*Exponential(Psi2*t) + 48*Exponential(Psi3*t)) + 24*t*(98 + 3*Psi4*t)*Power2(Psi3) + Psi4*Power3(Psi3)*Power3(t)) + Psi3*(3*Psi3*(876 + 765*Psi4*t + (-332 + 314*Psi4*t)*Exponential(Psi2*t) - 544*Exponential(Psi3*t)) + 447*Psi4*(-1 + Exponential(Psi3*t)) + 6*t*(608 + 45*Psi4*t)*Power2(Psi3) + 2*Psi4*Power3(Psi3)*Power3(t))*Power4(Psi2) + 2*Psi2*(1648*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(-168 + 25*Psi4*t + (168 + 67*Psi4*t)*Exponential(Psi2*t)))*Power4(Psi3) - (-1920*Psi4*(-1 + Exponential(Psi3*t)) + 480*Psi3*(-1 + 4*Psi4*t + Exponential(Psi3*t)) + 12*t*(-92 + 13*Psi4*t)*Power2(Psi3) + Psi4*Power3(Psi3)*Power3(t))*Power5(Psi2) - 184*Psi4*(-1 + Exponential(Psi2*t))*Power5(Psi3) + 2064*(-1 - Psi3*t + Exponential(Psi3*t))*Power6(Psi2)) + 6*Psi1*Psi4*(600*Psi3*(-1 + Exponential(Psi1*t)) + Psi2*(-176 + 39*Psi3*t + (176 + 37*Psi3*t)*Exponential(Psi1*t)))*Power2(Psi2 - Psi3)*Power4(Psi3)*Power7(Psi2) + Power3(Psi2)*Power5(Psi1)*(-2*Power3(Psi3)*(-(Psi4*(-6488 + 6473*Exponential(Psi1*t) + 15*Exponential(Psi3*t))) + Psi3*(-744 - 1775*Psi4*t + (636 - 2817*Psi4*t)*Exponential(Psi1*t) + (48 - 75*Psi4*t)*Exponential(Psi2*t) + 60*Exponential(Psi3*t)) + 6*t*(88 + 7*Psi4*t*(-23 + 26*Exponential(Psi1*t)))*Power2(Psi3) + 8*Psi4*Power3(Psi3)*Power3(t))*Power4(Psi2) + 2*Power3(Psi2)*(-3*Psi4*(-557 + 471*Exponential(Psi1*t) + 86*Exponential(Psi2*t)) - Psi3*(18*(6 + 505*Psi4*t) + (-24 + 10541*Psi4*t)*Exponential(Psi1*t) + 2*(-42 + 65*Psi4*t)*Exponential(Psi2*t)) - 6*t*(-80 + Psi4*t*(-59 + 106*Exponential(Psi1*t)))*Power2(Psi3) + 7*Psi4*Power3(Psi3)*Power3(t))*Power4(Psi3) - Power2(Psi3)*(-6*Psi4*(-9 + 8*Exponential(Psi1*t) + Exponential(Psi3*t)) - 2*Psi3*(-732 + 2945*Psi4*t + (624 + 2813*Psi4*t)*Exponential(Psi1*t) + 108*Exponential(Psi3*t)) - 12*t*(48 + Psi4*t*(-69 + 122*Exponential(Psi1*t)))*Power2(Psi3) + 9*Psi4*Power3(Psi3)*Power3(t))*Power5(Psi2) + 2*Power2(Psi2)*(-2*Psi4*(-6770 + 6979*Exponential(Psi1*t) - 209*Exponential(Psi2*t)) + Psi3*(-240 + 682*Psi4*t + 2*(162 + 517*Psi4*t)*Exponential(Psi1*t) + (-84 + 65*Psi4*t)*Exponential(Psi2*t)) + 24*t*(-8 + Psi4*t*(-27 + 35*Exponential(Psi1*t)))*Power2(Psi3))*Power5(Psi3) + 2*Power2(Psi3)*(288 + 48*Psi3*t + 93*Psi4*t - 96*Exponential(Psi3*t) - 246*Psi3*Psi4*Power2(t) + 3*Exponential(Psi1*t)*(-64 + 23*Psi4*t + 52*Psi3*Psi4*Power2(t)) + 5*Psi4*Power2(Psi3)*Power3(t))*Power6(Psi2) + 2*Psi2*(Psi4*(-5922 + 6143*Exponential(Psi1*t) - 221*Exponential(Psi2*t)) + Psi3*(96 + 3607*Psi4*t + 16*(-9 + 238*Psi4*t)*Exponential(Psi1*t) + (48 - 10*Psi4*t)*Exponential(Psi2*t)))*Power6(Psi3) + (24*Psi4*(-1 + Exponential(Psi3*t)) + 24*Psi3*(-4 - Psi4*t + 4*Exponential(Psi3*t)) - 24*t*(8 + Psi4*t)*Power2(Psi3) + Psi4*Power3(Psi3)*Power3(t))*Power7(Psi2) + 2*Psi4*(-2762 + 2731*Exponential(Psi1*t) + 31*Exponential(Psi2*t))*Power7(Psi3)) - 456*Psi4*(-1 + Exponential(Psi1*t))*Power2(Psi2 - Psi3)*Power5(Psi3)*Power8(Psi2) - 2*Power2(Psi2)*Power6(Psi1)*(-(Power3(Psi3)*(-(Psi4*(-3407 + 3863*Exponential(Psi1*t) - 516*Exponential(Psi2*t) + 60*Exponential(Psi3*t))) + Psi3*(-2436 + 1159*Psi4*t + 4*(546 + 107*Psi4*t)*Exponential(Psi1*t) + (12 - 34*Psi4*t)*Exponential(Psi2*t) + 240*Exponential(Psi3*t)) + 3*t*(-48 + Psi4*t*(-457 + 730*Exponential(Psi1*t)))*Power2(Psi3) + 2*Psi4*Power3(Psi3)*Power3(t))*Power4(Psi2)) + Power3(Psi2)*(-10*Psi4*(-159 + 167*Exponential(Psi1*t) - 8*Exponential(Psi2*t)) - Psi3*(16*(81 + 368*Psi4*t) + 2*(-588 + 4139*Psi4*t)*Exponential(Psi1*t) + 3*(-40 + 123*Psi4*t)*Exponential(Psi2*t)) + 6*t*(80 + Psi4*t*(-29 + 64*Exponential(Psi1*t)))*Power2(Psi3) + 8*Psi4*Power3(Psi3)*Power3(t))*Power4(Psi3) - Power2(Psi3)*(18*Psi4*(-41 + 44*Exponential(Psi1*t) - 3*Exponential(Psi3*t)) - Psi3*(-1224 + 3300*Psi4*t + (936 + 3485*Psi4*t)*Exponential(Psi1*t) + 6*(-16 + 25*Psi4*t)*Exponential(Psi2*t) + 384*Exponential(Psi3*t)) - 12*t*(-84 + Psi4*t*(-11 + 53*Exponential(Psi1*t)))*Power2(Psi3) + 17*Psi4*Power3(Psi3)*Power3(t))*Power5(Psi2) + Power2(Psi2)*(-(Psi4*(-4504 + 5081*Exponential(Psi1*t) - 577*Exponential(Psi2*t))) + Psi3*(4*(30 + 571*Psi4*t) + 6*(6 + 665*Psi4*t)*Exponential(Psi1*t) + 3*(-52 + 73*Psi4*t)*Exponential(Psi2*t)) + 3*t*(-96 + Psi4*t*(-179 + 216*Exponential(Psi1*t)))*Power2(Psi3))*Power5(Psi3) + Psi3*(Psi3*(252 + 69*Psi4*t + (84 - 18*Psi4*t)*Exponential(Psi1*t) - 336*Exponential(Psi3*t)) - 27*Psi4*(-1 + Exponential(Psi3*t)) + 6*t*(176 + 3*Psi4*t*(-26 + 29*Exponential(Psi1*t)))*Power2(Psi3) + 8*Psi4*Power3(Psi3)*Power3(t))*Power6(Psi2) + Psi2*(10*Psi4*(-333 + 368*Exponential(Psi1*t) - 35*Exponential(Psi2*t)) + Psi3*(-96 + 1490*Psi4*t + (-48 + 1249*Psi4*t)*Exponential(Psi1*t) + (144 - 34*Psi4*t)*Exponential(Psi2*t)))*Power6(Psi3) + 3*(32*Psi4*(-1 + Exponential(Psi3*t)) + 16*Psi3*(-3 - 2*Psi4*t + 3*Exponential(Psi3*t)) - 4*t*(28 + 5*Psi4*t)*Power2(Psi3) + Psi4*Power3(Psi3)*Power3(t))*Power7(Psi2) + 26*Psi4*(-1 + Exponential(Psi2*t))*Power7(Psi3) + 48*(-1 - Psi3*t + Exponential(Psi3*t))*Power8(Psi2)) - 2*Power8(Psi1)*(Power3(Psi3)*(Psi3*(-756 + 1097*Psi4*t - 20*(-15 + 82*Psi4*t)*Exponential(Psi1*t) + (696 - 533*Psi4*t)*Exponential(Psi2*t) - 240*Exponential(Psi3*t)) + Psi4*(-4739 + 1716*Exponential(Psi1*t) + 2963*Exponential(Psi2*t) + 60*Exponential(Psi3*t)) - 3*t*(-928 + Psi4*t*(-143 + 40*Exponential(Psi1*t)))*Power2(Psi3) + 7*Psi4*Power3(Psi3)*Power3(t))*Power4(Psi2) + Power3(Psi2)*(-6*Psi4*(-210 + 143*Exponential(Psi1*t) + 67*Exponential(Psi2*t)) + Psi3*(936 + 552*Psi4*t + (384 + 427*Psi4*t)*Exponential(Psi1*t) + 11*(-120 + 109*Psi4*t)*Exponential(Psi2*t)) + 12*t*(-20 + Psi4*t*(-19 + 32*Exponential(Psi1*t)))*Power2(Psi3) + Psi4*Power3(Psi3)*Power3(t))*Power4(Psi3) - Power2(Psi3)*(Psi3*(8*(-132 + 203*Psi4*t) + (1752 - 1999*Psi4*t)*Exponential(Psi1*t) + (-552 + 571*Psi4*t)*Exponential(Psi2*t) - 144*Exponential(Psi3*t)) + 6*Psi4*(-225 + 143*Exponential(Psi1*t) + 126*Exponential(Psi2*t) - 44*Exponential(Psi3*t)) + 6*t*(376 + Psi4*t*(-7 + 152*Exponential(Psi1*t)))*Power2(Psi3) + 9*Psi4*Power3(Psi3)*Power3(t))*Power5(Psi2) - Power2(Psi2)*(2243*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(204 + 461*Psi4*t + 3*(-68 + 145*Psi4*t)*Exponential(Psi2*t)) + 6*t*(8 + 7*Psi4*t)*Power2(Psi3))*Power5(Psi3) - Psi3*(6*Psi3*(396 - 47*Psi4*t + (-178 + 131*Psi4*t)*Exponential(Psi1*t) + (30 - 49*Psi4*t)*Exponential(Psi2*t) - 248*Exponential(Psi3*t)) + 582*Psi4*(-1 + Exponential(Psi3*t)) - 3*t*(-912 + Psi4*t*(-83 + 216*Exponential(Psi1*t)))*Power2(Psi3) + 7*Psi4*Power3(Psi3)*Power3(t))*Power6(Psi2) + 2*Psi2*(448*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(-24 - 19*Psi4*t + (24 + 23*Psi4*t)*Exponential(Psi2*t)))*Power6(Psi3) + 8*(-24*Psi4*(-1 + Exponential(Psi3*t)) - 6*Psi3*(-31 - 4*Psi4*t + 31*Exponential(Psi3*t)) + 6*t*(54 + Psi4*t)*Power2(Psi3) + Psi4*Power3(Psi3)*Power3(t))*Power7(Psi2) - 8*Psi4*(-1 + Exponential(Psi2*t))*Power7(Psi3) + 96*(-1 - Psi3*t + Exponential(Psi3*t))*Power8(Psi2)) + Psi2*Power7(Psi1)*(2*Power3(Psi3)*(Psi4*(-2027 + 3031*Exponential(Psi1*t) - 1094*Exponential(Psi2*t) + 90*Exponential(Psi3*t)) - Psi3*(-1152 + 1273*Psi4*t + 3*(472 + 769*Psi4*t)*Exponential(Psi1*t) + (-624 + 969*Psi4*t)*Exponential(Psi2*t) + 360*Exponential(Psi3*t)) - 6*t*(-336 + Psi4*t*(-163 + 260*Exponential(Psi1*t)))*Power2(Psi3) + 7*Psi4*Power3(Psi3)*Power3(t))*Power4(Psi2) + Power3(Psi2)*(-2*Psi4*(1371 + 1418*Exponential(Psi1*t) - 2789*Exponential(Psi2*t)) - 2*Psi3*(744 + 880*Psi4*t + 36*(-34 + 59*Psi4*t)*Exponential(Psi1*t) + (480 - 553*Psi4*t)*Exponential(Psi2*t)) + 12*Psi4*(-69 + 152*Exponential(Psi1*t))*Power2(Psi3)*Power2(t) + 9*Psi4*Power3(Psi3)*Power3(t))*Power4(Psi3) - Power2(Psi3)*(12*Psi4*(-270 + 258*Exponential(Psi1*t) + 43*Exponential(Psi2*t) - 31*Exponential(Psi3*t)) - 2*Psi3*(516 + 1363*Psi4*t + (-696 + 2950*Psi4*t)*Exponential(Psi1*t) + (-276 + 458*Psi4*t)*Exponential(Psi2*t) + 456*Exponential(Psi3*t)) + 96*t*(68 + Psi4*t*(-3 + 8*Exponential(Psi1*t)))*Power2(Psi3) + 41*Psi4*Power3(Psi3)*Power3(t))*Power5(Psi2) + 2*Power2(Psi2)*(2695*Psi4 + Psi3*(-192*Psi3 + 691*Psi4)*t + (-2630*Psi4 + Psi3*(36 - 139*Psi4*t))*Exponential(Psi2*t) - 234*Psi4*Power2(Psi3)*Power2(t) + Exponential(Psi1*t)*(-65*Psi4 + Psi3*(-36 + 2009*Psi4*t) + 192*Psi4*Power2(Psi3)*Power2(t)))*Power5(Psi3) + 2*Psi3*(-213*Psi4*(-1 + Exponential(Psi3*t)) + Psi3*(-888 + 132*Psi4*t - 132*(-7 + 4*Psi4*t)*Exponential(Psi1*t) + (-48 + 75*Psi4*t)*Exponential(Psi2*t) + 12*Exponential(Psi3*t)) + 12*t*(100 + Psi4*t*(-33 + 70*Exponential(Psi1*t)))*Power2(Psi3) + 2*Psi4*Power3(Psi3)*Power3(t))*Power6(Psi2) + 2*Psi2*(1119*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(9*(-16 + 19*Psi4*t) + 2*(72 + 11*Psi4*t)*Exponential(Psi2*t)))*Power6(Psi3) + 2*(204*Psi4*(-1 + Exponential(Psi3*t)) - 12*Psi3*(-29 + 17*Psi4*t + 29*Exponential(Psi3*t)) + 6*t*(80 - 13*Psi4*t)*Power2(Psi3) + 7*Psi4*Power3(Psi3)*Power3(t))*Power7(Psi2) - 386*Psi4*(-1 + Exponential(Psi2*t))*Power7(Psi3) + 480*(-1 - Psi3*t + Exponential(Psi3*t))*Power8(Psi2)) + (2*Power3(Psi2)*Power3(Psi3)*(Psi3*(744 + 2417*Psi4*t + (372 - 333*Psi4*t)*Exponential(Psi1*t) + 24*(-44 + 71*Psi4*t)*Exponential(Psi2*t) - 60*Exponential(Psi3*t)) + Psi4*(-5546 + 5531*Exponential(Psi2*t) + 15*Exponential(Psi3*t)) + 6*t*(264 + Psi4*t*(27 + 32*Exponential(Psi1*t)))*Power2(Psi3) + 2*Psi4*Power3(Psi3)*Power3(t)) + Power2(Psi3)*(2*Psi3*(-3*(380 + 951*Psi4*t) + (-744 + 666*Psi4*t)*Exponential(Psi1*t) + (1968 - 1799*Psi4*t)*Exponential(Psi2*t) - 84*Exponential(Psi3*t)) + 18*Psi4*(-127 + 108*Exponential(Psi2*t) + 19*Exponential(Psi3*t)) - 12*t*(-240 + Psi4*t*(1 + 64*Exponential(Psi1*t)))*Power2(Psi3) + Psi4*Power3(Psi3)*Power3(t))*Power4(Psi2) - 2*Power2(Psi2)*(4661*Psi4*(-1 + Exponential(Psi2*t)) - Psi3*(1140 + 77*Psi4*t + 3*(-380 + 91*Psi4*t)*Exponential(Psi2*t)) + 6*t*(16 + 7*Psi4*t)*Power2(Psi3))*Power4(Psi3) - 2*Psi3*(3*Psi3*(992 + 241*Psi4*t + (-124 + 111*Psi4*t)*Exponential(Psi1*t) + 4*(-11 + 3*Psi4*t)*Exponential(Psi2*t) - 824*Exponential(Psi3*t)) + 738*Psi4*(-1 + Exponential(Psi3*t)) - 6*t*(-1032 + Psi4*t*(-51 + 32*Exponential(Psi1*t)))*Power2(Psi3) + 7*Psi4*Power3(Psi3)*Power3(t))*Power5(Psi2) - 4*Psi2*(175*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(48 + 113*Psi4*t + (-48 + 71*Psi4*t)*Exponential(Psi2*t)))*Power5(Psi3) + 3*(Psi3*(848 + 872*Psi4*t - 848*Exponential(Psi3*t)) - 872*Psi4*(-1 + Exponential(Psi3*t)) + 64*t*(23 + 2*Psi4*t)*Power2(Psi3) + 3*Psi4*Power3(Psi3)*Power3(t))*Power6(Psi2) + 736*Psi4*(-1 + Exponential(Psi2*t))*Power6(Psi3) - 2112*(-1 - Psi3*t + Exponential(Psi3*t))*Power7(Psi2))*Power9(Psi1)) - 168*phi*Power3(Psi1)*(240*Power12(Psi1)*(-(Psi2*(9*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(4 + 3*Psi4*t + (-4 + 3*Psi4*t)*Exponential(Psi2*t)))*Power2(Psi3)) - (-4 + 4*Psi3*t - 6*Psi4*t + (4 - 3*Psi4*t)*Exponential(Psi2*t))*Power2(Psi2)*Power2(Psi3) + (Psi3*(4 - 3*Psi4*t - 4*Exponential(Psi3*t)) + 3*Psi4*(-1 + Exponential(Psi3*t)) + 8*t*Power2(Psi3))*Power3(Psi2) + 6*Psi4*(-1 + Exponential(Psi2*t))*Power3(Psi3) + 4*(-1 - Psi3*t + Exponential(Psi3*t))*Power4(Psi2)) + Power2(Psi2 - Psi3)*Power2(Psi3)*Power3(Psi1)*(Psi2*(28262*Psi4*(-1 + Exponential(Psi1*t)) + Psi3*(1008 + 4559*Psi4*t + 8*(-126 + 845*Psi4*t)*Exponential(Psi1*t)))*Power2(Psi3) + Psi3*Power2(Psi2)*(8191*Psi4*(-1 + Exponential(Psi1*t)) + Psi3*(-868 + 4669*Psi4*t + (868 + 5630*Psi4*t)*Exponential(Psi1*t)) + 2*t*(-96 + Psi4*t*(-73 + 240*Exponential(Psi1*t)))*Power2(Psi3)) + 4*(48*Psi4*(-1 + Exponential(Psi1*t)) + 4*Psi3*(6 + 41*Psi4*t + (-6 + 40*Psi4*t)*Exponential(Psi1*t)) + t*(48 + Psi4*t*(-13 + 30*Exponential(Psi1*t)))*Power2(Psi3))*Power3(Psi2) + 17950*Psi4*(-1 + Exponential(Psi1*t))*Power3(Psi3))*Power5(Psi2) - 6*Power11(Psi1)*(-(Power2(Psi2)*Power2(Psi3)*(1254*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(140 + 154*Psi4*t + 5*(-28 + 47*Psi4*t)*Exponential(Psi2*t)) + 5*t*(64 + Psi4*t)*Power2(Psi3))) + Psi3*(Psi3*(648 + 803*Psi4*t + 10*(-46 + 39*Psi4*t)*Exponential(Psi2*t) - 188*Exponential(Psi3*t)) + 61*Psi4*(-1 + Exponential(Psi3*t)) + 2*t*(24 + 5*Psi4*t)*Power2(Psi3))*Power3(Psi2) - Psi2*(-389*Psi4*(-1 + Exponential(Psi2*t)) + 5*Psi3*(64 + 33*Psi4*t + (-64 + 31*Psi4*t)*Exponential(Psi2*t)))*Power3(Psi3) + (Psi3*(404 - 484*Psi4*t - 404*Exponential(Psi3*t)) + 484*Psi4*(-1 + Exponential(Psi3*t)) + t*(864 - 5*Psi4*t)*Power2(Psi3))*Power4(Psi2) + 320*Psi4*(-1 + Exponential(Psi2*t))*Power4(Psi3) + 592*(-1 - Psi3*t + Exponential(Psi3*t))*Power5(Psi2)) - Power2(Psi1)*Power2(Psi2 - Psi3)*(Psi2*Psi3*(10299*Psi4*(-1 + Exponential(Psi1*t)) + 2*Psi3*(96 + 751*Psi4*t + (-96 + 1085*Psi4*t)*Exponential(Psi1*t))) + 11319*Psi4*(-1 + Exponential(Psi1*t))*Power2(Psi3) + 4*Power2(Psi2)*(324*Psi4*(-1 + Exponential(Psi1*t)) + 2*Psi3*(-24 + 83*Psi4*t + 4*(6 + 25*Psi4*t)*Exponential(Psi1*t)) + 5*Psi4*(-1 + 3*Exponential(Psi1*t))*Power2(Psi3)*Power2(t)))*Power3(Psi3)*Power6(Psi2) + Power10(Psi1)*(-(Power2(Psi3)*(6*Psi4*(-1429 + 1360*Exponential(Psi2*t) + 69*Exponential(Psi3*t)) - Psi3*(4*(805 + 904*Psi4*t) + 15*(-244 + 129*Psi4*t)*Exponential(Psi2*t) + 440*Exponential(Psi3*t)) + 8*t*(828 + 11*Psi4*t)*Power2(Psi3))*Power3(Psi2)) - Power2(Psi2)*(5137*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(5400 + 4157*Psi4*t + 15*(-360 + 269*Psi4*t)*Exponential(Psi2*t)) + 10*t*(48 + Psi4*t)*Power2(Psi3))*Power3(Psi3) + Psi3*(Psi3*(7360 + 4479*Psi4*t + 30*(-74 + 63*Psi4*t)*Exponential(Psi2*t) - 5140*Exponential(Psi3*t)) + 1791*Psi4*(-1 + Exponential(Psi3*t)) + 2*t*(5184 + 103*Psi4*t)*Power2(Psi3))*Power4(Psi2) + 2*Psi2*(4096*Psi4*(-1 + Exponential(Psi2*t)) + 5*Psi3*(-48 + 19*Psi4*t + 3*(16 + 7*Psi4*t)*Exponential(Psi2*t)))*Power4(Psi3) - 4*(Psi3*(95 + 1032*Psi4*t - 95*Exponential(Psi3*t)) - 1032*Psi4*(-1 + Exponential(Psi3*t)) + 3*t*(-88 + 9*Psi4*t)*Power2(Psi3))*Power5(Psi2) - 400*Psi4*(-1 + Exponential(Psi2*t))*Power5(Psi3) + 4320*(-1 - Psi3*t + Exponential(Psi3*t))*Power6(Psi2)) - Psi3*Power4(Psi1)*Power4(Psi2)*(-((-(Psi4*(-191 + 333*Exponential(Psi1*t) - 156*Exponential(Psi2*t) + 14*Exponential(Psi3*t))) + Psi3*(-4100 + 14352*Psi4*t + 4*(1056 + 4115*Psi4*t)*Exponential(Psi1*t) - 45*(4 + Psi4*t)*Exponential(Psi2*t) + 56*Exponential(Psi3*t)) + 12*t*(-280 + Psi4*t*(-43 + 170*Exponential(Psi1*t)))*Power2(Psi3))*Power3(Psi2)*Power3(Psi3)) - Power2(Psi3)*(-5*Psi3*(-176 + 903*Psi4*t + 102*(2 + 15*Psi4*t)*Exponential(Psi1*t) - 28*Exponential(Psi3*t)) - Psi4*(-17674 + 17611*Exponential(Psi1*t) + 63*Exponential(Psi3*t)) + 2*t*(1440 + Psi4*t*(-113 + 180*Exponential(Psi1*t)))*Power2(Psi3))*Power4(Psi2) + Power2(Psi2)*(-2*Psi4*(-20687 + 20872*Exponential(Psi1*t) - 185*Exponential(Psi2*t)) - Psi3*(4456 + 1505*Psi4*t + (-4876 + 5790*Psi4*t)*Exponential(Psi1*t) + 60*(7 + Psi4*t)*Exponential(Psi2*t)) + 2*t*(-624 + Psi4*t*(-217 + 750*Exponential(Psi1*t)))*Power2(Psi3))*Power4(Psi3) + Psi3*(Psi4*(-845 + 844*Exponential(Psi1*t) + Exponential(Psi3*t)) + Psi3*(-876 + 4016*Psi4*t + (488 + 3790*Psi4*t)*Exponential(Psi1*t) + 388*Exponential(Psi3*t)) + 24*t*(28 + Psi4*t*(-11 + 35*Exponential(Psi1*t)))*Power2(Psi3))*Power5(Psi2) + Psi2*(Psi4*(-7542 + 7825*Exponential(Psi1*t) - 283*Exponential(Psi2*t)) + Psi3*(1632 + 7206*Psi4*t + 2*(-936 + 5345*Psi4*t)*Exponential(Psi1*t) + 15*(16 + Psi4*t)*Exponential(Psi2*t)))*Power5(Psi3) + 4*(6*Psi3*(5*(4 + Psi4*t) + (-12 + 5*Psi4*t)*Exponential(Psi1*t) - 8*Exponential(Psi3*t)) - 12*Psi4*(-1 + Exponential(Psi3*t)) + t*(24 + Psi4*t*(-11 + 15*Exponential(Psi1*t)))*Power2(Psi3))*Power6(Psi2) + Psi4*(-15170 + 15131*Exponential(Psi1*t) + 39*Exponential(Psi2*t))*Power6(Psi3)) + 8*Psi1*Psi4*(459*Psi3*(-1 + Exponential(Psi1*t)) + Psi2*(-183 + 25*Psi3*t + (183 + 35*Psi3*t)*Exponential(Psi1*t)))*Power2(Psi2 - Psi3)*Power4(Psi3)*Power7(Psi2) + Power3(Psi2)*Power5(Psi1)*(-((-(Psi4*(-18403 + 18269*Exponential(Psi1*t) - 312*Exponential(Psi2*t) + 446*Exponential(Psi3*t))) + Psi3*(-4620 + 2893*Psi4*t + (3456 - 3480*Psi4*t)*Exponential(Psi1*t) - 60*(1 + 9*Psi4*t)*Exponential(Psi2*t) + 1224*Exponential(Psi3*t)) + 2*t*(1680 + Psi4*t*(-487 + 1620*Exponential(Psi1*t)))*Power2(Psi3))*Power3(Psi3)*Power4(Psi2)) - (2*Psi4*(-7153 + 6561*Exponential(Psi1*t) + 620*Exponential(Psi2*t) - 28*Exponential(Psi3*t)) + Psi3*(-360 + 18645*Psi4*t + 8*(62 + 3175*Psi4*t)*Exponential(Psi1*t) + 360*(-1 + 2*Psi4*t)*Exponential(Psi2*t) + 224*Exponential(Psi3*t)) + 2*t*(-3360 + Psi4*t*(-197 + 780*Exponential(Psi1*t)))*Power2(Psi3))*Power3(Psi2)*Power4(Psi3) + Power2(Psi3)*(8*Psi4*(-130 + 107*Exponential(Psi1*t) + 23*Exponential(Psi3*t)) + Psi3*(-6000 + 9479*Psi4*t + 4*(876 + 2065*Psi4*t)*Exponential(Psi1*t) + 2496*Exponential(Psi3*t)) + 6*t*(-256 + Psi4*t*(-73 + 340*Exponential(Psi1*t)))*Power2(Psi3))*Power5(Psi2) + Power2(Psi2)*(-(Psi4*(-23354 + 26319*Exponential(Psi1*t) - 2965*Exponential(Psi2*t))) + Psi3*(8*(-183 + 650*Psi4*t) + 6*(514 + 745*Psi4*t)*Exponential(Psi1*t) + 15*(-108 + 13*Psi4*t)*Exponential(Psi2*t)) + 4*t*(-768 + Psi4*t*(-169 + 570*Exponential(Psi1*t)))*Power2(Psi3))*Power5(Psi3) + Psi3*(Psi3*(2292 + 657*Psi4*t + 2*(-574 + 285*Psi4*t)*Exponential(Psi1*t) - 1144*Exponential(Psi3*t)) - 383*Psi4*(-1 + Exponential(Psi3*t)) + 2*t*(816 + Psi4*t*(-121 + 240*Exponential(Psi1*t)))*Power2(Psi3))*Power6(Psi2) + Psi2*(8*Psi4*(-1576 + 1831*Exponential(Psi1*t) - 255*Exponential(Psi2*t)) + Psi3*(288 + 6274*Psi4*t + 4*(-372 + 2155*Psi4*t)*Exponential(Psi1*t) - 15*(-80 + Psi4*t)*Exponential(Psi2*t)))*Power6(Psi3) - 12*(Psi3*(8 + 6*Psi4*t - 8*Exponential(Psi3*t)) - 6*Psi4*(-1 + Exponential(Psi3*t)) + t*(32 + Psi4*t)*Power2(Psi3))*Power7(Psi2) + 4*Psi4*(-1480 + 1417*Exponential(Psi1*t) + 63*Exponential(Psi2*t))*Power7(Psi3)) - 480*Psi4*(-1 + Exponential(Psi1*t))*Power2(Psi2 - Psi3)*Power5(Psi3)*Power8(Psi2) + Power8(Psi1)*((-(Psi4*(-11536 + 4048*Exponential(Psi1*t) + 6649*Exponential(Psi2*t) + 839*Exponential(Psi3*t))) + Psi3*(-960 - 217*Psi4*t + (-276 + 4010*Psi4*t)*Exponential(Psi1*t) + 15*(-104 + 203*Psi4*t)*Exponential(Psi2*t) + 2796*Exponential(Psi3*t)) + 2*t*(-5952 + Psi4*t*(-227 + 120*Exponential(Psi1*t)))*Power2(Psi3))*Power3(Psi3)*Power4(Psi2) - (2*Psi3*(6*(395 + 246*Psi4*t) + 112*(4 + 5*Psi4*t)*Exponential(Psi1*t) + 45*(-62 + 33*Psi4*t)*Exponential(Psi2*t) - 28*Exponential(Psi3*t)) - Psi4*(-59 + 2024*Exponential(Psi1*t) - 1951*Exponential(Psi2*t) - 14*Exponential(Psi3*t)) + 12*t*(-232 + Psi4*t*(-27 + 80*Exponential(Psi1*t)))*Power2(Psi3))*Power3(Psi2)*Power4(Psi3) + Power2(Psi3)*(Psi4*(-3439 + 2024*Exponential(Psi1*t) + 4620*Exponential(Psi2*t) - 3205*Exponential(Psi3*t)) + Psi3*(324 + 6484*Psi4*t + (3240 - 4660*Psi4*t)*Exponential(Psi1*t) + 30*(-68 + 21*Psi4*t)*Exponential(Psi2*t) - 1524*Exponential(Psi3*t)) + 24*t*(332 + Psi4*t*(-11 + 100*Exponential(Psi1*t)))*Power2(Psi3))*Power5(Psi2) + Power2(Psi2)*(6824*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(2520 + 989*Psi4*t + 15*(-168 + 61*Psi4*t)*Exponential(Psi2*t)) + 10*t*(48 + 5*Psi4*t)*Power2(Psi3))*Power5(Psi3) - 2*Psi3*(-2119*Psi4*(-1 + Exponential(Psi3*t)) + Psi3*(-3816 + 2757*Psi4*t + (1034 - 885*Psi4*t)*Exponential(Psi1*t) + 255*(-2 + 3*Psi4*t)*Exponential(Psi2*t) + 3292*Exponential(Psi3*t)) + 4*t*(-1020 + Psi4*t*(-43 + 210*Exponential(Psi1*t)))*Power2(Psi3))*Power6(Psi2) - 2*Psi2*(952*Psi4*(-1 + Exponential(Psi2*t)) + 5*Psi3*(-48 - Psi4*t + (48 + 9*Psi4*t)*Exponential(Psi2*t)))*Power6(Psi3) - 24*(Psi3*(299 - 50*Psi4*t - 299*Exponential(Psi3*t)) + 50*Psi4*(-1 + Exponential(Psi3*t)) + 392*t*Power2(Psi3))*Power7(Psi2) + 80*Psi4*(-1 + Exponential(Psi2*t))*Power7(Psi3) - 1920*(-1 - Psi3*t + Exponential(Psi3*t))*Power8(Psi2)) + Power2(Psi2)*Power6(Psi1)*((-(Psi4*(-7725 + 10161*Exponential(Psi1*t) - 3590*Exponential(Psi2*t) + 1154*Exponential(Psi3*t))) + Psi3*(5*(-2104 + 2525*Psi4*t) + 16*(499 + 255*Psi4*t)*Exponential(Psi1*t) - 30*(32 + 9*Psi4*t)*Exponential(Psi2*t) + 3496*Exponential(Psi3*t)) + 6*t*(-720 + Psi4*t*(-251 + 990*Exponential(Psi1*t)))*Power2(Psi3))*Power3(Psi3)*Power4(Psi2) - (-(Psi4*(-5967 + 8210*Exponential(Psi1*t) - 2159*Exponential(Psi2*t) - 84*Exponential(Psi3*t))) - Psi3*(40*(61 + 215*Psi4*t) + 2*(-2048 + 9175*Psi4*t)*Exponential(Psi1*t) + 15*(88 + 127*Psi4*t)*Exponential(Psi2*t) + 336*Exponential(Psi3*t)) + 8*t*(600 + Psi4*t*(-23 + 150*Exponential(Psi1*t)))*Power2(Psi3))*Power3(Psi2)*Power4(Psi3) - Power2(Psi3)*(-12*Psi4*(-31 + 113*Exponential(Psi1*t) + 13*Exponential(Psi2*t) - 95*Exponential(Psi3*t)) + Psi3*(4*(-2575 + 2691*Psi4*t) + (5056 + 9010*Psi4*t)*Exponential(Psi1*t) + 105*(-4 + 9*Psi4*t)*Exponential(Psi2*t) + 5664*Exponential(Psi3*t)) + 8*t*(-1260 + Psi4*t*(-23 + 195*Exponential(Psi1*t)))*Power2(Psi3))*Power5(Psi2) - Power2(Psi2)*(-(Psi4*(-6187 + 8639*Exponential(Psi1*t) - 2452*Exponential(Psi2*t))) + Psi3*(1184 + 6370*Psi4*t + 2*(98 + 5025*Psi4*t)*Exponential(Psi1*t) + 15*(-92 + 53*Psi4*t)*Exponential(Psi2*t)) + 8*t*(-456 + Psi4*t*(-73 + 210*Exponential(Psi1*t)))*Power2(Psi3))*Power5(Psi3) - Psi3*(Psi3*(2960 + 1821*Psi4*t + (-932 + 570*Psi4*t)*Exponential(Psi1*t) - 2028*Exponential(Psi3*t)) - 1379*Psi4*(-1 + Exponential(Psi3*t)) + 2*t*(2976 + Psi4*t*(-247 + 750*Exponential(Psi1*t)))*Power2(Psi3))*Power6(Psi2) - Psi2*(Psi4*(-5477 + 8044*Exponential(Psi1*t) - 2567*Exponential(Psi2*t)) + Psi3*(-1728 + 2846*Psi4*t + 16*(-27 + 175*Psi4*t)*Exponential(Psi1*t) - 15*(-144 + 7*Psi4*t)*Exponential(Psi2*t)))*Power6(Psi3) + 4*(-144*Psi4*(-1 + Exponential(Psi3*t)) + Psi3*(-23 + 144*Psi4*t + 23*Exponential(Psi3*t)) + 3*t*(88 + 5*Psi4*t)*Power2(Psi3))*Power7(Psi2) - 127*Psi4*(-1 + Exponential(Psi2*t))*Power7(Psi3) - 288*(-1 - Psi3*t + Exponential(Psi3*t))*Power8(Psi2)) + Psi2*Power7(Psi1)*(-((-(Psi4*(-3107 + 7400*Exponential(Psi1*t) - 5709*Exponential(Psi2*t) + 1416*Exponential(Psi3*t))) + Psi3*(-6548 + 10591*Psi4*t + 6*(814 + 1115*Psi4*t)*Exponential(Psi1*t) + 15*(-192 + 235*Psi4*t)*Exponential(Psi2*t) + 4544*Exponential(Psi3*t)) + 2*t*(-6720 + Psi4*t*(-577 + 1980*Exponential(Psi1*t)))*Power2(Psi3))*Power3(Psi3)*Power4(Psi2)) + (Psi3*(1496 + 2259*Psi4*t - 36*(-98 + 115*Psi4*t)*Exponential(Psi1*t) + 15*(-320 + 43*Psi4*t)*Exponential(Psi2*t) - 224*Exponential(Psi3*t)) - 4*Psi4*(1797 + 1012*Exponential(Psi1*t) - 2795*Exponential(Psi2*t) - 14*Exponential(Psi3*t)) + 12*t*(-80 + Psi4*t*(-43 + 200*Exponential(Psi1*t)))*Power2(Psi3))*Power3(Psi2)*Power4(Psi3) - 2*Power2(Psi3)*(2*Psi4*(-706 + 896*Exponential(Psi1*t) + 495*Exponential(Psi2*t) - 685*Exponential(Psi3*t)) - Psi3*(5*(-508 + 367*Psi4*t) + 4*(58 + 875*Psi4*t)*Exponential(Psi1*t) + 30*(-14 + 41*Psi4*t)*Exponential(Psi2*t) + 2728*Exponential(Psi3*t)) + t*(8160 + Psi4*t*(-107 + 600*Exponential(Psi1*t)))*Power2(Psi3))*Power5(Psi2) + Power2(Psi2)*(Psi4*(7227 + 232*Exponential(Psi1*t) - 7459*Exponential(Psi2*t)) + 5*Psi3*(-212 + 331*Psi4*t + (-64 + 940*Psi4*t)*Exponential(Psi1*t) + (276 - 9*Psi4*t)*Exponential(Psi2*t)) + 2*t*(-1056 + Psi4*t*(-133 + 240*Exponential(Psi1*t)))*Power2(Psi3))*Power5(Psi3) + 2*Psi3*(-1506*Psi4*(-1 + Exponential(Psi3*t)) + Psi3*(-764 + 2028*Psi4*t + (606 - 435*Psi4*t)*Exponential(Psi1*t) + 75*(-2 + 3*Psi4*t)*Exponential(Psi2*t) + 308*Exponential(Psi3*t)) + 4*t*(660 + Psi4*t*(-62 + 285*Exponential(Psi1*t)))*Power2(Psi3))*Power6(Psi2) + Psi2*(1734*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(7*(-240 + 73*Psi4*t) + 15*(112 + Psi4*t)*Exponential(Psi2*t)))*Power6(Psi3) - 2*(-780*Psi4*(-1 + Exponential(Psi3*t)) + 4*Psi3*(-343 + 195*Psi4*t + 343*Exponential(Psi3*t)) + 3*t*(-352 + 15*Psi4*t)*Power2(Psi3))*Power7(Psi2) - 526*Psi4*(-1 + Exponential(Psi2*t))*Power7(Psi3) + 1440*(-1 - Psi3*t + Exponential(Psi3*t))*Power8(Psi2)) + ((Psi3*(3780 + 6065*Psi4*t + (784 - 780*Psi4*t)*Exponential(Psi1*t) + 195*(-20 + 17*Psi4*t)*Exponential(Psi2*t) - 664*Exponential(Psi3*t)) + Psi4*(-14945 + 14751*Exponential(Psi2*t) + 194*Exponential(Psi3*t)) + 30*t*(144 + Psi4*t*(3 + 16*Exponential(Psi1*t)))*Power2(Psi3))*Power3(Psi2)*Power3(Psi3) - Power2(Psi3)*(-1836*Psi4*(-1 + Exponential(Psi3*t)) + Psi3*(3616 + 9329*Psi4*t - 8*(-196 + 195*Psi4*t)*Exponential(Psi1*t) + 30*(-200 + 159*Psi4*t)*Exponential(Psi2*t) + 816*Exponential(Psi3*t)) + 6*t*(-832 + 5*Psi4*t*(-5 + 32*Exponential(Psi1*t)))*Power2(Psi3))*Power4(Psi2) - Power2(Psi2)*(10818*Psi4*(-1 + Exponential(Psi2*t)) - Psi3*(2760 + 293*Psi4*t + 15*(-184 + 65*Psi4*t)*Exponential(Psi2*t)) + 10*t*(96 + 7*Psi4*t)*Power2(Psi3))*Power4(Psi3) + Psi3*(-3695*Psi4*(-1 + Exponential(Psi3*t)) + Psi3*(-9388 + 1641*Psi4*t + (784 - 780*Psi4*t)*Exponential(Psi1*t) + 30*(-10 + 27*Psi4*t)*Exponential(Psi2*t) + 8904*Exponential(Psi3*t)) + 2*t*(-9072 + 5*Psi4*t*(-29 + 48*Exponential(Psi1*t)))*Power2(Psi3))*Power5(Psi2) - 2*Psi2*(634*Psi4*(-1 + Exponential(Psi2*t)) + 5*Psi3*(96 + 47*Psi4*t + (-96 + 33*Psi4*t)*Exponential(Psi2*t)))*Power5(Psi3) + 8*(Psi3*(808 + 225*Psi4*t - 808*Exponential(Psi3*t)) - 225*Psi4*(-1 + Exponential(Psi3*t)) + 3*t*(368 + 5*Psi4*t)*Power2(Psi3))*Power6(Psi2) + 800*Psi4*(-1 + Exponential(Psi2*t))*Power6(Psi3) - 960*(-1 - Psi3*t + Exponential(Psi3*t))*Power7(Psi2))*Power9(Psi1)) - 2520*Psi1*Power3(phi)*(24*Power12(Psi1)*(-(Psi2*(9*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(4 + 3*Psi4*t + (-4 + 3*Psi4*t)*Exponential(Psi2*t)))*Power2(Psi3)) - (-4 + 4*Psi3*t - 6*Psi4*t + (4 - 3*Psi4*t)*Exponential(Psi2*t))*Power2(Psi2)*Power2(Psi3) + (Psi3*(4 - 3*Psi4*t - 4*Exponential(Psi3*t)) + 3*Psi4*(-1 + Exponential(Psi3*t)) + 8*t*Power2(Psi3))*Power3(Psi2) + 6*Psi4*(-1 + Exponential(Psi2*t))*Power3(Psi3) + 4*(-1 - Psi3*t + Exponential(Psi3*t))*Power4(Psi2)) - Psi4*Power2(Psi2 - Psi3)*Power3(Psi3)*Power4(Psi1)*Power4(Psi2)*(6*Psi2*Psi3*(-696 + 215*Psi3*t + (696 + 185*Psi3*t)*Exponential(Psi1*t)) + 1980*(-1 + Exponential(Psi1*t))*Power2(Psi3) + 2*Power2(Psi2)*(-636 + 791*Psi3*t - 90*Power2(Psi3)*Power2(t) + Exponential(Psi1*t)*(636 + 649*Psi3*t + 50*Power2(Psi3)*Power2(t))) + 2*t*(138 - 64*Psi3*t + 2*(51 + 14*Psi3*t)*Exponential(Psi1*t) + 3*Power2(Psi3)*Power2(t))*Power3(Psi2) + (-9 + 2*Psi3*t + 3*Exponential(Psi1*t))*Power2(t)*Power4(Psi2)) + Psi4*Power2(Psi2 - Psi3)*Power3(Psi1)*(4*Psi2*Psi3*(-720 + 209*Psi3*t + (720 + 169*Psi3*t)*Exponential(Psi1*t)) + 2400*(-1 + Exponential(Psi1*t))*Power2(Psi3) + 4*Power2(Psi2)*(Exponential(Psi1*t)*(120 + 102*Psi3*t + 7*Power2(Psi3)*Power2(t)) - 2*(60 - 69*Psi3*t + 8*Power2(Psi3)*Power2(t))) + t*(36 - 18*Psi3*t + 6*(4 + Psi3*t)*Exponential(Psi1*t) + Power2(Psi3)*Power2(t))*Power3(Psi2))*Power3(Psi3)*Power5(Psi2) - 6*Power11(Psi1)*(-(Power2(Psi2)*Power2(Psi3)*(132*Psi4*(-1 + Exponential(Psi2*t)) + 2*Psi3*(8 + 7*Psi4*t + (-8 + 15*Psi4*t)*Exponential(Psi2*t)) + t*(32 + 3*Psi4*t)*Power2(Psi3))) + Psi3*(Psi3*(64 + 85*Psi4*t + (-48 + 43*Psi4*t)*Exponential(Psi2*t) - 16*Exponential(Psi3*t)) + 4*Psi4*(-1 + Exponential(Psi3*t)) + 6*Psi4*Power2(Psi3)*Power2(t))*Power3(Psi2) - Psi2*(-44*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(32 + 19*Psi4*t + (-32 + 13*Psi4*t)*Exponential(Psi2*t)))*Power3(Psi3) + (Psi3*(48 - 52*Psi4*t - 48*Exponential(Psi3*t)) + 52*Psi4*(-1 + Exponential(Psi3*t)) - 3*t*(-32 + Psi4*t)*Power2(Psi3))*Power4(Psi2) + 32*Psi4*(-1 + Exponential(Psi2*t))*Power4(Psi3) + 64*(-1 - Psi3*t + Exponential(Psi3*t))*Power5(Psi2)) + Power2(Psi2 - Psi3)*Power2(Psi3)*Power3(Psi2)*Power5(Psi1)*(20*Psi2*Psi4*(-150 + 53*Psi3*t + 2*(75 + 23*Psi3*t)*Exponential(Psi1*t))*Power2(Psi3) + 2*Psi3*Psi4*Power2(Psi2)*(-708 + 1095*Psi3*t - 130*Power2(Psi3)*Power2(t) + Exponential(Psi1*t)*(708 + 993*Psi3*t + 85*Power2(Psi3)*Power2(t))) + 720*Psi4*(-1 + Exponential(Psi1*t))*Power3(Psi3) + 2*Power3(Psi2)*(-36*Psi4*(-1 + Exponential(Psi1*t)) + 4*Psi3*(-12 + 85*Psi4*t + 2*(6 + 37*Psi4*t)*Exponential(Psi1*t)) + Psi4*(-171 + 100*Exponential(Psi1*t))*Power2(Psi3)*Power2(t) + 7*Psi4*Power3(Psi3)*Power3(t)) + 4*Psi3*Psi4*(-16 + 3*Psi3*t + 7*Exponential(Psi1*t))*Power2(t)*Power4(Psi2) + Psi3*Psi4*Power3(t)*Power5(Psi2)) - Power2(Psi2 - Psi3)*Power2(Psi3)*Power3(Psi2)*(2*Psi4*(-480 + 208*Psi3*t + 3*(160 + 51*Psi3*t)*Exponential(Psi1*t) - Psi3*t*Exponential(Psi2*t))*Power2(Psi3) + Psi2*Psi3*Psi4*(-492 + 1460*Psi3*t + 6*Psi3*t*Exponential(Psi2*t) - 205*Power2(Psi3)*Power2(t) + Exponential(Psi1*t)*(492 + 1534*Psi3*t + 137*Power2(Psi3)*Power2(t))) + 2*Power3(Psi2)*(-48 + 48*Psi3*t - 21*Psi4*t - 72*Psi3*Psi4*Power2(t) + Exponential(Psi1*t)*(48 - 15*Psi4*t + 50*Psi3*Psi4*Power2(t)) + 14*Psi4*Power2(Psi3)*Power3(t)) + 2*Power2(Psi2)*(-132*Psi4*(-1 + Exponential(Psi1*t)) + 3*Psi3*(-48 + 113*Psi4*t + 3*(16 + 41*Psi4*t)*Exponential(Psi1*t)) + 2*Psi4*(-112 + 85*Exponential(Psi1*t))*Power2(Psi3)*Power2(t) + 8*Psi4*Power3(Psi3)*Power3(t)) + 6*Psi3*Psi4*Power3(t)*Power4(Psi2))*Power6(Psi1) - 3*Psi4*Power2(Psi1)*Power2(Psi2 - Psi3)*(4*Psi2*Psi3*(-80 + 23*Psi3*t + (80 + 17*Psi3*t)*Exponential(Psi1*t)) + 504*(-1 + Exponential(Psi1*t))*Power2(Psi3) + Power2(Psi2)*(-20 + 24*Psi3*t - 3*Power2(Psi3)*Power2(t) + Exponential(Psi1*t)*(20 + 16*Psi3*t + Power2(Psi3)*Power2(t))))*Power3(Psi3)*Power6(Psi2) + 2*Power10(Psi1)*(-(Power2(Psi2)*(192*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(240 + 209*Psi4*t + (-240 + 211*Psi4*t)*Exponential(Psi2*t)) + t*(48 + 11*Psi4*t)*Power2(Psi3))*Power3(Psi3)) - Power2(Psi3)*Power3(Psi2)*(516*Psi4*(-1 + Exponential(Psi2*t)) - Psi3*(144 + 151*Psi4*t + (-144 + 41*Psi4*t)*Exponential(Psi2*t)) + 2*t*(144 + 7*Psi4*t)*Power2(Psi3) + Psi4*Power3(Psi3)*Power3(t)) + Psi3*(3*Psi3*(112 + 107*Psi4*t + (-48 + 49*Psi4*t)*Exponential(Psi2*t) - 64*Exponential(Psi3*t)) + 48*Psi4*(-1 + Exponential(Psi3*t)) + t*(432 + 61*Psi4*t)*Power2(Psi3) + 2*Psi4*Power3(Psi3)*Power3(t))*Power4(Psi2) + Psi2*(420*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(-48 + Psi4*t + (48 + 23*Psi4*t)*Exponential(Psi2*t)))*Power4(Psi3) - (-264*Psi4*(-1 + Exponential(Psi3*t)) + 24*Psi3*(-4 + 11*Psi4*t + 4*Exponential(Psi3*t)) + 12*t*(-16 + 3*Psi4*t)*Power2(Psi3) + Psi4*Power3(Psi3)*Power3(t))*Power5(Psi2) - 24*Psi4*(-1 + Exponential(Psi2*t))*Power5(Psi3) + 288*(-1 - Psi3*t + Exponential(Psi3*t))*Power6(Psi2)) + 12*Psi1*Psi4*(40*Psi3*(-1 + Exponential(Psi1*t)) + Psi2*(-10 + 3*Psi3*t + 2*(5 + Psi3*t)*Exponential(Psi1*t)))*Power2(Psi2 - Psi3)*Power4(Psi3)*Power7(Psi2) + Psi2*Power7(Psi1)*(2*Power3(Psi3)*(396*Psi4*(-1 + Exponential(Psi1*t)) - Psi3*(-96 + 53*Psi4*t + 2*(72 + 127*Psi4*t)*Exponential(Psi1*t) + (-48 + 65*Psi4*t)*Exponential(Psi2*t)) - 2*t*(-96 + Psi4*t*(-101 + 84*Exponential(Psi1*t)))*Power2(Psi3) + 7*Psi4*Power3(Psi3)*Power3(t))*Power4(Psi2) + Power3(Psi2)*(-24*Psi4*(4 + 15*Exponential(Psi1*t) - 19*Exponential(Psi2*t)) - 4*Psi3*(48 + 94*Psi4*t + 8*(-9 + 17*Psi4*t)*Exponential(Psi1*t) + (24 - 47*Psi4*t)*Exponential(Psi2*t)) + 2*Psi4*(-67 + 95*Exponential(Psi1*t))*Power2(Psi3)*Power2(t) + 9*Psi4*Power3(Psi3)*Power3(t))*Power4(Psi3) - Power2(Psi3)*(408*Psi4*(-1 + Exponential(Psi1*t)) - 4*Psi3*(72 + 76*Psi4*t + 8*(-9 + 22*Psi4*t)*Exponential(Psi1*t) + 3*Psi4*t*Exponential(Psi2*t)) + 2*t*(288 + Psi4*t*(26 + 33*Exponential(Psi1*t)))*Power2(Psi3) + 41*Psi4*Power3(Psi3)*Power3(t))*Power5(Psi2) + 2*Psi4*Power2(Psi2)*(360 + 110*Psi3*t - (348 + 41*Psi3*t)*Exponential(Psi2*t) - 42*Power2(Psi3)*Power2(t) + 3*Exponential(Psi1*t)*(-4 + 81*Psi3*t + 7*Power2(Psi3)*Power2(t)))*Power5(Psi3) + 2*Psi3*(-3*Psi3*(64 + 17*Psi4*t + (-48 + 23*Psi4*t)*Exponential(Psi1*t) - 16*Exponential(Psi3*t)) - 12*Psi4*(-1 + Exponential(Psi3*t)) + Psi4*(-76 + 85*Exponential(Psi1*t))*Power2(Psi3)*Power2(t) + 2*Psi4*Power3(Psi3)*Power3(t))*Power6(Psi2) + 12*Psi2*Psi4*(-28 + 3*Psi3*t + (28 + Psi3*t)*Exponential(Psi2*t))*Power6(Psi3) + 2*(12*Psi3*(4 + Psi4*t - 4*Exponential(Psi3*t)) - 12*Psi4*(-1 + Exponential(Psi3*t)) + 3*t*(32 + 3*Psi4*t)*Power2(Psi3) + 7*Psi4*Power3(Psi3)*Power3(t))*Power7(Psi2) - 48*Psi4*(-1 + Exponential(Psi2*t))*Power7(Psi3)) + Psi2*(-(Power3(Psi2)*Power3(Psi3)*(48*Psi4*(-28 + 9*Exponential(Psi1*t) + 19*Exponential(Psi2*t)) + Psi3*(-192 + 338*Psi4*t + (96 - 390*Psi4*t)*Exponential(Psi1*t) + (96 + 52*Psi4*t)*Exponential(Psi2*t)) - t*(-576 + Psi4*t*(-189 + 31*Exponential(Psi1*t)))*Power2(Psi3) + 14*Psi4*Power3(Psi3)*Power3(t))) + 2*Power2(Psi3)*(108*Psi4*(-1 + Exponential(Psi1*t)) + Psi3*(4*(-36 + 47*Psi4*t) - 3*(-80 + 79*Psi4*t)*Exponential(Psi1*t) + (-96 + 109*Psi4*t)*Exponential(Psi2*t)) + t*(192 + Psi4*t*(23 + 95*Exponential(Psi1*t)))*Power2(Psi3) + 9*Psi4*Power3(Psi3)*Power3(t))*Power4(Psi2) - 2*Power2(Psi2)*(-36*Psi4*(-8 + 3*Exponential(Psi1*t) + 5*Exponential(Psi2*t)) + 3*Psi3*(32 + 21*Psi4*t + (16 + 17*Psi4*t)*Exponential(Psi1*t) + (-48 + 46*Psi4*t)*Exponential(Psi2*t)) + 2*Psi4*(-17 + 21*Exponential(Psi1*t))*Power2(Psi3)*Power2(t) + Psi4*Power3(Psi3)*Power3(t))*Power4(Psi3) + Psi3*(6*Psi3*(112 + 22*Psi4*t + (-48 + 31*Psi4*t)*Exponential(Psi1*t) - Psi4*t*Exponential(Psi2*t) - 64*Exponential(Psi3*t)) + 96*Psi4*(-1 + Exponential(Psi3*t)) - t*(-864 + Psi4*t*(-133 + 137*Exponential(Psi1*t)))*Power2(Psi3) + 14*Psi4*Power3(Psi3)*Power3(t))*Power5(Psi2) + 2*Psi2*Psi4*(-264 + 55*Psi3*t + (264 + 65*Psi3*t)*Exponential(Psi2*t) + 7*Power2(Psi3)*Power2(t))*Power5(Psi3) - 8*(Psi3*(36 + 21*Psi4*t - 36*Exponential(Psi3*t)) - 21*Psi4*(-1 + Exponential(Psi3*t)) + 9*t*(8 + Psi4*t)*Power2(Psi3) + 2*Psi4*Power3(Psi3)*Power3(t))*Power6(Psi2) - 2*Psi4*(120 + 7*Psi3*t)*(-1 + Exponential(Psi2*t))*Power6(Psi3) + 96*(-1 - Psi3*t + Exponential(Psi3*t))*Power7(Psi2))*Power8(Psi1) - 60*Psi4*(-1 + Exponential(Psi1*t))*Power2(Psi2 - Psi3)*Power5(Psi3)*Power8(Psi2) + (2*Power3(Psi2)*Power3(Psi3)*(684*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(96 + 301*Psi4*t + (48 - 39*Psi4*t)*Exponential(Psi1*t) + 2*(-72 + 133*Psi4*t)*Exponential(Psi2*t)) + t*(192 + Psi4*t*(40 + 21*Exponential(Psi1*t)))*Power2(Psi3) + 2*Psi4*Power3(Psi3)*Power3(t)) + Power2(Psi3)*(456*Psi4*(-1 + Exponential(Psi2*t)) + 12*Psi3*(-8*(3 + 7*Psi4*t) + (-16 + 13*Psi4*t)*Exponential(Psi1*t) - 5*(-8 + 7*Psi4*t)*Exponential(Psi2*t)) - 2*t*(-192 + Psi4*t*(11 + 42*Exponential(Psi1*t)))*Power2(Psi3) + Psi4*Power3(Psi3)*Power3(t))*Power4(Psi2) - 2*Power2(Psi2)*(636*Psi4*(-1 + Exponential(Psi2*t)) - 2*Psi3*(72 + 11*Psi4*t + (-72 + 7*Psi4*t)*Exponential(Psi2*t)) + 5*Psi4*Power2(Psi3)*Power2(t))*Power4(Psi3) - 2*Psi3*(3*Psi3*(128 + 58*Psi4*t + (-16 + 13*Psi4*t)*Exponential(Psi1*t) + (-16 + 17*Psi4*t)*Exponential(Psi2*t) - 96*Exponential(Psi3*t)) + 72*Psi4*(-1 + Exponential(Psi3*t)) - 3*t*(-256 + Psi4*t*(-26 + 7*Exponential(Psi1*t)))*Power2(Psi3) + 7*Psi4*Power3(Psi3)*Power3(t))*Power5(Psi2) - 2*Psi2*Psi4*(-36 + 29*Psi3*t + (36 + 19*Psi3*t)*Exponential(Psi2*t))*Power5(Psi3) + 3*(-144*Psi4*(-1 + Exponential(Psi3*t)) - 16*Psi3*(-4 - 9*Psi4*t + 4*Exponential(Psi3*t)) + 4*t*(32 + 9*Psi4*t)*Power2(Psi3) + 3*Psi4*Power3(Psi3)*Power3(t))*Power6(Psi2) + 96*Psi4*(-1 + Exponential(Psi2*t))*Power6(Psi3) - 384*(-1 - Psi3*t + Exponential(Psi3*t))*Power7(Psi2))*Power9(Psi1)) + 105*Power4(phi)*(144*Power12(Psi1)*(-(Psi2*(9*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(4 + 3*Psi4*t + (-4 + 3*Psi4*t)*Exponential(Psi2*t)))*Power2(Psi3)) - (-4 + 4*Psi3*t - 6*Psi4*t + (4 - 3*Psi4*t)*Exponential(Psi2*t))*Power2(Psi2)*Power2(Psi3) + (Psi3*(4 - 3*Psi4*t - 4*Exponential(Psi3*t)) + 3*Psi4*(-1 + Exponential(Psi3*t)) + 8*t*Power2(Psi3))*Power3(Psi2) + 6*Psi4*(-1 + Exponential(Psi2*t))*Power3(Psi3) + 4*(-1 - Psi3*t + Exponential(Psi3*t))*Power4(Psi2)) - Psi4*Power2(Psi2 - Psi3)*Power3(Psi3)*Power4(Psi1)*Power4(Psi2)*(288*Psi2*Psi3*(29*(-3 + Psi3*t) + 3*(29 + 7*Psi3*t)*Exponential(Psi1*t)) + 11880*(-1 + Exponential(Psi1*t))*Power2(Psi3) + 24*Power2(Psi2)*(-318 + 432*Psi3*t - 58*Power2(Psi3)*Power2(t) + Exponential(Psi1*t)*(318 + 288*Psi3*t + 19*Power2(Psi3)*Power2(t))) + 24*t*(76 - 42*Psi3*t + 2*(22 + 5*Psi3*t)*Exponential(Psi1*t) + 3*Power2(Psi3)*Power2(t))*Power3(Psi2) - Power2(t)*(72 - 24*Psi3*t - 12*Exponential(Psi1*t) + Power2(Psi3)*Power2(t))*Power4(Psi2)) + 12*Psi4*Power2(Psi2 - Psi3)*Power3(Psi1)*(12*Psi2*Psi3*(-120 + 38*Psi3*t + 5*(24 + 5*Psi3*t)*Exponential(Psi1*t)) + 1200*(-1 + Exponential(Psi1*t))*Power2(Psi3) + 2*Power2(Psi2)*(-120 + 152*Psi3*t - 21*Power2(Psi3)*Power2(t) + Exponential(Psi1*t)*(120 + 88*Psi3*t + 5*Power2(Psi3)*Power2(t))) + t*(20 - 12*Psi3*t + 2*(5 + Psi3*t)*Exponential(Psi1*t) + Power2(Psi3)*Power2(t))*Power3(Psi2))*Power3(Psi3)*Power5(Psi2) - 144*Power11(Psi1)*(-(Power2(Psi2)*Power2(Psi3)*(33*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(4 + 3*Psi4*t + (-4 + 8*Psi4*t)*Exponential(Psi2*t)) + t*(8 + Psi4*t)*Power2(Psi3))) + Psi3*(Psi3*(16 + 21*Psi4*t + (-12 + 11*Psi4*t)*Exponential(Psi2*t) - 4*Exponential(Psi3*t)) + Psi4*(-1 + Exponential(Psi3*t)) + 2*Psi4*Power2(Psi3)*Power2(t))*Power3(Psi2) - Psi2*(-11*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(8 + 5*Psi4*t + (-8 + 3*Psi4*t)*Exponential(Psi2*t)))*Power3(Psi3) - (-13*Psi4*(-1 + Exponential(Psi3*t)) + Psi3*(-12 + 13*Psi4*t + 12*Exponential(Psi3*t)) + t*(-24 + Psi4*t)*Power2(Psi3))*Power4(Psi2) + 8*Psi4*(-1 + Exponential(Psi2*t))*Power4(Psi3) + 16*(-1 - Psi3*t + Exponential(Psi3*t))*Power5(Psi2)) - 2*Power2(Psi2 - Psi3)*Power2(Psi3)*Power3(Psi2)*Power5(Psi1)*(-36*Psi2*Psi4*(-250 + 94*Psi3*t + (250 + 71*Psi3*t)*Exponential(Psi1*t))*Power2(Psi3) - 24*Psi3*Psi4*Power2(Psi2)*(-177 + 297*Psi3*t - 41*Power2(Psi3)*Power2(t) + Exponential(Psi1*t)*(177 + 225*Psi3*t + 17*Power2(Psi3)*Power2(t))) - 2160*Psi4*(-1 + Exponential(Psi1*t))*Power3(Psi3) - 12*Power3(Psi2)*(-18*Psi4*(-1 + Exponential(Psi1*t)) + 6*Psi3*(-4 + 31*Psi4*t + (4 + 22*Psi4*t)*Exponential(Psi1*t)) + 2*Psi4*(-55 + 19*Exponential(Psi1*t))*Power2(Psi3)*Power2(t) + 7*Psi4*Power3(Psi3)*Power3(t)) + 2*Psi3*Psi4*Power2(t)*(126 - 36*Psi3*t - 30*Exponential(Psi1*t) + Power2(Psi3)*Power2(t))*Power4(Psi2) + Psi3*Psi4*(-6 + Psi3*t)*Power3(t)*Power5(Psi2)) + Power2(Psi2 - Psi3)*Power2(Psi3)*Power3(Psi2)*(-288*Psi4*(-20 + 9*Psi3*t + (20 + 6*Psi3*t)*Exponential(Psi1*t))*Power2(Psi3) - 36*Psi2*Psi3*Psi4*(-82 + 264*Psi3*t - 42*Power2(Psi3)*Power2(t) + Exponential(Psi1*t)*(82 + 236*Psi3*t + 19*Power2(Psi3)*Power2(t))) - 48*Power2(Psi2)*(-33*Psi4*(-1 + Exponential(Psi1*t)) + Psi3*(-36 + 93*Psi4*t + 12*(3 + 7*Psi4*t)*Exponential(Psi1*t)) + 2*Psi4*(-35 + 17*Exponential(Psi1*t))*Power2(Psi3)*Power2(t) + 4*Psi4*Power3(Psi3)*Power3(t)) + 8*Psi3*Psi4*(-9 + Psi3*t)*Power3(t)*Power4(Psi2) - 6*Power3(Psi2)*(-96 + 96*Psi3*t - 48*Psi4*t - 184*Psi3*Psi4*Power2(t) + 4*Exponential(Psi1*t)*(24 - 6*Psi4*t + 19*Psi3*Psi4*Power2(t)) + 56*Psi4*Power2(Psi3)*Power3(t) - Psi4*Power3(Psi3)*Power4(t)) + Psi3*Psi4*Power4(t)*Power5(Psi2))*Power6(Psi1) - 12*Psi4*Power2(Psi1)*Power2(Psi2 - Psi3)*(8*Psi2*Psi3*(-60 + 19*Psi3*t + (60 + 11*Psi3*t)*Exponential(Psi1*t)) + 756*(-1 + Exponential(Psi1*t))*Power2(Psi3) + Power2(Psi2)*(-30 + 40*Psi3*t - 6*Power2(Psi3)*Power2(t) + Exponential(Psi1*t)*(30 + 20*Psi3*t + Power2(Psi3)*Power2(t))))*Power3(Psi3)*Power6(Psi2) + Power10(Psi1)*(-48*Power2(Psi2)*(48*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(60 + 52*Psi4*t + (-60 + 53*Psi4*t)*Exponential(Psi2*t)) + 4*t*(3 + Psi4*t)*Power2(Psi3))*Power3(Psi3) - 24*Power2(Psi3)*Power3(Psi2)*(258*Psi4*(-1 + Exponential(Psi2*t)) - 2*Psi3*(36 + 41*Psi4*t + (-36 + 7*Psi4*t)*Exponential(Psi2*t)) + 8*t*(18 + Psi4*t)*Power2(Psi3) + Psi4*Power3(Psi3)*Power3(t)) + 48*Psi2*(105*Psi4*(-1 + Exponential(Psi2*t)) + Psi3*(-12 - Psi4*t + (12 + 7*Psi4*t)*Exponential(Psi2*t)))*Power4(Psi3) + Psi3*Power4(Psi2)*(144*Psi3*(28 + 26*Psi4*t + (-12 + 13*Psi4*t)*Exponential(Psi2*t) - 16*Exponential(Psi3*t)) + 576*Psi4*(-1 + Exponential(Psi3*t)) + 192*t*(27 + 5*Psi4*t)*Power2(Psi3) + 48*Psi4*Power3(Psi3)*Power3(t) + Psi4*Power4(Psi3)*Power4(t)) - 2*(-1584*Psi4*(-1 + Exponential(Psi3*t)) + 144*Psi3*(-4 + 11*Psi4*t + 4*Exponential(Psi3*t)) + 288*t*(-4 + Psi4*t)*Power2(Psi3) + 12*Psi4*Power3(Psi3)*Power3(t) + Psi4*Power4(Psi3)*Power4(t))*Power5(Psi2) - 288*Psi4*(-1 + Exponential(Psi2*t))*Power5(Psi3) + (-3456 - 3456*Psi3*t + 3456*Exponential(Psi3*t) + Psi4*Power3(Psi3)*Power4(t))*Power6(Psi2)) + 120*Psi1*Psi4*(24*Psi3*(-1 + Exponential(Psi1*t)) + Psi2*(-6 + 2*Psi3*t + (6 + Psi3*t)*Exponential(Psi1*t)))*Power2(Psi2 - Psi3)*Power4(Psi3)*Power7(Psi2) + Psi2*(-24*Power2(Psi2)*(-18*Psi4*(-8 + 3*Exponential(Psi1*t) + 5*Exponential(Psi2*t)) + 6*Psi3*(8 + 6*Psi4*t + 4*(1 + Psi4*t)*Exponential(Psi1*t) + (-12 + 11*Psi4*t)*Exponential(Psi2*t)) + 18*Psi4*(-1 + Exponential(Psi1*t))*Power2(Psi3)*Power2(t) + Psi4*Power3(Psi3)*Power3(t))*Power4(Psi3) + Power3(Psi2)*Power3(Psi3)*(-288*Psi4*(-28 + 9*Exponential(Psi1*t) + 19*Exponential(Psi2*t)) + 48*Psi3*(24 - 32*Psi4*t + 3*(-4 + 15*Psi4*t)*Exponential(Psi1*t) - (12 + 13*Psi4*t)*Exponential(Psi2*t)) + 36*t*(-96 + Psi4*t*(-38 + 5*Exponential(Psi1*t)))*Power2(Psi3) - 168*Psi4*Power3(Psi3)*Power3(t) + Psi4*Power4(Psi3)*Power4(t)) + 6*Power2(Psi3)*Power4(Psi2)*(216*Psi4*(-1 + Exponential(Psi1*t)) - 24*Psi3*(12 - 13*Psi4*t + 2*(-10 + 9*Psi4*t)*Exponential(Psi1*t) + (8 - 10*Psi4*t)*Exponential(Psi2*t)) + 4*t*(96 + Psi4*t*(20 + 39*Exponential(Psi1*t)))*Power2(Psi3) + 36*Psi4*Power3(Psi3)*Power3(t) + Psi4*Power4(Psi3)*Power4(t)) - 3*Psi3*(-48*Psi3*(28 + 6*Psi4*t + (-12 + 7*Psi4*t)*Exponential(Psi1*t) - 16*Exponential(Psi3*t)) - 192*Psi4*(-1 + Exponential(Psi3*t)) + 12*t*(-144 + Psi4*t*(-26 + 19*Exponential(Psi1*t)))*Power2(Psi3) - 56*Psi4*Power3(Psi3)*Power3(t) + 3*Psi4*Power4(Psi3)*Power4(t))*Power5(Psi2) + 96*Psi2*Psi4*(-33 + 6*Psi3*t + (33 + 9*Psi3*t)*Exponential(Psi2*t) + Power2(Psi3)*Power2(t))*Power5(Psi3) - 4*(-252*Psi4*(-1 + Exponential(Psi3*t)) - 36*Psi3*(-12 - 7*Psi4*t + 12*Exponential(Psi3*t)) + 144*t*(6 + Psi4*t)*Power2(Psi3) + 48*Psi4*Power3(Psi3)*Power3(t) + Psi4*Power4(Psi3)*Power4(t))*Power6(Psi2) - 96*Psi4*(15 + Psi3*t)*(-1 + Exponential(Psi2*t))*Power6(Psi3) + 6*(-96 - 96*Psi3*t + 96*Exponential(Psi3*t) + Psi4*Power3(Psi3)*Power4(t))*Power7(Psi2))*Power8(Psi1) - 360*Psi4*(-1 + Exponential(Psi1*t))*Power2(Psi2 - Psi3)*Power5(Psi3)*Power8(Psi2) - 4*Psi2*Power7(Psi1)*(-3*Power3(Psi2)*(-12*Psi4*(4 + 15*Exponential(Psi1*t) - 19*Exponential(Psi2*t)) - 2*Psi3*(48 + 106*Psi4*t + 3*(-24 + 43*Psi4*t)*Exponential(Psi1*t) + (24 - 52*Psi4*t)*Exponential(Psi2*t)) + 2*Psi4*(-40 + 39*Exponential(Psi1*t))*Power2(Psi3)*Power2(t) + 9*Psi4*Power3(Psi3)*Power3(t))*Power4(Psi3) + Power3(Psi3)*Power4(Psi2)*(-1188*Psi4*(-1 + Exponential(Psi1*t)) + 36*Psi3*(-8 + 7*Psi4*t + (12 + 19*Psi4*t)*Exponential(Psi1*t) + (-4 + 5*Psi4*t)*Exponential(Psi2*t)) + 6*t*(-96 + Psi4*t*(-121 + 71*Exponential(Psi1*t)))*Power2(Psi3) - 42*Psi4*Power3(Psi3)*Power3(t) + Psi4*Power4(Psi3)*Power4(t)) + Power2(Psi3)*(612*Psi4*(-1 + Exponential(Psi1*t)) - 18*Psi3*(8*(3 + 4*Psi4*t) + (-24 + 53*Psi4*t)*Exponential(Psi1*t)) + 6*t*(144 + Psi4*t*(16 + 11*Exponential(Psi1*t)))*Power2(Psi3) + 123*Psi4*Power3(Psi3)*Power3(t) + Psi4*Power4(Psi3)*Power4(t))*Power5(Psi2) - 6*Psi4*Power2(Psi2)*(180 + 68*Psi3*t - 2*(87 + 13*Psi3*t)*Exponential(Psi2*t) - 25*Power2(Psi3)*Power2(t) + 3*Exponential(Psi1*t)*(-2 + 38*Psi3*t + 3*Power2(Psi3)*Power2(t)))*Power5(Psi3) - 4*Psi3*(-9*Psi3*(16 + 5*Psi4*t + (-12 + 5*Psi4*t)*Exponential(Psi1*t) - 4*Exponential(Psi3*t)) - 9*Psi4*(-1 + Exponential(Psi3*t)) + 3*Psi4*(-23 + 17*Exponential(Psi1*t))*Power2(Psi3)*Power2(t) + 3*Psi4*Power3(Psi3)*Power3(t) + Psi4*Power4(Psi3)*Power4(t))*Power6(Psi2) - 24*Psi2*Psi4*(-21 + 2*Psi3*t + (21 + Psi3*t)*Exponential(Psi2*t))*Power6(Psi3) + (36*Psi4*(-1 + Exponential(Psi3*t)) + 36*Psi3*(-4 - Psi4*t + 4*Exponential(Psi3*t)) - 36*t*(8 + Psi4*t)*Power2(Psi3) - 42*Psi4*Power3(Psi3)*Power3(t) + Psi4*Power4(Psi3)*Power4(t))*Power7(Psi2) + 72*Psi4*(-1 + Exponential(Psi2*t))*Power7(Psi3) + Psi4*Power3(Psi3)*Power4(t)*Power8(Psi2)) - 2*(-12*Power3(Psi2)*Power3(Psi3)*(342*Psi4*(-1 + Exponential(Psi2*t)) - 6*Psi3*(-8 - 23*Psi4*t + (-4 + 3*Psi4*t)*Exponential(Psi1*t) + (12 - 24*Psi4*t)*Exponential(Psi2*t)) + 3*t*(32 + 3*Psi4*t*(3 + Exponential(Psi1*t)))*Power2(Psi3) + 2*Psi4*Power3(Psi3)*Power3(t)) + 24*Power2(Psi2)*(159*Psi4*(-1 + Exponential(Psi2*t)) + 9*Psi3*(-4 - Psi4*t + 4*Exponential(Psi2*t)) + Psi4*Power2(Psi3)*Power2(t))*Power4(Psi3) + Power2(Psi3)*Power4(Psi2)*(-1368*Psi4*(-1 + Exponential(Psi2*t)) - 24*Psi3*(-36 - 83*Psi4*t + 6*(-4 + 3*Psi4*t)*Exponential(Psi1*t) + (60 - 52*Psi4*t)*Exponential(Psi2*t)) + 72*t*(-16 + Psi4*t*(2 + 3*Exponential(Psi1*t)))*Power2(Psi3) - 6*Psi4*Power3(Psi3)*Power3(t) + Psi4*Power4(Psi3)*Power4(t)) + 12*Psi3*(6*Psi3*(32 + 14*Psi4*t + (-4 + 3*Psi4*t)*Exponential(Psi1*t) + (-4 + 5*Psi4*t)*Exponential(Psi2*t) - 24*Exponential(Psi3*t)) + 36*Psi4*(-1 + Exponential(Psi3*t)) - t*(-384 + Psi4*t*(-49 + 9*Exponential(Psi1*t)))*Power2(Psi3) + 7*Psi4*Power3(Psi3)*Power3(t))*Power5(Psi2) + 24*Psi2*Psi4*(-9 + 7*Psi3*t + (9 + 5*Psi3*t)*Exponential(Psi2*t))*Power5(Psi3) - 3*(-432*Psi4*(-1 + Exponential(Psi3*t)) - 48*Psi3*(-4 - 9*Psi4*t + 4*Exponential(Psi3*t)) + 48*t*(8 + 3*Psi4*t)*Power2(Psi3) + 18*Psi4*Power3(Psi3)*Power3(t) + Psi4*Power4(Psi3)*Power4(t))*Power6(Psi2) - 288*Psi4*(-1 + Exponential(Psi2*t))*Power6(Psi3) + 2*(-576 - 576*Psi3*t + 576*Exponential(Psi3*t) + Psi4*Power3(Psi3)*Power4(t))*Power7(Psi2))*Power9(Psi1))))/Psi4;
}
